/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_COLLECTION_H
#define O2SCL_COLLECTION_H

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <map>
#include <set>
#include <sstream>
#include <o2scl/misc.h>
#include <o2scl/err_hnd.h>
#include <o2scl/test_mgr.h>
#include <o2scl/text_file.h>
#include <o2scl/file_format.h>
#include <o2scl/file_detect.h>
#include <o2scl/ovector_tlate.h>
#include <o2scl/omatrix_tlate.h>
#include <o2scl/ovector_cx_tlate.h>
#include <o2scl/omatrix_cx_tlate.h>
/* \file collection.h
   \brief Brief desc.
*/

#ifndef DOXYGENP
namespace o2scl {
#endif

  /// An entry in a collection
  typedef struct {
    /// The pointer to the object
    void *data;
    /// The first size parameter
    int size;
    /// The second size parameter
    int size2;
    /// True if the collection owns this object
    bool owner;
    /// A pointer to the corresponding \ref io_base object
    class io_base *iop;
  } collection_entry;

  /// A pointer output structure
  typedef struct {
    /// The name of the pointer
    std::string name;
    /// Pointer to the collection entry
    collection_entry *ep;
    /// True if the pointer has been written to the file
    bool output;
  } pointer_output;
  
  /// A pointer input structure
  typedef struct {
    /// The name of the pointer
    std::string name;
    /// The pointer
    void **ptr;
    /// The type of the object pointed to
    std::string stype;
  } pointer_input;

  class cinput;
  class coutput;

  /** 
      \brief I/O base class



      This class is necessary so that the collection method source code
      and the io_base method source code doesn't have to go in header files.
      
      \todo Should the remove() functions be moved to class collection?
  */
  class io_base {
   
#ifndef DOXYGEN_INTERNAL

  protected:
    /// for io_type_info.add_type().
    friend class io_type_info;
   
    /// for stat_in and stat_out
    friend class collection;

    friend class cinput;
    friend class coutput;
    friend class io_manager;

    /// A pointer to the type manager
    static class io_manager *iom;
    
    /// A count of the number of objects
    static int objs_count;

    /// Automatically create an object for stat_in
    virtual int stat_in_noobj(cinput *co, in_file_format *ins);
  
    /// Automatically create an object for stat_out
    virtual int stat_out_noobj(coutput *co, out_file_format *outs);
  
    /** \brief Allocate memory and input an object
     */
    virtual int in_wrapper(cinput *co, in_file_format *ins, void *&vp);
    
    /** \brief Allocate memory and input an array of objects
    */
    virtual int in_wrapper(cinput *co, in_file_format *ins, void *&vp, 
			   int &sz);

    /** \brief Allocate memory and input a 2-d array of objects 
    */
    virtual int in_wrapper(cinput *co, in_file_format *ins, void *&vp, 
			   int &sz, int &sz2);
    
    /** \brief Internal function to output an object (or an array
	or 2-d array) 
    */
    virtual int out_wrapper(coutput *co, out_file_format *outs, 
			    void *vp, int sz, int sz2);

    /// Input an object (no memory allocation)
    virtual int object_in_void(cinput *cin, in_file_format *ins, void *op,
			       std::string &name);
    
    /// Input an array of objects (no memory allocation)
    virtual int object_in_void(cinput *cin, in_file_format *ins, void *op, 
			       int sz, std::string &name);
    
    /// Input a 2-d array of objects (no memory allocation)
    virtual int object_in_void(cinput *cin, in_file_format *ins, void *op, 
			       int sz, int sz2, std::string &name);
    
    /// Input an object (no memory allocation)
    virtual int object_in_mem_void(cinput *cin, in_file_format *ins, 
				   void *&op, std::string &name);
    
    /// Input an array of objects (no memory allocation)
    virtual int object_in_mem_void(cinput *cin, in_file_format *ins, 
				   void *&op, int &sz, std::string &name);
    
    /// Input a 2-d array of objects (no memory allocation)
    virtual int object_in_mem_void(cinput *cin, in_file_format *ins, 
				   void *&op, int &sz, int &sz2,
				   std::string &name);
    
    /// Output an object, an array of objects, or a 2-d array of objects
    virtual int object_out_void(coutput *cout, out_file_format *outs, 
				void *op, int sz, int sz2, 
				std::string name="");

    /** \brief Store the value of \c sw given in the constructor so
	that we know if we need to remove the type in the destructor
    */
    int sw_store;

    /** \name Functions to remove the memory that was allocated for an object
     */
    //@{
    /// Remove the memory for an object
    virtual int remove(void *vp) { return 0; }
    
    /// Remove the memory for an array of objects
    virtual int remove_arr(void *vp) { 
      return 0; 
    }

    /// Remove the memory for a 2-dimensional array of objects
    virtual int remove_2darr(void *vp, int sz) { return 0; }
    //@}

#endif
  
  public:

    /** 
	\brief Create a new I/O object 
	
	If \c sw is different from zero, then the type will not
	be added to the io_manager. This is useful if you want
	an object to be its own I/O class, in which case you
	may want to make sure that the io_manager only tries
	to add the type once. There is no need to have an I/O 
	object for every instance of a particular type.
    */
    io_base(int sw=0);

    /** \brief Create a new object only if an I/O object for type \c t
	is not yet present
    */
    io_base(const char *t);

    virtual ~io_base();
    
    /// \name Functions to be overloaded in descendants of io_base
    //@{
    /// Return the type of an object
    virtual const char *type() { return "io_base"; }
    
    /// If true, then the object contains static data
    virtual bool has_static_data() { return false; }
    //@}
    
    /** \name Functions useful for in in() and out()
     */
    //@{
    /// Input a pointer
    virtual int pointer_in(cinput *co, in_file_format *ins, void **pp, 
			   std::string &stype);

    /** 
	\brief Output an object to \c outs of type \c stype 

	This is useful for to output a pointer to an object in the
	out() or stat_out() functions for a class. The data for
	the object which is pointed to is separate from the object
	and is only referred to once if more than one objects
	point to it.
    */
    virtual int pointer_out(coutput *co, out_file_format *outs, void *ptr, 
			    std::string stype);
    //@}
    
  };
 
  /** 
      \brief Manage I/O type information


      
      This class is automatically created, utilized, and destroyed by
      \ref io_base.
  */
  class io_manager {

  public:

    /** 
	\brief Add a type to the list

	Unfortunately, add_type() cannot ensure that no type is added
	more than once, since the type is not specified until the 
	entire constructor hierarchy has been executed and add_type()
	is called at the top of this hierarchy.
    */
    int add_type(io_base *iop);

    /// Return 0 if \c iop points to a valid type
    int is_type(io_base *iop);

    /// Add type \c iop to the manager assuming the type name \c t
    int add_type(io_base *iop, const char *t);

    /// Remove a type from the list
    int remove_type(io_base *iop);

#ifndef DOXYGEN_INTERNAL

  protected:

    /// So that io_base can access add_type().
    friend class io_base;

    /// So that collections can access get_ptr
    friend class collection;

    /// So that cinputs can access get_ptr()
    friend class cinput;

    /// So that coutputs can access get_ptr()
    friend class coutput;

    // So that io_type_info can access tlist
    friend class io_type_info;

    /// Get a pointer to type \c stype 
    io_base *get_ptr(std::string stype);
  
    /// The list of types in the form of io_base pointers
    std::vector<io_base *> tlist;
  
    /// A useful definition for iterating through types
    typedef std::vector<io_base *>::iterator titer;

    /// Empty constructor
    io_manager() {};
  
#endif

  };

  // ----------------------------------------------------------------
  // Start of class io_type_info
  // ----------------------------------------------------------------

  /** 
      \brief User interface to provide I/O type information


  */
  class io_type_info : public io_base {

#ifndef DOXYGEN_INTERNAL

  protected:

    /// So that collection::fout() can access static_fout().
    friend class collection;
    
    /// Output the static information for the I/O types
    int static_fout(coutput *co, out_file_format *out);
    
    /// Output the static information for the I/O types not in the list
    int static_fout_restricted(coutput *co, out_file_format *out,
			       std::set<std::string,string_comp> list);
  
    /// A useful definition for iterating through types
    typedef std::vector<io_base *>::iterator titer;

#endif
  
  public:

    io_type_info();

    virtual ~io_type_info();

    /** \name Type manipulation 
     */
    //@{

    /// Return 0 if \c stype is a valid I/O type
    int is_type(std::string stype);

    /** 
	\brief Remove \c stype from the list of valid I/O types

	This method is dangerous, as it can't check to ensure that
	no collection has remaining objects of the type to be
	removed.
     */
    int remove_type(std::string stype);

    /** 
	\brief Remove all types in the list of valid I/O types

	This method is dangerous as it doesn't ensure that
	all collections are empty.
    */
    virtual int clear_types();

    /// Print a summary of valid types to the \c outs stream
    void type_summary(std::ostream *outs, bool pointers=false);

    /// Add an I/O type to the list
    int add_type(io_base *iop) { return iom->add_type(iop); }
    //@}

  };
  
  // ----------------------------------------------------------------
  // End of class io_type_info
  // ----------------------------------------------------------------

  // ----------------------------------------------------------------
  // Start of class collection
  // ----------------------------------------------------------------

  /** 
      \brief Collection of objects



      By default, the fout() functions alphabetize the objects by name,
      but this is not a requirement for files read using fin().
    
      Important issues:
      1. Pointers are not set until after an entire file is read so that
      objects that are pointed to may occur anywhere in a file. This
      means that the information that is pointed to cannot be used in
      the io_tlate_d::input() function.
      
      \todo
      - If pointer_in gets a null pointer it does nothing. Should we
      replace this behaviour by two pointer_in() functions. One which
      does nothing if it gets a null pointer, and one which will
      go ahead and set the pointer to null. This is useful for 
      output object which have default values to be used if
      they are given a null pointer.
      - More testing on rewrite() function.
      - Think more about adding arrays of pointers? pointers to arrays?
      - Modify static data output so that if no objects of a type
      are included, then no static data is output for that type?
      (No, it's too hard to go through all objects looking for
      an object of a particular type).
      
      \bug
      - Ensure that the user cannot add a object with a name of ptrXXX.
      - Test_type does not test handle static data or pointers.
      - Check strings and words for characters that we can't handle
      - The present version of a text-file requires
      strings to contain at least one printable character.
      - Ensure that all matching is done by both type and name if 
      possible.

      Structure:
      collection::fout() does the following:
      - create an object of type coutput
      - Add all objects in the list to the pointer map \c ptr_map 
      (with \c output=true ) so that they can be referred to by
      pointers later
      - Output all static data using io_type_info::static_fout()
      - Output all of the items in the list \c plist  (see
      below). Any pointers which are not already in \c ptr_map  are
      added at this point (with \c output=false 
      - Call coutput::pointer_map_fout() to output all objects that
      were referred to but not in the list

      To output individual items, collection::fout() does the following:
      - Call either io_base::out_wrapper() or io_base::out_hc_wrapper()
      - In turn, these functions call io_base::output(), which the user
      has overloaded
      - If the function io_base::output() calls io_tlate::object_out() then the
      io_base::output() function appropriate for that object is called.
      No type or name information is included, but size integers are
      included if the object is a 1- or 2-d array.
      - If the function io_base::output() calls io_base::pointer_out(), then 
      the object is searched for in the \c ptr_map . If it is 
      not there, then the object is added and assigned a name. The
      type and name are then output. If the pointer is NULL, then
      both the type and the name are set to \c null .

  */
  class collection : public io_base {

#ifndef DOXYGEN_INTERNAL
  
  protected:

    friend class io_base;
    friend class cinput;
    friend class coutput;
  
    /// A convenient iterator definition for the collection
    typedef std::map<std::string,collection_entry,string_comp>::iterator 
      piter;
  
    /// The actual collection
    std::map<std::string,collection_entry,string_comp> plist;

#endif

  public:
    
    collection() : io_base(1) {
    };
    ~collection();

    /** \name Output to file methods */
    //@{

    /// Output entire list to \c outs 
    int fout(out_file_format *outs);

    /// Output entire list to a textfile named \c filename
    int fout(std::string filename);
    //@}
    
    /** \name Input from file methods 
	If \c overwrt is true, then any objects which
	already exist with the same name are overwritten
	with the objects in the file. The collection
	owns all the objects read. (Since it created them,
	the collection assumes it ought to be responsible
	to destroy them.)
    */
    //@{
    /// Read a collection from text file named \c file_name
    int fin(std::string file_name, bool overwrt=false, int verbose=0);

    /// Read a collection from \c ins
    int fin(in_file_format *ins, bool overwrt=false, int verbose=0);
    //@}

    /** \name Miscellaneous methods
     */
    //@{

    /// Test the output for type \c stype.
    int test_type(o2scl::test_mgr &t, std::string stype, void *obj,
		  void *&newobj, bool scrout=false);

    /** 
	\brief Update a file containing a collection
 
	This method loads the file from "fin" and produces a file at
	"fout" containing all of the objects from "fin", updated by
	their new values in the present list if possible. Then, it adds
	to the end of "fout" any objects in the present list that were
	not originally contained in "fin".
    */
    int rewrite(std::string in_name, std::string out_name);

    /** \brief Force the collection to assume that the ownership
	of \c name is external.

	This allows the user to take over ownership of the object
	named \c name. This is particularly useful if the object is
	read from a file (since then object is owned initially by the
	collection), and you want to delete the collection, but retain
	the object.  
    */
    int disown(std::string name);

    /** \brief Summarize contents of collection
     */
    int summary(std::ostream *out, bool show_addresses=false);

    /** 
	\brief Remove an object for the collection

	Free the memory \c name if it is owned by the collection 
	and then remove it from the collection.  
    */
    int remove(std::string name);
    
    /** \brief Remove all objects from the list
     */
    void clear();

    /** 
	\brief Count number of objects */
    int npa();

    //@}

    /** \name Generic add methods
	
	If \c overwrt is true, then any objects which already exist
	with the same name as \c name are overwritten. If owner=true,
	then the collection will own the memory allocated for the
	object and will free that memory with delete when the object
	is removed or the collection is deleted.
    */
    //@{
    int add(std::string name, io_base *tio, void *vec, int sz=0, int sz2=0,
	    bool overwrt=true, bool owner=false);
    
    int add(std::string name, std::string stype, 
	    void *vec, int sz=0, int sz2=0,
	    bool overwrt=true, bool owner=false);
    //@}

    /** 
	\name Generic get methods 
    */
    //@{
    /** 
	\brief Get an object
    */
    int get(std::string tname, void *&vec);
    /** 
	\brief Get an array of objects
    */
    int get(std::string tname, void *&vec, int &sz);
    /** 
	\brief Get a 2-d array of objects
    */
    int get(std::string tname, void *&vec, int &sz, int &sz2);
    /** 
	\brief Get an object and its type
    */
    int get(std::string tname, std::string &stype, void *&vec);
    /** 
	\brief Get an array of objects and their type
    */
    int get(std::string tname, std::string &stype, void *&vec, int &sz);
    /** 
	\brief Get a 2-d array of objects and their type
    */
    int get(std::string tname, std::string &stype, void *&vec, int &sz, 
	    int &sz2);
    /** 
	\brief Get an object (alternative form)
    */
    void *get(std::string name) {
      void *vp;
      get(name,vp);
      return vp;
    }
    //@}

    /// Output object of type \c stype and name \c name to output \c tof
    int get_type(text_out_file &tof, std::string stype, std::string name);

    /// Output object with name \c name to output \c tof
    int get(text_out_file &tof, std::string &stype, std::string name);
    
    /// Set object named \c name with input from \c tif
    int set(std::string name, text_in_file &tif);

    /// Set object named \c name with input from \c val
    int set(std::string name, std::string val);

    /** \name Input and output of individual objects */
    //@{
    /** 
	\brief Output one object to a file

	This does not disturb any objects in the collection. The 
	pointer specified does not need to be in the collection and
	is not added to the collection.
    */
    int out_one(out_file_format *outs, std::string stype, std::string name,
    		void *vp, int sz=0, int sz2=0);

    /** 
	\brief Output one object to a file

	This does not disturb any objects in the collection. The 
	pointer specified does not need to be in the collection and
	is not added to the collection.
    */
    int out_one(std::string fname, std::string stype, std::string name,
    		void *vp, int sz=0, int sz2=0);

    /** 
	\brief Input one object from a file with name \c name 
      
	This does not disturb any objects in the collection. The
	pointer specified does not need to be in the collection and is
	not added to the collection.
    */
    int in_one_name(in_file_format *ins, std::string stype, std::string name,
		    void *&vp, int &sz, int &sz2);
  
    /** 
	\brief Input one object from a file
      
	This does not disturb any objects in the collection. The
	pointer specified does not need to be in the collection and is
	not added to the collection.
    */
    int in_one(in_file_format *ins, std::string stype, std::string &name,
	       void *&vp, int &sz, int &sz2);

    /** 
	\brief Input one object from a file
      
	This does not disturb any objects in the collection. The
	pointer specified does not need to be in the collection and is
	not added to the collection.
    */
    int in_one(std::string fname, std::string stype, std::string &name,
	       void *&vp, int &sz, int &sz2);
    //@}

    /** 
	\brief An iterator for stepping through a collection
     */
    class iterator {
      
#ifndef DOXYGEN_INTERNAL

    protected:
      
      friend class collection;

      /// Create an iterator from the STL iterator
      iterator(piter p) { pit=p; }
      
      /// Local storage for the STL iterator
      piter pit;

#endif
      
    public:
      
      /// Prefix increment
      iterator operator++() {
	pit++;
	return *this;
      }

      /// Postfix increment
      iterator operator++(int unused) {
	pit++;
	return *this;
      }

      /// Prefix decrement
      iterator operator--() {
	pit--;
	return *this;
      }
      
      /// Dereference
      collection_entry *operator->() const {
	return &(pit->second);
      };

      /// Return the name of the collection entry
      std::string name() {
	return pit->first;
      }

      friend int operator==(const iterator &i1, const iterator &i2);
      
      friend int operator!=(const iterator &i1, const iterator &i2);
      
    };

    /** 
	\brief An iterator for stepping through the entries
	in a collection of a particular type
     */
    class type_iterator {
      
#ifndef DOXYGEN_INTERNAL

    protected:
      
      friend class collection;
      
      /// Local storage for the type
      std::string ltype;

      /// Store a pointer to the collection
      collection *lcop;
      
      /// Constructor
      type_iterator(piter p, std::string type, collection *cop) { 
	pit=p; 
	ltype=type;
	lcop=cop;
	while (pit!=lcop->plist.end() && 
	       pit->second.iop->type()!=ltype) pit++;
      }
      
      /// The STL iterator
      piter pit;

#endif

    public:

      /// Prefix increment
      type_iterator operator++() {
	if (pit!=lcop->plist.end()) pit++;
	while (pit!=lcop->plist.end() && 
	       pit->second.iop->type()!=ltype) pit++;
	return *this;
      }

      /// Postfix increment
      type_iterator operator++(int unused) {
	if (pit!=lcop->plist.end()) pit++;
	while (pit!=lcop->plist.end() && 
	       pit->second.iop->type()!=ltype) pit++;
	return *this;
      }

      /// Dereference
      collection_entry *operator->() const {
	return &(pit->second);
      };

      /// Return the name of the collection entry
      std::string name() {
	return pit->first;
      }

      friend int operator==(const type_iterator &i1, const type_iterator &i2);
      
      friend int operator!=(const type_iterator &i1, const type_iterator &i2);
      
    };

    /** \name Iterator functions
     */
    //@{
    /// Return an iterator to the start of the collection
    iterator begin() {
      return iterator(plist.begin());
    }

    /// Return an iterator to the end of the collection
    iterator end() {
      return iterator(plist.end());
    }

    /** \brief Return an iterator to the first element of type \c utype 
	in the collection
    */
    type_iterator begin(std::string utype) {
      return type_iterator(plist.begin(),utype,this);
    }
    
    /// Return an iterator to the end of the collection
    type_iterator end(std::string utype) {
      return type_iterator(plist.end(),utype,this);
    }
    //@}
    
  };
  // ----------------------------------------------------------------
  // End of class collection
  // ----------------------------------------------------------------

  /// Equality comparison for two iterators
  int operator==(const collection::iterator &i1, 
		 const collection::iterator &i2);
  
  /// Inequality comparison for two iterators
  int operator!=(const collection::iterator &i1, 
		 const collection::iterator &i2);

  /// Equality comparison for two iterators
  int operator==(const collection::type_iterator &i1, 
		 const collection::type_iterator &i2);
  
  /// Inequality comparison for two iterators
  int operator!=(const collection::type_iterator &i1, 
		 const collection::type_iterator &i2);

  /** 
      \brief Class to control object input


  */
  class cinput {

  public:

    /// Input an object
    int object_in(std::string type, in_file_format *ins, void *vp,
		  std::string &name);

    /// Input an array of objects
    int object_in(std::string type, in_file_format *ins, void *vp,
		  int sz, std::string &name);

    /// Input a 2-d array of objects
    int object_in(std::string type, in_file_format *ins, void *vp,
		  int sz, int sz2, std::string &name);
    
    /// Input an object, allocating memory first
    int object_in_mem(std::string type, in_file_format *ins, void *&vp,
		      std::string &name);
    
    /// Input an array of objects, allocating memory first
    int object_in_mem(std::string type, in_file_format *ins, void *&vp,
		      int &sz, std::string &name);
    
    /// Input a 2-d array of objects, allocating memory first
    int object_in_mem(std::string type, in_file_format *ins, void *&vp,
		      int &sz, int &sz2, std::string &name);

#ifndef DOXYGEN_INTERNAL

  protected:

    /// Create a new input object for a collection 
    cinput(collection *co) {
      cop=co;
    }

    friend class collection;
    friend class io_base;
  
    /// The pointers that need to be set
    std::vector<pointer_input> input_ptrs;

    /// An iterator for the input pointers
    typedef std::vector<pointer_input>::iterator ipiter;
  
    /// Assign all of the pointers read with the appropriate objects
    int assign_pointers(collection *co);

    /// The pointer to the collection stored in the constructor
    collection *cop;

#endif

  };

  /** 
      \brief Class to control object output


  */
  class coutput {

  public:

    /// Output an object
    int object_out(std::string type, out_file_format *outs,
		   void *op, int sz=0, int sz2=0, std::string name="");

#ifndef DOXYGEN_INTERNAL

  protected:

    friend class collection;
    friend class io_base;
  
    /// Create a new object from a pointer to a collection
    coutput(class collection *co) {
      cop=co;
      npointers=0;
    };

    /// Order the pointers by numeric value
    struct ltptr {
      /// Returns  \f$ p_1 < p_2 \f$ 
      bool operator()(const void * p1, const void * p2) const {
	return p1<p2;
      }
    };
    
    /// The list pointers to object to be written to the file
    std::map<void *,pointer_output,ltptr> ptr_map;
    
    /// A convenient iterator for the pointer list
    typedef std::map<void *,pointer_output,ltptr>::iterator pmiter;
    
    /** 
	\brief Look for an object in the collection given a pointer
	
	Lookup the pointer \c vp  in the collection, and return
	its name and collection_entry
    */
    int pointer_lookup(void *vp, std::string &name, collection_entry *&ep);

    /// The pointer to the collection stored in the constructor
    collection *cop;
  
    /// Output all of the remaining pointers to 'out'
    int pointer_map_fout(out_file_format *out);
    
    /** 
	\brief Keep track of the number of pointers added to ptr_map
    
	These are counted for the purposes of making a unique name. 
	This is initialized
	in fout() and incremented in io_base::pointer_out
    */
    int npointers;
    
#endif

  };

  /** 
      
  \brief A template for adding I/O classes


  */
  template <class object> class io_vtlate : public io_base {
    
  public:
    
    /** \name Functions to be overloaded 
	
	These functions should be overloaded in all descendants of
	io_tlate.
    */
    //@{
    
    /** \brief The name of the type to be processed */
    virtual const char *type() { return "io_tlate"; }
    
    /** \brief Method for reading an object from \c ins  */
    virtual int input(cinput *cin, in_file_format *ins, object *op) 
      { return 0; }
    
    /** 
	\brief Method for writing an object to \c outs  */
    virtual int output(coutput *cout, out_file_format *outs, object *op) 
      { return 0; }
    //@}
    
    /** \name Functions to be overloaded for static data
	
	These functions should be overloaded in all descendants of
	io_tlate which control I/O for classes which contain static
	data.
    */
    //@{
    
    /** \brief \c true  if the object contains static I/O data */
    virtual bool has_static_data() { return false; }
    
    /** \brief Method for reading static data for an object 
	from \c ins 
	
	One must be careful about objects which set the static data in
	their constructors. An object is automatically created in order
	to read its static data. This means that if the static data is
	set in the constructor, then possibly useful information will be
	overwritten through the creation of this temporary object. 
	
	If one needs to set static data in the constructor of a
	singleton object, then the create() and remove() functions
	should be empty and a separate pointer to the singleton should
	be provided instead of void *vp.
	
	This is only used if has_static_data() returns true;
    */
    virtual int stat_input(cinput *cin, in_file_format *ins, object *op) 
      { return 0; }
    
    /** \brief Method for writing static data for an object 
	to \c outs  
    */
    virtual int stat_output(coutput *cout, out_file_format *outs, 
			    object *op) 
      { return 0; }
    //@}
  };
  
  /** \brief A template for adding I/O classes (documents template
      io_tlate) 



      Note that the generic interface here only works with
      pointers, not with the actual objects themselves. This
      is important, because it avoids the problem of I/O
      for an object with private copy and assigment operators.
      For basic types (bool, char, double, int, etc.),
      some additional add() and get() functions are defined.
      
  */
  template <class object> class io_tlate : public io_base {
      
  public:
      
    /// Create an I/O class for type \c object.
    io_tlate() : io_base() {};

    /** 
	\brief Create an I/O class for type \c object only if another
	object of type \c t is not yet present
    */
    io_tlate(const char *t) : io_base(t) {};

    /** \name Functions to be overloaded 

	These functions should be overloaded in all descendants of
	io_tlate.
    */
    //@{
    
    /** \brief The name of the type to be processed */
    virtual const char *type() { return "io_tlate"; }

    /** \brief Method for reading an object from \c ins  */
    virtual int input(cinput *cin, in_file_format *ins, object *op) 
      { return 0; }

    /** 
	\brief Method for writing an object to \c outs  */
    virtual int output(coutput *cout, out_file_format *outs, object *op) 
      { return 0; }
    //@}

    /** \name Functions to be overloaded for static data
	
	These functions should be overloaded in all descendants of
	io_tlate which control I/O for classes which contain static
	data.
    */
    //@{

    /** \brief \c true  if the object contains static I/O data */
    virtual bool has_static_data() { return false; }

    /** \brief Method for reading static data for an object 
	from \c ins 

	One must be careful about objects which set the static data in
	their constructors. An object is automatically created in order
	to read its static data. This means that if the static data is
	set in the constructor, then possibly useful information will be
	overwritten through the creation of this temporary object. 
      
	If one needs to set static data in the constructor of a
	singleton object, then the create() and remove() functions
	should be empty and a separate pointer to the singleton should
	be provided instead of void *vp.

	This is only used if has_static_data() returns true;
    */
    virtual int stat_input(cinput *cin, in_file_format *ins, object *op) 
      { return 0; }
    
    /** \brief Method for writing static data for an object 
	to \c outs  
    */
    virtual int stat_output(coutput *cout, out_file_format *outs, 
			    object *op) 
      { return 0; }
    //@}

#ifndef DOXYGEN_INTERNAL

  protected:  
    
    virtual int object_in_void(cinput *cin, in_file_format *ins, void *op,
			       std::string &name) {
      return object_in(cin,ins,(object *)op,name);
    }
    
    virtual int object_in_void(cinput *cin, in_file_format *ins, void *op, 
			       int sz, std::string &name) {
      return object_in(cin,ins,(object *)op,sz,name);
    }
    
    virtual int object_in_void(cinput *cin, in_file_format *ins, void *op, 
			       int sz, int sz2, std::string &name) {
      return object_in(cin,ins,(object **)op,sz,sz2,name);
    }

    virtual int object_in_mem_void(cinput *cin, in_file_format *ins, 
				   void *&vp, std::string &name) {
      object *op;
      int ret=object_in_mem(cin,ins,op,name);
      vp=(void *)op;
      return ret;
    }
    
    virtual int object_in_mem_void(cinput *cin, in_file_format *ins, void *&vp,
				   int &sz,std::string &name) {
      object *op;
      int ret=object_in_mem(cin,ins,op,sz,name);
      vp=(void *)op;
      return ret;
    }
    
    virtual int object_in_mem_void(cinput *cin, in_file_format *ins, void *&vp,
				   int &sz, int &sz2, std::string &name) {
      object **op;
      int ret=object_in_mem(cin,ins,op,sz,sz2,name);
      vp=(void *)op;
      return ret;
    }

    virtual int object_out_void(coutput *cout, out_file_format *outs, 
				void *op, int sz=0, int sz2=0, 
				std::string name="") {
      if (sz2==0) {
	return object_out(cout,outs,(object *)op,sz,name);
      } 
      return object_out(cout,outs,(object **)op,sz,sz2,name);
    }

    virtual int stat_in_noobj(cinput *cin, in_file_format *ins) { 
      int ret;
      object *op=new object;
      ret=stat_input(cin,ins,op);
      return ret; 
    };
  
    virtual int stat_out_noobj(coutput *cout, out_file_format *outs) {
      int ret;
      object *op=new object;
      ret=stat_output(cout,outs,op);
      return ret; 
    };

    int in_wrapper(cinput *cin, in_file_format *ins, void *&vp) {
      object *op;
      if (vp==NULL) {
	op=new object;
      } else {
	op=(object *)vp;
      }
      input(cin,ins,op);
      if (vp==NULL) vp=(void *)op;
      return 0;
    }
  
    int in_wrapper(cinput *cin, in_file_format *ins, void *&vp, int &sz) {
      object *op;
      ins->int_in(sz);
      if (vp==NULL) {
	op=new object[sz];
      } else {
	op=(object *)vp;
      }
      for(int i=0;i<sz;i++) {
	input(cin,ins,&(op[i]));
      }
      if (vp==NULL) vp=(void *)op;
      return 0;
    }
  
    int in_wrapper(cinput *cin, in_file_format *ins, void *&vp, int &sz, 
		   int &sz2) {
      object **op;
      ins->int_in(sz);
      ins->int_in(sz2);
      if (vp==NULL) {
	op=new object *[sz];
      } else {
	op=(object **)vp;
      }
      for(int i=0;i<sz;i++) {
	if (vp==NULL) op[i]=new object[sz2];
	for(int j=0;j<sz2;j++) {
	  input(cin,ins,&(op[i][j]));
	}
      }
      if (vp==NULL) vp=(void *)op;
      return 0;
    }
  
    int out_wrapper(coutput *cout, out_file_format *outs, void *vp, int sz, 
		    int sz2) {
      int i, j;
      std::string stype;
      object *op=(object *)vp;

      if (sz==0 && sz2==0) {
	output(cout,outs,op);
      } else if (sz2==0) {
	outs->int_out(sz);
	for(i=0;i<sz;i++) {
	  output(cout,outs,&(op[i]));
	}
      } else {
	object **aop=(object **)vp;
	outs->int_out(sz);
	outs->int_out(sz2);
	for(i=0;i<sz;i++) {
	  for(j=0;j<sz2;j++) {
	    output(cout,outs,&(aop[i][j]));
	  }
	}
      }
      return 0;
    }

    /// Remove the memory for an object
    virtual int remove(void *vp) {
      object *op=(object *)vp;
      delete op;
      return 0;
    }  
      
    /// Remove the memory for an array of objects
    virtual int remove_arr(void *vp) {
      object *op=(object *)vp;
      delete[] op;
      return 0;
    }  
    
    /// Remove the memory for a 2-dimensional array of objects
    virtual int remove_2darr(void *vp, int sz) {
      object **op=(object **)vp;
      for(int i=0;i<sz;i++) {
	delete[] op[i];
      }
      delete[] op;
      return 0;
    }  

    /// Static input for an object
    virtual int stat_in_wrapper(cinput *cin, in_file_format *ins, void *vp) {
      object *op=(object *)vp;
      stat_input(cin,ins,op);
      return 0;
    }
  
    /// Static output for an object
    virtual int stat_out_wrapper(coutput *cout, out_file_format *outs, 
				 void *vp) {
      object *op=(object *)vp;
      outs->start_object(type());
      stat_output(cout,outs,op);
      outs->end_object();
      return 0;
    }

#endif

  public:

    /** \name Input functions
     */
    //@{

    /** \brief Read an object from \c ins  */
    virtual int object_in(cinput *cin, in_file_format *ins, object *op,
			  std::string &name) {
      int i, j;
      
      std::string typ;
      ins->start_object(typ,name);
      input(cin,ins,op);
      ins->end_object();
      return 0;
    }
    
    /** \brief Read an array of objects from \c ins */
    virtual int object_in(cinput *cin, in_file_format *ins, object *op, 
			  int sz, std::string &name) {
      int i;
    
      std::string typ;
      ins->start_object(typ,name);
      ins->int_in(sz);
      for(i=0;i<sz;i++) {
	input(cin,ins,&op[i]);
      }
      ins->end_object();
      return 0;
    }
    
    /** \brief Read a 2-d array of objects from \c ins */
    virtual int object_in(cinput *cin, in_file_format *ins, object **op, 
			  int sz, int sz2, std::string &name) {
      int i, j;
      
      std::string typ;
      ins->start_object(typ,name);
      ins->int_in(sz);
      ins->int_in(sz2);
	
      for(i=0;i<sz;i++) {
	object *opx=op[i];
	for(j=0;j<sz2;j++) {
	  input(cin,ins,&opx[j]);
	}
      }
      ins->end_object();
      return 0;
    }
    
    /** \brief Create memory for a 2-d array of objects 
	and read it from \c ins  
    */
    template<size_t N>
      int object_in(cinput *co, in_file_format *ins, 
		    object op[][N], int sz, std::string &name) {
      int i, j;
	
      std::string typ;
      ins->start_object(typ,name);
      ins->int_in(sz);
      ins->int_in(((int)N));
	
      for(i=0;i<sz;i++) {
	for(j=0;j<N;j++) {
	  input(co,ins,&op[i][j]);
	}
      }
      ins->end_object();

      return 0;
    }

    /** \brief Create memory for an object and read it from \c ins  */
    virtual int object_in_mem(cinput *cin, in_file_format *ins, 
			      object *&op, std::string &name) {
      int i, j;
      op=new object;
      std::string typ;
      ins->start_object(typ,name);
      input(cin,ins,op);
      ins->end_object();
      return 0;
    }
    
    /** \brief Create memory for an object and read it from \c ins  */
    virtual int object_in_mem(cinput *cin, in_file_format *ins, object *&op, 
			      int &sz, std::string &name) {
      int i, j;
      
      std::string typ;
      ins->start_object(typ,name);
      ins->int_in(sz);
      op=new object[sz];
      for(i=0;i<sz;i++) {
	input(cin,ins,&op[i]);
      }
      ins->end_object();
      return 0;
    }
    
    /** \brief Create memory for an object and read it from \c ins  */
    virtual int object_in_mem(cinput *cin, in_file_format *ins, 
			      object **&op, int &sz, int &sz2,
			      std::string &name) {
      int i, j;
      
      std::string typ;
      ins->start_object(typ,name);
      ins->int_in(sz);
      ins->int_in(sz2);
      
      op=new object *[sz];
      for(i=0;i<sz;i++) {
	op[i]=new object[sz2];
	for(j=0;j<sz2;j++) {
	  input(cin,ins,&op[i][j]);
	}
      }
      ins->end_object();
      
      return 0;
    }
    
    /** \brief Create memory for a 2-d array of objects 
	and read it from \c ins  
	
	Note that you must specify in advance the size \c N.
    */
    template<size_t N>
      int object_in_mem(cinput *co, in_file_format *ins, 
			object op[][N], int &sz, std::string &name) {
      int i, j;
      
      std::string typ;
      ins->start_object(typ,name);
      ins->int_in(sz);
      ins->int_in(((int)N));
      
      // commented out on 3/10/05 - i don't think this should be here
      //	op=new object *[sz];
      for(i=0;i<sz;i++) {
	// commented out on 3/10/05 - i don't think this should be here
	//	  op[i]=new object[N];
	for(j=0;j<N;j++) {
	  input(co,ins,&op[i][j]);
	}
      }
      ins->end_object();
      
      return 0;
    }
    
    //@}
    
    /** \name Output functions
     */
    //@{
    /** \brief Output an object (or an array of objects) to \c outs 
     */
    virtual int object_out(coutput *cout, out_file_format *outs, object *op, 
			   int sz=0, std::string name="") {
      if (sz==0) {
	outs->start_object(type(),name);
	output(cout,outs,op);
      } else {
	outs->start_object(((std::string)(type()))+"[]",name);
	outs->int_out(sz);
	for(int i=0;i<sz;i++) {
	  output(cout,outs,&op[i]);
	}
      }
      outs->end_object();
      return 0;
    }
      
    /** \brief Output an object (or an array of objects) to \c outs 
     */
    virtual int object_out(coutput *cout, out_file_format *outs, object **op, 
			   int sz, int sz2, std::string name="") {
      int i, j;
      
      outs->start_object(((std::string)(type()))+"[][]",name);
      outs->int_out(sz);
      outs->int_out(sz2);
      for(i=0;i<sz;i++) {
	for(j=0;j<sz2;j++) {
	  output(cout,outs,&(op[i][j]));
	}
      }
      outs->end_object();
      return 0;
    }
      
    /** \brief Output a 2-d array of objects to \c outs 
     */
    template<size_t N>
      int object_out(coutput *cout, out_file_format *outs, 
		     object op[][N], int sz, std::string name="") {
      int i, j;
      
      outs->start_object(((std::string)(type()))+"[][]",name);
      outs->int_out(sz);
      outs->int_out(N);
      for(i=0;i<sz;i++) {
	for(j=0;j<N;j++) {
	  output(cout,outs,&op[i][j]);
	}
      }
      outs->end_object();
      return 0;
    }
    //@}
    
    /** \name Memory allocation
     */
    //@{
    /** \brief Create memory for an object */
    virtual int mem_alloc(object *&op) {
      op=new object;
      return 0;
    }

    /** \brief Create memory for an object */
    virtual int mem_alloc_arr(object *&op, int sz) {
      op=new object[sz];
      return 0;
    }

    /** \brief Create memory for an object */
    virtual int mem_alloc_2darr(object **&op, int sz, int sz2) {
      op=new object *[sz];
      for(int i=0;i<sz;i++) {
	op[i]=new object[sz2];
      }
      return 0;
    }
    //@}
    
    /** \name Add and get objects from a collection
     */
    //@{

    /** \brief Add an object(s) to a collection */
    int add(collection &coll, std::string name, object *op, int sz=0, 
	    bool overwrt=true, bool owner=false) {
      return coll.add(name,this,op,sz,0,overwrt,owner);
    }

    /** \brief Add an object(s) to a collection */
    int add_2darray(collection &coll, std::string name, object **op, int sz, 
		    int sz2, bool overwrt=true, bool owner=false) {
      return coll.add(name,this,op,sz,sz2,overwrt,owner);
    }

    /** \brief Get an object(s) from a collection */
    int get(collection &coll, std::string tname, object *&op) {
      void *vp;
      int ret=coll.get(tname,vp);
      if (ret==0) {
	op=(object *)vp;
      } else {
	op=NULL;
	return ret;
      }
      return 0;
    }
    
    /** \brief Get an object(s) from a collection */
    int get(collection &co, std::string tname, object *&op, int &sz) {
      void *vp;
      int ret=co.get(tname,vp,sz);
      if (ret==0) {
	op=(object *)vp;
      } else {
	op=NULL;
	return ret;
      }
      return 0;
    }
    
    /** \brief Get an object(s) from a collection */
    int get(collection &co, std::string tname, object **&op, int &sz,
	    int &sz2) {
      void *vp;
      int ret=co.get(tname,vp,sz,sz2);
      if (ret==0) {
	op=(object **)vp;
      } else {
	op=NULL;
	return ret; 
      }
      return 0;
    }
    //@}

    /** \name Other functions
     */
    //@{

    /// Free the memory associated with an object 
    virtual int mem_free(object *op) { return remove((void *)op); }  
    
    /// Free the memory associated with an array of objects
    virtual int mem_free_arr(object *op) { return remove_arr((void *)op); }  
    
    /// Free the memory associated with a 2-d array of objects
    virtual int mem_free_2darr(object **op, int sz) 
      { return remove_2darr((void *)op,sz); }  
    
    //@}

  }; 


  template<> int io_tlate<bool>::input
    (cinput *co, in_file_format *ins, bool *dp);
  template<> int io_tlate<bool>::output
    (coutput *co, out_file_format *outs, bool *dp);
  template<> const char *io_tlate<bool>::type();
  
  /** 
      \brief I/O object for bool variables


  */
#ifdef DOXYGENP
  class bool_io_type : public io_tlate<object>
#else
    class bool_io_type : public io_tlate<bool>
#endif
    {
    public:

      /// Desc
      bool_io_type(const char *t) : io_tlate<bool>(t) {};

      bool_io_type() : io_tlate<bool>() {};

      /// Add a bool to a collection
      int addb(collection &co, std::string name, bool x, bool overwrt=true);
      /// Get a bool from a collection
      bool getb(collection &co, std::string tname);
      /// Get a bool from a collection
      int get_def(collection &co, std::string tname, bool &op, bool def=false);
    };

  template<> int io_tlate<char>::input
    (cinput *co, in_file_format *ins, char *dp);
  template<> int io_tlate<char>::output
    (coutput *co, out_file_format *outs, char *dp);
  template<> const char *io_tlate<char>::type();
  
  /** 
      \brief I/O object for char variables


  */
#ifdef DOXYGENP
  class char_io_type : public io_tlate<object>
#else
    class char_io_type : public io_tlate<char> 
#endif
    {
    public:
      /// Desc
      char_io_type(const char *t) : io_tlate<char>(t) {};

      char_io_type() : io_tlate<char>() {};

      /// Add a char to a collection
      int addc(collection &co, std::string name, char x, bool overwrt=true);
      /** 
	  \brief Get a char from a collection
	  
	  Some older systems have trouble with functions named \c getc,
	  so this is named \c getcc instead.
       */
      char getcc(collection &co, std::string tname);
      /// Get a char from a collection
      int get_def(collection &co, std::string tname, char &op, char def='x');
    };

  template<> int io_tlate<double>::input
    (cinput *co, in_file_format *ins, double *dp);
  template<> int io_tlate<double>::output
    (coutput *co, out_file_format *outs, double *dp);
  template<> const char *io_tlate<double>::type();
  
  /** 
      \brief I/O object for double variables


  */
#ifdef DOXYGENP
  class double_io_type : public io_tlate<object>
#else
    class double_io_type : public io_tlate<double>
#endif
    {
    public:
      /// Desc
      double_io_type(const char *t) : io_tlate<double>(t) {};

      double_io_type() : io_tlate<double>() {};

      /// Add a double to a collection
      int addd(collection &co, std::string name, double x, bool overwrt=true);
      /// Get a double from a collection
      double getd(collection &co, std::string tname);
      /// Get a double from a collection
      int get_def(collection &co, std::string tname, double &op, 
		  double def=0.0);
    };
  
  template<> int io_tlate<int>::input
    (cinput *co, in_file_format *ins, int *dp);
  template<> int io_tlate<int>::output
    (coutput *co, out_file_format *outs, int *dp);
  template<> const char *io_tlate<int>::type();
  
  /** 
      \brief I/O object for int variables


  */
#ifdef DOXYGENP
  class int_io_type : public io_tlate<object> 
#else
    class int_io_type : public io_tlate<int>
#endif
    {
    public:
      /// Desc
      int_io_type(const char *t) : io_tlate<int>(t) {};

      int_io_type() : io_tlate<int>() {};

      /// Add a int to a collection
      int addi(collection &co, std::string name, int x, bool overwrt=true);
      /// Get a int from a collection
      int geti(collection &co, std::string tname);
      /// Get a int from a collection
      int get_def(collection &co, std::string tname, int &op, 
		  int def=0);
    };

  template<> int io_tlate<unsigned long int>::input
    (cinput *co, in_file_format *ins, unsigned long int *dp);
  template<> int io_tlate<unsigned long int>::output
    (coutput *co, out_file_format *outs, unsigned long int *dp);
  template<> const char *io_tlate<unsigned long int>::type();
  
  /** 
      \brief I/O object for long variables


  */
#ifdef DOXYGENP
  class long_io_type : public io_tlate<object> 
#else
    class long_io_type : public io_tlate<unsigned long int>
#endif
    {
    public:
      /// Desc
      long_io_type(const char *t) : io_tlate<unsigned long int>(t) {};

      long_io_type() : io_tlate<unsigned long int>() {};

      /// Add a long to a collection
      int addl(collection &co, std::string name, unsigned long int x, 
	       bool overwrt=true);
      /// Get a long from a collection
      int getl(collection &co, std::string tname);
      /// Get a long from a collection
      int get_def(collection &co, std::string tname, unsigned long int &op, 
		  unsigned long int def=0);
    };
  
  template<> int io_tlate<std::string>::input
    (cinput *co, in_file_format *ins, std::string *dp);
  template<> int io_tlate<std::string>::output
    (coutput *co, out_file_format *outs, std::string *dp);
  template<> const char *io_tlate<std::string>::type();
  
  /** 
      \brief I/O object for string variables


  */
#ifdef DOXYGENP
  class string_io_type : public io_tlate<object> 
#else
    class string_io_type : public io_tlate<std::string>
#endif
    {
    public:
      /// Desc
      string_io_type(const char *t) : io_tlate<std::string>(t) {};

      string_io_type() : io_tlate<std::string>() {};

      /// Add a string to a collection
      int adds(collection &co, std::string name, std::string s, 
	       bool overwrt=true);
      /// Get a string from a collection
      std::string gets(collection &co, std::string tname);
      /// Get a string from a collection
      int get_def(collection &co, std::string tname, std::string &op, 
		  std::string def="");
    };
  
  /** 
      \brief I/O object for words


  */
#ifdef DOXYGENP
  class word_io_type : public io_tlate<object>
#else
    class word_io_type : public io_tlate<std::string>
#endif
    {
    public:
      /// Desc
      word_io_type(const char *t) : io_tlate<std::string>(t) {};

      word_io_type() : io_tlate<std::string>() {};

      /// Desc
      int input(cinput *co, in_file_format *ins, std::string *dp);
      /// Desc
      int output(coutput *co, out_file_format *outs, std::string *dp);
      /// Add a string to a collection
      int addw(collection &co, std::string name, std::string w, 
	       bool overwrt=true);
      /// Get a word from a collection
      std::string getw(collection &co, std::string tname);
      /// Get a word from a collection
      int get_def(collection &co, std::string tname, std::string &op, 
		  std::string def="");
      /// Desc
      const char *type() { return "word"; }
    };

#ifdef NEVER_DEFINED
  /** 
      \brief I/O for ovector


  */
  template<> int io_tlate<ovector>::input
    (cinput *co, in_file_format *ins, ovector *dp);
  template<> int io_tlate<ovector>::output
    (coutput *co, out_file_format *outs, ovector *dp);
  template<> const char *io_tlate<ovector>::type();

  /** 
      \brief I/O for omatrix


  */
  template<> int io_tlate<omatrix>::input
    (cinput *co, in_file_format *ins, omatrix *dp);
  template<> int io_tlate<omatrix>::output
    (coutput *co, out_file_format *outs, omatrix *dp);
  template<> const char *io_tlate<omatrix>::type();

  /** 
      \brief I/O for ovector_int


  */
  template<> int io_tlate<ovector_int>::input
    (cinput *co, in_file_format *ins, ovector_int *dp);
  template<> int io_tlate<ovector_int>::output
    (coutput *co, out_file_format *outs, ovector_int *dp);
  template<> const char *io_tlate<ovector_int>::type();

  /** 
      \brief I/O for omatrix_int


  */
  template<> int io_tlate<omatrix_int>::input
    (cinput *co, in_file_format *ins, omatrix_int *dp);
  template<> int io_tlate<omatrix_int>::output
    (coutput *co, out_file_format *outs, omatrix_int *dp);
  template<> const char *io_tlate<omatrix_int>::type();

  /** 
      \brief I/O for ovector_cx


  */
  template<> int io_tlate<ovector_cx>::input
    (cinput *co, in_file_format *ins, ovector_cx *dp);
  template<> int io_tlate<ovector_cx>::output
    (coutput *co, out_file_format *outs, ovector_cx *dp);
  template<> const char *io_tlate<ovector_cx>::type();

  /** 
      \brief I/O for omatrix_cx


  */
  template<> int io_tlate<omatrix_cx>::input
    (cinput *co, in_file_format *ins, omatrix_cx *dp);
  template<> int io_tlate<omatrix_cx>::output
    (coutput *co, out_file_format *outs, omatrix_cx *dp);
  template<> const char *io_tlate<omatrix_cx>::type();

  typedef io_tlate<ovector> ovector_io_type;
  typedef io_tlate<omatrix> omatrix_io_type;
  typedef io_tlate<ovector_int> ovector_int_io_type;
  typedef io_tlate<omatrix_int> omatrix_int_io_type;
  typedef io_tlate<ovector_cx> ovector_cx_io_type;
  typedef io_tlate<omatrix_cx> omatrix_cx_io_type;

#endif

#ifndef DOXYGENP  
}
#endif

#endif
