/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_STRING_CONV_H
#define O2SCL_STRING_CONV_H
/** \file string_conv.h
    \brief Various string conversion functions
*/

#include <iostream>
#include <cmath>
#include <string>
#include <fstream>
#include <sstream>
#include <o2scl/lib_settings.h>
#include <gsl/gsl_ieee_utils.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Convert a pointer to a string 

      This uses an \c ostringstream to convert a pointer to a string
      and is architecture-dependent.
  */
  std::string ptos(void *p);

  /** \brief Convert an integer to a string 
   */
  std::string itos(int x);

  /** \brief Convert a double to a string 
   */
  std::string dtos(double x, int prec=6, bool auto_prec=false);

  /** \brief Returns the number of characters required to display the 
      exponent of \c x in scientific mode

      This usually returns 2 or 3, depending on whether or not the 
      absolute magnitude of the exponent is greater than 100.
   */
  size_t size_of_exponent(double x);

  /** 
      \brief Convert a double to a string using a specified format
      
      \todo Add error checking to this function using \c if(strout << x)
  */
  std::string dtos(double x, std::ostream &format);

  /** 
      \brief Convert a string to an integer 
      
      If this function fails it will call set_err() and return zero.
  */
  int stoi(std::string s);

  /** 
      \brief Convert a string to a double 

      If this function fails it will call set_err() and return zero.
  */
  double stod(std::string s);

  /** 
      \brief Convert a double to a Latex-like string

      The parameter \c sigfigs is the number of significant figures
      to give in the string. The parameters \c ex_min and \c ex_max
      represent the base-10 logarithm of the smallest and largest 
      numbers to be represented without the string 
      \verbatim
      $\times 10^{\mathrm{ex}}$
      \endverbatim
      where \c ex is the relevant exponent.  The number zero is always
      converted to "0".

      Note that this function does not warn the user if the number
      of significant figures requested is larger than the machine 
      precision.

      \todo Fix to ensure final zeros are printed properly
      if requested
  */
  std::string double_to_latex(double x, int sigfigs=5, int ex_min=-2, 
			      int ex_max=3);

  /** 
      \brief Convert a double to a HTML-like string

      This uses \c \&times; and \c <sup></sup> to convert a double
      to a string for use on the web. 

      The parameter \c sigfigs is the number of significant figures to
      give in the string. The parameters \c ex_min and \c ex_max
      represent the base-10 logarithm of the smallest and largest
      numbers to be represented without exponential notation. The
      number zero is always converted to "0". Superscripts are
      implemented using (can't use greater than size in documentation)

      Note that this function does not warn the user if the number
      of significant figures requested is larger than the machine 
      precision.
  */
  std::string double_to_html(double x, int sigfigs=5, int ex_min=-2, 
			     int ex_max=3);
  
  /** 
      \brief Convert a double to a string containing IEEE representation
      
      Modeled after the GSL function \c gsl_ieee_fprintf_double(), but
      converts to a \c string instead of a \c FILE *.
  */
  std::string double_to_ieee_string(double *x);

#ifndef DOXYGENP
}
#endif

#endif

