/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/table.h>

using namespace std;
using namespace o2scl;

table::table(int cmaxlines) {
  nlines=0;
  intp_set=false;
  maxlines=cmaxlines;

  intp1=&cintp1;
  intp2=&cintp2;
}

table::~table() {
  aiter it;

  for(it=atree.begin();it!=atree.end();it++) {
    if (it->second.owner==true) {
      ovector *op=(ovector *)(it->second.dat);
      op->free();
    }
  }
  atree.clear();
  if (intp_set==true) {
    delete si;
  }
}

int table::summary(ostream *out, int ncol) const {

  if (constants.size()==1) {
    (*out) << "1 constant:" << endl;
  } else {
    (*out) << constants.size() << " constants:" << endl;
  }
  std::map<std::string,double>::const_iterator mit;
  for(mit=constants.begin();mit!=constants.end();mit++) {
    (*out) << mit->first << " " << mit->second << endl;
  }

  // Output number of columns and preprend column numbers
  int nh=get_ncolumns(), nh2;
  if (nh==1) {
    (*out) << "1 column: " << endl;
  } else {
    (*out) << nh << " columns: " << endl;
  }
  string *h=new string[nh], *h2;
  for(int i=0;i<nh;i++) {
    h[i]=itos(i)+". "+get_column_name(i);
  }
  
  // Convert to string with width 'ncol'
  screenify(h,nh,h2,nh2,ncol);
  
  // Output column names
  for(int i=0;i<nh2;i++) {
    (*out) << h2[i] << endl;
  }
  
  if (get_nlines()==0) (*out) << "No lines of data." << endl;
  else if (get_nlines()==1) (*out) << "One line of data." << endl;
  (*out) << get_nlines() << " lines of data." << endl;
  
  delete[] h;
  delete[] h2;
  
  return 0;
}

int table::check_synchro() const {
  if (atree.size()!=alist.size()) {
    set_err("Size of table and list do not match.",gsl_esanity);
    return gsl_esanity;
  }
  for(aciter it=atree.begin();it!=atree.end();it++) {
    if (it->second.index!=alist[it->second.index]->second.index) {
      set_err((((string)"Problem with iterator for entry '")+
	       it->first+"' in list.").c_str(),gsl_esanity);
    }
  }
  for(int i=0;i<((int)atree.size());i++) {
    if (alist[i]->second.index!=i) {
      set_err((((string)"Problem with index of entry ")+
	       itos(i)+" in list.").c_str(),gsl_esanity);
      return gsl_esanity;
    }
  }
  return gsl_success;
}

int table::reset_list() {
  aiter it;
  for(it=atree.begin();it!=atree.end();it++) {
    alist[it->second.index]=it;
  }
  return gsl_success;
}

const gsl_vector *table::get_gsl_vector(std::string name) const {
  aciter it=atree.find(name);
  if (it==atree.end()) {
    set_err((((string)"Column '")+name+
	     " not found in make_gsl_vector(). "+
	     "Returning NULL pointer.").c_str(),
    	    gsl_notfound);
    return NULL;
  }
  return (const gsl_vector *)(it->second.dat);
}

int table::ch_owner(std::string name, bool ow) {
  aiter it=atree.find(name);
  if (it==atree.end()) {
    set_err((((string)"Column '")+name+" not found in ch_owner().").c_str(),
    	    gsl_notfound);
    return gsl_notfound;
  }
  it->second.owner=ow;
  return gsl_success;
}

bool table::get_owner(std::string name) const {
  aciter it=atree.find(name);
  if (it==atree.end()) {
    set_err((((string)"Column '")+name+" not found in get_owner().").c_str(),
    	    gsl_notfound);
    return false;
  }
  return it->second.owner;
}

table *table::subtable(string list, size_t top, size_t bottom,
		       bool linked) {
  int sublines, i;
  table *at;
  string head;
  aiter it;
  
  if (top>bottom) {
    size_t tmp=bottom;
    bottom=top;
    top=tmp;
  }
  sublines=bottom-top+1;
  if (nlines==0) {
    set_err("Can't make a subtable of an empty table. Returning NULL.",
    	    gsl_einval);
    return NULL;
  }
  if (bottom+1>nlines) {
    set_err("Requested row beyond nlines. Adjusting.",gsl_einval);
    bottom=nlines-1;
  }

  istringstream *is;
  is=new istringstream(list);

  if (linked==true) {
    at=new table(sublines);
    while((*is) >> head) {
      it=atree.find(head);
      if (it==atree.end()) {
	set_err
	  ((((string)"Couldn't find column named ")+head+
	    " in table::subtable(). Returning NULL pointer.").c_str(),
	   gsl_einval);
	delete at;
	return NULL;
      }
      ovector_subvector *as=new ovector_subvector(*it->second.dat,
						  top,((size_t)sublines));
      at->new_column(it->first,as);
      at->ch_owner(it->first,true);
    }
    if (at->get_ncolumns()==0) {
      set_err("Making a subtable with no columns.",gsl_einval);
    }
  } else {
    at=new table(sublines);
    while((*is) >> head) {
      it=atree.find(head);
      if (it==atree.end()) {
	set_err
	  ((((string)"Couldn't find column named ")+head+
	    " in table::subtable(). Returning NULL pointer.").c_str(),
	   gsl_einval);
	delete at;
	return NULL;
      }
      at->new_column(head);
      ovector_view *dcol=at->get_column(head);
      for(i=0;i<sublines;i++) {
	(*dcol)[i]=(*it->second.dat)[i+top];
      }
    } 
    if (at->get_ncolumns()==0) {
      set_err("Making a subtable with no columns.",gsl_einval);
    }
  }
  at->nlines=sublines;

  delete is;

  return at;
}

double *table::create_array(std::string scol) const {
  aciter its=atree.find(scol), itd;
  if (its==atree.end()) {
    set_err((((string)"Column '")+scol+" not found in copy_column().").c_str(),
    	    gsl_notfound);
    return NULL;
  }
  double *x=new double[nlines];
  for(size_t i=0;i<nlines;i++) {
    x[i]=(*its->second.dat)[i];
  }
  return x;
}

int table::copy_column(std::string src, std::string dest) {
  aiter its=atree.find(src), itd;
  if (its==atree.end()) {
    set_err((((string)"Column '")+src+" not found in copy_column().").c_str(),
    	    gsl_notfound);
    return gsl_notfound;
  }
  col s;
  s.dat=new ovector(maxlines);
  s.owner=true;
  s.index=0;
  atree.insert(make_pair(dest,s));
  itd=atree.find(dest);
  for(int i=0;i<((int)nlines);i++) {
    (*itd->second.dat)[i]=(*its->second.dat)[i];
  }
  return gsl_success;
}

int table::init_column(std::string scol, double val) {
  if (!finite(val)) {
    set_err("Value not finite in init_column()",gsl_einval);
    return gsl_einval;
  }
  aiter it=atree.find(scol);
  if (it==atree.end()) {
    set_err((((string)"Column '")+scol+"' not found in init_column()").c_str(),
    	    gsl_notfound);
    return gsl_notfound;
  }
  for(size_t i=0;i<nlines;i++) {
    (*(it->second.dat))[i]=val;
  }
  if (intp_set && (scol==intp_colx || scol==intp_coly)) {
    intp_set=false;
    delete si;
  }
  return gsl_success;
}

int table::sortd_comp(const void *a, const void *b) {
  sortd *sa=(sortd *)a, *sb=(sortd *)b;
  if (sa->val<sb->val) return -1;
  else if (sa->val>sb->val) return 1;
  return 0;
}

int table::sort_table(string scol) {
  int i, j;
  ovector **newdata;
  aiter it=atree.find(scol), itj;
  if (it==atree.end()) {
    set_err((((string)"Column '")+scol+" not found in sort_table().").c_str(),
    	    gsl_notfound);
    return gsl_notfound;
  }
  
  sortd *sa=new sortd[nlines];
  for(i=0;i<((int)nlines);i++) {
    sa[i].val=(*it->second.dat)[i];
    sa[i].indx=i;
  }
  qsort(sa,nlines,sizeof(sortd),sortd_comp);
  
  newdata=new ovector *[atree.size()];
  for(i=0;i<((int)atree.size());i++) {
    newdata[i]=new ovector(maxlines);
  }
  
  for(j=0,itj=atree.begin();itj!=atree.end();itj++,j++) {
    for(i=0;i<((int)nlines);i++) {
      (*newdata[j])[i]=(*itj->second.dat)[sa[i].indx];
    }
    ovector *op=(ovector *)(itj->second.dat);
    op->free();
    itj->second.dat=newdata[j];
  }
  delete[] newdata;
  delete[] sa;
  
  if (intp_set) {
    intp_set=false;
    delete si;
  }

  return gsl_success;
}

int table::sort_column(string scol) {
  int i;
  aiter it=atree.find(scol);
  if (it==atree.end()) {
    set_err((((string)"Column '")+scol+" not found in sort_column().").c_str(),
    	    gsl_notfound);
    return gsl_notfound;
  }

  sortd *sa=new sortd[nlines];
  for(i=0;i<((int)nlines);i++) {
    sa[i].val=(*it->second.dat)[i];
    sa[i].indx=i;
  }
  qsort(sa,nlines,sizeof(sortd),sortd_comp);

  for(i=0;i<((int)nlines);i++) {
    (*it->second.dat)[i]=sa[i].val;
  }

  delete[] sa;

  if (intp_set && (scol==intp_colx || scol==intp_coly)) {
    intp_set=false;
    delete si;
  }

  return gsl_success;
}

int table::rename_column(string olds, string news) {
  // We cannot just rename the column since it is in a map,
  // so we have to remove it and re-insert it, while
  // retaining the original ownership.
  bool ow=get_owner(olds);
  ovector_view *dat=get_column(olds);

  delete_column(olds);
  new_column(news,dat);
  ch_owner(news,ow);
  return gsl_success;
}

table::aiter table::get_iterator(std::string lname) {
  aiter it=atree.find(lname);
  if (it==atree.end()) {
    set_err((((string)"Column '")+lname+
	     " not found in get_iterator().").c_str(),
    	    gsl_notfound);
  }
  return it;
}

table::col *table::get_col_struct(std::string lname) {
  aiter it=atree.find(lname);
  if (it==atree.end()) {
    set_err((((string)"Column '")+lname+
	     " not found in get_col_struct().").c_str(),
    	    gsl_notfound);
    return NULL;
  }
  return &(it->second);
}

int table::delete_column(string scol) {

  // Find the tree iterator for the element we want to erase
  aiter it=atree.find(scol);
  if (it==atree.end()) {
    set_err((((string)"Column '")+scol+
	     " not found in delete_column().").c_str(),
    	    gsl_notfound);
    return gsl_notfound;
  }

  // Find the corresponding list iterator
  aviter vit=alist.begin();
  vit+=it->second.index;
  
  // Find the last element in the list and it's corresponding table
  // entry. Change it's index to the index of the column to be
  // deleted.
  alist[alist.size()-1]->second.index=it->second.index;

  // Erase the elements from the list and the tree
  atree.erase(it);
  alist.erase(vit);

  // Reset the list to reflect the proper iterators
  reset_list();

  return gsl_success;
}

int table::delete_row(string scol, double val) {
  int irow=lookup(scol,val);
  if (irow==gsl_notfound) return gsl_notfound;

  return delete_row(irow);
}

int table::delete_row(size_t irow) {
  int i;
  aiter it;
  for(it=atree.begin();it!=atree.end();it++) {
    for(i=irow;i<((int)nlines-1);i++) {
      (*it->second.dat)[i]=(*it->second.dat)[i+1];
    }
  }
  nlines--;
  if (intp_set==true) {
    delete si;
    intp_set=false;
  }
  return gsl_success;
}

size_t table::ordered_lookup(string scol, double val) {
  int ret;
  if (!finite(val)) {
    set_err("Value not finite in ordered_lookup()",gsl_einval);
    return gsl_einval;
  }
  aiter it=atree.find(scol);
  if (it==atree.end()) {
    set_err((((string)"Column '")+scol+
	     " not found in ordered_lookup().").c_str(),
    	    gsl_notfound);
    return gsl_notfound;
  }
  
  ret=se.ordered_lookup(val,nlines,*it->second.dat);
  return ret;
}

size_t table::mlookup(string scol, double val, vector<double> &results,
		      double threshold) const {
  size_t i;
  if (!finite(val)) {
    set_err("Value not finite in lookup()",gsl_einval);
    return gsl_einval;
  }
  aciter it=atree.find(scol);
  if (it==atree.end()) {
    set_err((((string)"Column '")+scol+" not found in lookup().").c_str(),
    	    gsl_notfound);
    return gsl_notfound;
  }
  if (threshold==0.0) {
    for(i=0;i<nlines;i++) {
      if ((*it->second.dat)[i]==val) {
	results.push_back(i);
      }
    }
  } else {
    for(i=0;i<nlines;i++) {
      if (fabs((*it->second.dat)[i]-val)<threshold) {
	results.push_back(i);
      }
    }
  }
  return results.size();
}

size_t table::lookup(string scol, double val) const {
  int i, indx=0;
  if (!finite(val)) {
    set_err("Value not finite in lookup()",gsl_einval);
    return gsl_einval;
  }
  aciter it=atree.find(scol);
  if (it==atree.end()) {
    set_err((((string)"Column '")+scol+" not found in lookup().").c_str(),
	    gsl_notfound);
    return gsl_notfound;
  }
  return it->second.dat->lookup(val);
}

double table::lookup_val(string scol, double val, string scol2) const {
  int i, indx=0;
  if (!finite(val)) {
    set_err("Value not finite in lookup()",gsl_einval);
    return gsl_einval;
  }
  aciter it=atree.find(scol);
  if (it==atree.end()) {
    set_err((((string)"Column '")+scol+" not found in lookup().").c_str(),
	    gsl_notfound);
    return gsl_notfound;
  }
  return get(scol2,it->second.dat->lookup(val));
}

size_t table::lookup(int icol, double val) const {
  return lookup(get_column_name(icol),val);
}

int table::set_nlines(size_t il) {

  // Double the number of lines until we have enough
  bool error=false;
  if (il>maxlines) {
    int ret=inc_maxlines(il-maxlines);
    if (ret!=0) error=true;
  }

  // If the increase failed, then stop
  if (error==true) {
    add_err_ret("Failed to increase maxlines in set_nlines()",gsl_efailed);
  }

  // Now that maxlines is large enough, set the number of lines 
  nlines=il;
  
  // Reset the interpolation object for future interpolations
  if (intp_set) {
    intp_set=false;
    delete si;
  }

  return gsl_success;
}

int table::set_nlines_auto(size_t il) {
  
  // Double the number of lines until we have enough
  bool error=false;
  while(il>maxlines && error==false) {
    int ret;
    if (maxlines==0) ret=inc_maxlines(5);
    else ret=inc_maxlines(maxlines);
    if (ret!=0) error=true;
  }

  // If the increase failed, then stop
  if (error==true) {
    add_err_ret("Failed to increase maxlines in set_nlines_auto()",
		gsl_efailed);
  }

  // Now that maxlines is large enough, set the number of lines 
  nlines=il;
  
  // Reset the interpolation object for future interpolations
  if (intp_set) {
    intp_set=false;
    delete si;
  }

  return gsl_success;
}

double table::deriv(size_t ix, double x0, size_t iy) {
  return deriv(get_column_name(ix),x0,get_column_name(iy));
}

double table::deriv(string sx, double x0, string sy) {
  double ret;
  aiter itx=atree.find(sx), ity=atree.find(sy);
  if (itx==atree.end() || ity==atree.end()) {
    set_err((((string)"Columns '")+sx+"' or '"+sy+
	     "' not found in deriv().").c_str(),
    	    gsl_notfound);
    return 0.0;
  }
  if (!finite(x0)) {
    set_err("x0 not finite in deriv().",gsl_einval);
    return gsl_einval;
  }
  if (intp_set==false || sx!=intp_colx || sy!=intp_coly) {
    if (intp_set==true) {
      delete si;
    } else {
      intp_set=true;
    }
    si=new sm_interp_vec(*intp1,*intp2,
			 nlines,*itx->second.dat,*ity->second.dat);
    intp_colx=sx;
    intp_coly=sy;
  }
  ret=si->deriv(x0);
  return ret;
}
  
int table::deriv(string x, string y, string yp) {
  aiter itx, ity, ityp;
  int ix=0, iy=0;
  new_column(yp);

  itx=atree.find(x);
  ity=atree.find(y);
  ityp=atree.find(yp);

  if (itx==atree.end() || ity==atree.end() || ityp==atree.end()) {
    set_err("Column not found in deriv().", 
    	    gsl_notfound);
    return gsl_einval;
  }
  
  lookup_column(x,ix);
  lookup_column(y,iy);
  for(int i=0;i<((int)nlines);i++) {
    (*ityp->second.dat)[i]=deriv(ix,(*itx->second.dat)[i],iy);
  }
  
  return gsl_success;
}

int table::integ(string x, string y, string ynew) {
  aiter itx, ity, itynew;
  int ix=0, iy=0;
  new_column(ynew);

  itx=atree.find(x);
  ity=atree.find(y);
  itynew=atree.find(ynew);

  if (itx==atree.end() || ity==atree.end() || itynew==atree.end()) {
    set_err("Column not found in deriv().", 
    	    gsl_notfound);
    return gsl_einval;
  }
  
  lookup_column(x,ix);
  lookup_column(y,iy);
  for(size_t i=0;i<nlines;i++) {
    (*itynew->second.dat)[i]=integ(ix,(*itx->second.dat)[0],
				 (*itx->second.dat)[i],iy);
  }
  
  return gsl_success;
}

double table::deriv2(size_t ix, double x0, size_t iy) {
  return deriv2(get_column_name(ix),x0,get_column_name(iy));
}
 
double table::deriv2(string sx, double x0, string sy) {
  double ret;
  aiter itx=atree.find(sx), ity=atree.find(sy);
  if (itx==atree.end() || ity==atree.end()) {
    set_err((((string)"Columns '")+sx+"' or '"+sy+
	     "' not found in deriv2().").c_str(),
    	    gsl_notfound);
    return 0.0;
  }
  if (!finite(x0)) {
    set_err("x0 not finite in deriv().",gsl_einval);
    return gsl_einval;
  }
  if (intp_set==false || sx!=intp_colx || sy!=intp_coly) {
    if (intp_set==true) {
      delete si;
    } else {
      intp_set=true;
    }
    si=new sm_interp_vec(*intp1,*intp2,
			 nlines,*itx->second.dat,*ity->second.dat);
    intp_colx=sx;
    intp_coly=sy;
  }
  ret=si->deriv2(x0);
  return ret;
}
  
int table::deriv2(string x, string y, string yp) {
  aiter itx, ity, ityp;
  int ix=0, iy=0;
  new_column(yp);

  itx=atree.find(x);
  ity=atree.find(y);
  ityp=atree.find(yp);

  if (itx==atree.end() || ity==atree.end() || ityp==atree.end()) {
    set_err("Column not found in deriv2().", 
	    gsl_notfound);
    return gsl_einval;
  }
  
  lookup_column(x,ix);
  lookup_column(y,iy);
  for(int i=0;i<((int)nlines);i++) {
    (*ityp->second.dat)[i]=deriv2(ix,(*itx->second.dat)[i],iy);
  }
  
  return gsl_success;
}

double table::interp(size_t ix, double x0, size_t iy) {
  return interp(get_column_name(ix),x0,get_column_name(iy));
}

double table::interp(string sx, double x0, string sy) {
  double ret;
  aiter itx=atree.find(sx), ity=atree.find(sy);
  if (itx==atree.end() || ity==atree.end()) {
    set_err((((string)"Columns '")+sx+"' or '"+sy+
	     "' not found in interp().").c_str(),
    	    gsl_notfound);
    return 0.0;
  }
  if (!finite(x0)) {
    set_err("x0 not finite in interp().",gsl_einval);
    return gsl_einval;
  }
  if (intp_set==false || sx!=intp_colx || sy!=intp_coly) {
    if (intp_set==true) {
      delete si;
    } else {
      intp_set=true;
    }
    si=new sm_interp_vec(*intp1,*intp2,
			 nlines,*itx->second.dat,*ity->second.dat);
    intp_colx=sx;
    intp_coly=sy;
  }
  ret=si->interp(x0);
  return ret;
}

double table::interp_const(string sx, double x0, string sy) const {
  double ret;
  aciter itx=atree.find(sx), ity=atree.find(sy);
  if (itx==atree.end() || ity==atree.end()) {
    set_err((((string)"Columns '")+sx+"' or '"+sy+
	     "' not found in interp().").c_str(),
    	    gsl_notfound);
    return 0.0;
  }
  if (!finite(x0)) {
    set_err("x0 not finite in interp().",gsl_einval);
    return gsl_einval;
  }
  sm_interp_vec *sic=new sm_interp_vec(*intp1,*intp2,nlines,
				      *itx->second.dat,*ity->second.dat);
  ret=sic->interp(x0);
  delete sic;
  return ret;
}

double table::integ(size_t ix, double x1, double x2, size_t iy) {
  return integ(get_column_name(ix),x1,x2,
	       get_column_name(iy));
}

double table::integ(string sx, double x1, double x2, string sy) {
  double ret;
  aiter itx=atree.find(sx), ity=atree.find(sy);
  if (itx==atree.end() || ity==atree.end()) {
    set_err((((string)"Columns '")+sx+"' or '"+sy+
	     "' not found in integ().").c_str(),
    	    gsl_notfound);
    return 0.0;
  }
  if (!finite(x1) || !finite(x2)) {
    set_err("x1 or x2 not finite in integ().",gsl_einval);
    return gsl_einval;
  }
  if (intp_set==false || sx!=intp_colx || sy!=intp_coly) {
    if (intp_set==true) {
      delete si;
    } else {
      intp_set=true;
    }
    si=new sm_interp_vec(*intp1,*intp2,
			 nlines,*itx->second.dat,*ity->second.dat);
    intp_colx=sx;
    intp_coly=sy;
  }
  ret=si->integ(x1,x2);
  return ret;
}

void table::zero_table() {
  aiter it;
  for(it=atree.begin();it!=atree.end();it++) {
    for(int j=0;j<((int)nlines);j++) {
      (*it->second.dat)[j]=0.0;
    }
  }

  if (intp_set) {
    intp_set=false;
    delete si;
  }

  return;
}

void table::clear_table() {
  aiter it;
  for(it=atree.begin();it!=atree.end();it++) {
    if (it->second.owner==true) {
      ovector *op=(ovector *)(it->second.dat);
      op->free();
    }
  }
  atree.clear();
  alist.clear();
  nlines=0;
  if (intp_set==true) {
    delete si;
    intp_set=false;
  }

  return;
}

int table::inc_maxlines(size_t llines) {
  int j;
  ovector_view *newcol, *oldcol;
  aiter it;

  // Check to ensure that all of the columns not owned by the table
  // have enough space for the increase
  if (llines>0) {
    for(it=atree.begin();it!=atree.end();it++) {
      if (it->second.owner==false && 
	  it->second.dat->size()<(maxlines+llines)) {
	std::string s=((string)"column '")+it->first+
	  "' not owned by the table ran out of space.";
	set_err(s.c_str(),gsl_einval);
	return gsl_einval;
      }
    }
  } else {
    // Quietly do nothing
    return 0;
  }

  // Add space to all of the columns that are owned by the table
  // by copying the data to new arrays
  for(it=atree.begin();it!=atree.end();it++) {
    if (it->second.owner==true) {
      newcol=new ovector(maxlines+llines);
      
      // Copy data to new array
      for(j=0;j<((int)maxlines);j++) {
	(*newcol)[j]=(*it->second.dat)[j];
      }
      
      // Swap arrays and delete the old one
      oldcol=it->second.dat;
      it->second.dat=newcol;
      delete oldcol;
    }
  }
  
  maxlines+=llines;

  return 0;
}

ovector_view *table::get_column(string scol) {
  ovector_view *tmp;
  aciter it=atree.find(scol);
  if (it==atree.end()) {
    set_err((((string)"Column '")+scol+
	     "' not found in get_column(). Returning NULL pointer.").c_str(),
    	    gsl_notfound);
    tmp=NULL;
  } else {
    tmp=it->second.dat;
  }
  return tmp;
}

const ovector_view *table::get_column_const(string scol) const {
  ovector_view *tmp;
  aciter it=atree.find(scol);
  if (it==atree.end()) {
    set_err((((string)"Column '")+scol+
	     "' not found in get_column(). Returning NULL pointer.").c_str(),
	    gsl_notfound);
    tmp=NULL;
  } else {
    tmp=it->second.dat;
  }
  return tmp;
}

int table::get_row(string scol, double val, ovector &row) const {
  
  int irow=lookup(scol,val);
  if (irow==gsl_notfound) {
    set_err((((string)"Column '")+scol+
	     "' not found in get_row(). Returning NULL pointer.").c_str(),
    	    gsl_notfound);
    return gsl_notfound;
  } 
  return get_row(irow,row);
}

int table::get_row(size_t irow, ovector &row) const {

  if (irow+1>nlines) {
    set_err((((string)"Row '")+ itos(irow)+
	     "' not found in get_row(). Returning NULL pointer.").c_str(),
    	    gsl_notfound);
    return gsl_notfound;
  }
  
  int i;
  aciter it;
  row.allocate(atree.size());
  for(i=0,it=atree.begin();it!=atree.end();it++,i++) {
    row[i]=(*it->second.dat)[irow];
  }
  return 0;
}

double table::get(string scol, size_t row) const {
  double tmp;
  aciter it=atree.find(scol);
  if (it==atree.end()) {
    set_err((((string)"Column '")+scol+
	     "' not found in get().").c_str(),
    	    gsl_notfound);
    tmp=0.0;
  } else {
    tmp=(*it->second.dat)[row];
  }
  return tmp;
}

int table::set(string scol, size_t row, double val) {
  if (!finite(val)) {
    set_err("value not finite in set2(string,int,double).",gsl_einval);
    return gsl_einval;
  }
  
  if (maxlines==0) inc_maxlines(row+1);
  while(row>maxlines-1) inc_maxlines(maxlines);
  if (row>=nlines) set_nlines_auto(row+1);

  aiter it=atree.find(scol);
  if (it==atree.end()) {
    new_column(scol);
    it=atree.find(scol);
  }
  if (it==atree.end()) {
    set_err((((string)"Refused to add column '")+scol+
	     "' in set2(string,int,double).").c_str(),
    	    gsl_notfound);
    return gsl_notfound;
  }

  (*it->second.dat)[row]=val;
  return gsl_success;
}

int table::set(size_t icol, size_t row, double val) {
  if (!finite(val)) {
    set_err("value not finite in set(int,int,double).",gsl_einval);
    return gsl_einval;
  }
  if (icol>=atree.size()) {
    set_err("Invalid column in set(int,int,double)",gsl_einval);
    return gsl_einval;
  }
  
  while(row>maxlines-1) inc_maxlines(maxlines);
  if (row>=nlines) set_nlines_auto(row+1);
  
  ovector_view *colp=alist[icol]->second.dat;
  (*colp)[row]=val;
  return gsl_success;
}

double table::get(size_t icol, size_t row) {
  if (icol>=atree.size()) {
    set_err("Invalid column in get(size_t,size_t)",gsl_einval);
    return 0.0;
  }
  if (row>=nlines) {
    set_err("Invalid row in get(size_t,size_t)",gsl_einval);
    return 0.0;
  }
  ovector_view *colp=alist[icol]->second.dat;
  return (*colp)[row];
}
	     
int table::new_row(size_t n) {
  if (nlines>=maxlines) inc_maxlines(maxlines);
  
  nlines++;
  for(int i=((int)nlines)-2;i>=((int)n);i--) {
    copy_row(i,i+1);
  }

  if (intp_set) {
    intp_set=false;
    delete si;
  }

  return 0;
}

int table::copy_row(size_t src, size_t dest) {
  for(int i=0;i<((int)atree.size());i++) {
    set(i,dest,get(i,src));
  }
  return 0;
}

int table::insert_data(size_t n, size_t nv, double *v) {

  if (maxlines==0) inc_maxlines(5);
  if (nlines>=maxlines) inc_maxlines(maxlines);

  if (intp_set) {
    intp_set=false;
    delete si;
  }
  
  if (new_row(n)==0 && nv<=atree.size()) {
    
    for(size_t i=0;i<nv;i++) {
      (*this)[i][n]=v[i];
    }
    return 0;
  }
  return gsl_einval;
}

int table::insert_data(size_t n, size_t nv, double **v) {
  if (maxlines==0) inc_maxlines(5);
  if (nlines>=maxlines) inc_maxlines(maxlines);

  if (intp_set) {
    intp_set=false;
    delete si;
  }

  if (new_row(n)==0 && nv<=(atree.size())) {
    
    for(size_t i=0;i<nv;i++) {
      (*this)[i][n]=*(v[i]);
    }
    return 0;
  }
  return gsl_einval;
}

int table::line_of_names(string newheads) {
  int ret=0;
  string head;

  istringstream *is=new istringstream(newheads);
  
  while((*is) >> head) {
    error_update(ret,new_column(head));
  } 
  delete is;
  
  if (ret!=0) {
    add_err_ret("At least one new column failed in line_of_names().",
		gsl_efailed);
  }

  return gsl_success;
}

string table::get_column_name(size_t icol) const {
  if (icol+1>atree.size()) {
    return "";
  }
  return alist[icol]->first;
}

string table::get_sorted_name(size_t icol) {
  if (icol+1>atree.size()) {
    return "";
  }
  aiter it=atree.begin();
  for(size_t i=0;i<icol;i++) it++;
  return it->first;
}

int table::lookup_column(string lname, int &index) {
  aiter it=atree.find(lname);
  if (it==atree.end()) {
    // We choose not to call the error handler here in case the 
    // user wants to use this function to test to see if a 
    // column is already present in the table
    index=0;
    return gsl_notfound;
  }
  index=it->second.index;
  return 0;
}

int table::new_column(string head) {
  int tmp;
  if (lookup_column(head,tmp)!=gsl_notfound) {
    set_err_ret((((string)"Column '")+head+
		 "' already present in new_column().").c_str(),
		gsl_einval);
  }
  for(int i=0;i<((int)head.size());i++) {
    if (head[i]==' ' || head[i]=='\t' || head[i]=='\n' || head[i]=='\r' 
	|| head[i]=='\v' || head[i]=='\f') {
      set_err_ret("Invalid column name.",gsl_einval);
    }
  }
  col s;
  s.dat=new ovector(maxlines);
  s.owner=true;
  s.index=atree.size();
  atree.insert(make_pair(head,s));
  aiter it=atree.find(head);
  alist.push_back(it);
  return 0;
}

int table::new_column(string head, ovector_view *ldat) {
  int tmp;
  // Make sure the column name doesn't already appear
  if (lookup_column(head,tmp)!=gsl_notfound) {
    set_err_ret("Column name already present in table.",gsl_einval);
  }
  // Make sure the column name is valid
  for(int i=0;i<((int)head.size());i++) {
    if (head[i]==' ' || head[i]=='\t' || head[i]=='\n' || head[i]=='\r' 
	|| head[i]=='\v' || head[i]=='\f') {
      set_err_ret("Invalid column name.",gsl_einval);
    }
  }
  // Ensure that the column doesn't have more data 
  if (ldat->size()>maxlines) {
    set_err_ret("New column has more data than the table.",gsl_efailed);
  }
  // Add the column to the table
  col s;
  s.dat=ldat;
  s.owner=false;
  s.index=atree.size();
  aiter ait=atree.insert(make_pair(head,s)).first;
  alist.push_back(ait);
  // What to do if the column is smaller than the table??
  return 0;
}

double table::max(string scol) const {
  double ret=0.0;
  int i;
  const ovector_view *dcol=get_column_const(scol);
  if (dcol!=NULL) {
    bool setb=false;
    for(i=0;i<((int)nlines);i++) {
      if (finite((*dcol)[i])) {
	if (setb==false) {
	  ret=(*dcol)[i];
	  setb=true;
	} else if ((*dcol)[i]>ret) {
	  ret=(*dcol)[i];
	}
      }
    }
    if (setb==false) {
      set_err("No finite values found in table::max().",gsl_efailed);
      return 0.0;
    }
    return ret;
  }
  set_err((((string)"Column '")+scol+"' not found in max().").c_str(),
  	  gsl_notfound);
  return 0.0;
}

double table::min(string scol) const {
  double ret=0.0;
  int i;
  const ovector_view *dcol=get_column_const(scol);
  if (dcol!=NULL) {
    bool setb=false;
    for(i=0;i<((int)nlines);i++) {
      if (finite((*dcol)[i])) {
	if (setb==false) {
	  ret=(*dcol)[i];
	  setb=true;
	} else if ((*dcol)[i]<ret) {
	  ret=(*dcol)[i];
	}
      }
    }
    if (setb==false) {
      set_err("No finite values found in table::min().",gsl_efailed);
      return 0.0;
    }
    return ret;
  }
  set_err((((string)"Column '")+scol+"' not found in min().").c_str(),
  	  gsl_notfound);
  return 0.0;
}

int table::add_col_from_table(std::string loc_index, table &source,
			      std::string src_index, std::string src_col,
			      std::string dest_col) {

  if (dest_col=="") dest_col=src_col;

  // Add the new column
  new_column(dest_col);
  if (err_hnd->get_errno()!=0) {
    add_err_ret("Failed to create new column in add_col_from_table().",
		err_hnd->get_errno());
  }
  
  // Fill the new column
  for(size_t i=0;i<nlines;i++) {
    set(dest_col,i,source.interp(src_index,get(loc_index,i),src_col));
  }
  
  return 0;
}

/// Input specialization
template<> int io_tlate<table>::input
(cinput *co, o2scl::in_file_format *ins, table *ta) {
  int nheads, ncons, i, j, itmp;
  std::string name;
  double val;
    
  ins->int_in(nheads,"n_cols");
  ins->int_in(itmp,"n_lines");
  
  ta->nlines=itmp;
  ta->maxlines=ta->nlines;
  
  ins->int_in(ncons);
  
  for(i=0;i<ncons;i++) {
    ins->word_in(name);
    ins->double_in(val);
    ta->add_constant(name,val);
  }
  
  for(i=0;i<nheads;i++) {
    ins->word_in(name);
    if (ta->new_column(name)!=0) {
      // If the column addition failed (usually because of 
      // a repeated name), then give up mercilessly.
      return 1;
    }
  }
    
  for(j=0;j<((int)ta->nlines);j++) {
    for(i=0;i<nheads;i++) {
      ins->double_in((*ta)[i][j]);
    }
  }
    
  return 0;
}

/// Output specialization
template<> int io_tlate<table>::output
(coutput *co, o2scl::out_file_format *outs, 
 table *at) {
  int i,j;
  outs->int_out(at->atree.size(),"n_cols");
  outs->int_out(at->nlines,"n_lines");
  outs->int_out(at->constants.size());
  outs->end_line();
  std::map<std::string,double>::iterator mit;
  for(mit=at->constants.begin();mit!=at->constants.end();mit++) {
    outs->word_out(mit->first);
    outs->double_out(mit->second);
    outs->end_line();
  }
  for(i=0;i<((int)(at->atree.size()));i++) {
    outs->word_out(at->get_column_name(i));
  }
  outs->end_line();
  for(j=0;j<((int)(at->nlines));j++) {
    for(i=0;i<((int)at->atree.size());i++) {
      outs->double_out((*at)[i][j]);
    }
    outs->end_line();
  }
  return 0;
}

/// \fn Type specialization
template<> const char *io_tlate<table>::type() { return "table"; }
