/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_NAMBUJL_EOS_H
#define O2SCL_NAMBUJL_EOS_H

#include <o2scl/part.h>
#include <cmath>
#include <o2scl/constants.h>
#include <o2scl/quark_eos.h>
#include <o2scl/mm_funct.h>
#include <o2scl/gsl_mroot_hybrids.h>
#include <o2scl/inte.h>
#include <o2scl/gsl_inte_qag.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Nambu Jona-Lasinio EOS at zero temperature

      Calculates everything from the quark condensates ([uds].qq) and the
      chemical potentials ([uds].mu). If "fromqq" is set to false, then
      instead it calculates everything from the dynamical masses
      ([uds].ms) and the chemical potentials. L, G, K, and B0 are fixed
      constants.  [uds].pr returns the pressure due to the Fermi-gas
      contribution plus the bag pressure contribution. [uds.ed] is the
      energy density for each quark so that e.g. u.ed+u.pr=u.mu*u.n. B0
      should be fixed using calc_B0() beforehand to ensure that the energy
      density and pressure of the vacuum is zero.

      The functions set_parameters() should be called first.

      The code is based on \ref Buballa99.

      The Lagrangian is
      \f[
      {\cal L} = \bar q ( i \partial{\hskip-2.0mm}/ - {\hat m_0}) q \;+\;
      G \sum_{k=0}^8 [\,({\bar q}\lambda_k q)^2 + ({\bar q}
      i\gamma_5\lambda_k q)^2\,] + {\cal L}_6
      \f]
      \f[
      {\cal L}_6 = - K \,[ \,{\rm det}_f ({\bar
      q}(1+\gamma_5) q) + {\rm det}_f ({\bar q}(1-\gamma_5) q) \,] \, .
      \f]

      And the corresponding thermodynamic potential is
      \f[
      \Omega = \Omega_{FG} + \Omega_{Int}
      \f]
      where \f$\Omega_{FG}\f$ is the Fermi gas contribution and
      \f[
      \frac{\Omega_{\mathrm{Int}}}{V} = - 2 N_c \sum_{i=u,d,s}
      \int \frac {d^3p}{(2\pi)^3} \sqrt{m_i^2 + p^2} +
      \frac{\Omega_{V}}{V}
      \f]
      \f[
      \frac{\Omega_{V}}{V} = 
      \sum_{i=u,d,s} 2 G \langle\bar{q}_i q_i \rangle^2
      - 4 K \langle \bar{q}_u q_u \rangle \langle \bar{q}_d q_d \rangle
      \langle \bar{q}_s q_s \rangle + B_0\,.
      \f]
      where \f$B_0\f$ is a constant defined to ensure that the 
      energy density and the pressure of the vacuum is zero.

      Unlike \ref Buballa99, the bag constant, \f$\Omega_{Int}/V\f$ is
      defined without the term
      \f[
      \sum_{i=u,d,s} 2 N_C \int_0^{\Lambda} 
      \frac{d^3 p}{(2 \pi)^3} \sqrt{ m_{0,i}^2+p^2 } ~dp
      \f]
      since this allows an easier comparison to the finite temperature
      EOS. 
      The constant \f$B_0\f$ in this case 
      is therefore significantly 
      larger, but the energy density and pressure are still zero in
      the vacuum. 

      The Feynman-Hellman theorem (\ref Bernard88 ), gives
      \f[
      \left< \bar{q} q \right> = \frac{\partial m^{*}}{\partial m}
      \f]

      The functions calc_e() and calc_p() never return a value other
      than zero, but will give nonsensical results for nonsensical
      inputs.

      <b>Finite T documentation</b>

      Calculates everything from the quark condensates ([uds].qq) and the
      chemical potentials ([uds].mu). If "fromqq" is set to false, then
      instead it calculates everything from the dynamical masses
      ([uds].ms) and the chemical potentials. L, G, K, and B0 are fixed
      constants.  [uds].pr returns the pressure due to the Fermi-gas
      contribution plus the bag pressure contribution. [uds.ed] is the
      energy density for each quark so that e.g. u.ed+u.pr=u.mu*u.n. B0 is
      fixed to ensure that the energy density and pressure of the vacuum
      is zero.  

      This implementation includes contributions from antiquarks.
      
      References: \ref Buballa99, \ref Hatsuda94.
  */

  class nambujl_eos : public quark_eos {
  public:

    /** \brief Set the parameters and the bag constant \c B0 
      
	This function allows the user to specify the momentum cutoff,
	\c lambda, the four-fermion coupling \c fourferm 
	and the six-fermion coupling from the 't Hooft interaction
	\c sixferm . If 0.0 is given for any of the values, then
	the default is used (\f$ \Lambda=602.3/(\hbar c),
	G=1.835/\Lambda^2, K=12.36/\Lambda^5 \f$).

	The value of the shift in the bag constant \c B0 is
	automatically calculated to ensure that the energy density and
	the pressure of the vacuum are zero. The functions
	set_quarks() and set_thermo() can be used before hand to
	specify the \ref quark and \ref thermo objects.
    */
    virtual int set_parameters(double lambda=0.0, double fourferm=0.0, 
			       double sixferm=0.0);
    
    /** 
	\brief Accuracy limit for Fermi integrals for finite temperature

	\ref limit is used for the finite temperature integrals to
	ensure that no numbers larger than exp(limit) or smaller than
	exp(-limit) are avoided. (Default: 20)
    */
    double limit;
  
    /** 
	\brief Calculate from quark condensates if true (default true)
    
	If this is false, then computations are performed using
	the effective masses as inputs
     */
    bool fromqq;

    nambujl_eos();

    /** 
	\brief Equation of state as a function of chemical potentials
	
	This function automatically solves the gap equations
    */
    virtual int calc_p(quark &u, quark &d, quark &s, thermo &lth);

    /** \brief Equation of state as a function of chemical potentials at 
	finite temperature

	This function automatically solves the gap equations
    */
    virtual int calc_temp_p(quark &u, quark &d, quark &s, 
			    const double T, thermo &th);

    /** \brief Equation of state and gap equations as a function of 
	chemical potential
    */
    virtual int calc_eq_p(quark &u, quark &d, quark &s, double &gap1, 
			  double &gap2, double &gap3, thermo &lth);
    
    /** \brief Equation of state and gap equations as a function of
	the densities
    */
    virtual int calc_eq_e(quark &u, quark &d, quark &s, double &gap1, 
			  double &gap2, double &gap3, thermo &lth);
    
    /** \brief Equation of state and gap equations
	as a function of chemical potentials
    */
    int calc_eq_temp_p(quark &tu, quark &td, quark &ts,
		       double &gap1, double &gap2, double &gap3,
		       thermo &qb, const double temper);
    
    /** \brief Calculates gap equations in \c y as a function of the 
	constituent masses in \c x

	The function utilizes the \ref quark objects which can
	be specified in set_quarks() and the \ref thermo object
	which can be specified in eos::set_thermo().
    */
    int gapfunms(size_t nv, const ovector_view &x, ovector_view &y, void *&pa);
    
    /** \brief Calculates gap equations in \c y as a function of the 
	quark condensates in \c x

	The function utilizes the \ref quark objects which can
	be specified in set_quarks() and the \ref thermo object
	which can be specified in eos::set_thermo().
    */
    int gapfunqq(size_t nv, const ovector_view &x, ovector_view &y, void *&pa);

    /** \brief Calculates gap equations in \c y as a function of the 
	constituent masses in \c x

	The function utilizes the \ref quark objects which can
	be specified in set_quarks() and the \ref thermo object
	which can be specified in eos::set_thermo().
    */
    int gapfunmsT(size_t nv, const ovector_view &x, ovector_view &y, 
		  void *&pa);
    
    /** \brief Calculates gap equations in \c y as a function of the 
	quark condensates in \c x

	The function utilizes the \ref quark objects which can
	be specified in set_quarks() and the \ref thermo object
	which can be specified in eos::set_thermo().
    */
    int gapfunqqT(size_t nv, const ovector_view &x, ovector_view &y, 
		  void *&pa);

    /** 
	\name The default quark masses

	These are the values from \ref Buballa99 which were used
	to fix the pion and kaon decay constants, and the pion,
	kaon, and eta prime masses. They are set in the constructor
	and are in units of \f$ \mathrm{fm}^{-1} \f$ .
    */
    //@{
    double up_default_mass;
    double down_default_mass;
    double strange_default_mass;
    //@}

    /** 
	\brief Set the quark objects to use

	The quark objects are used in gapfunms(), gapfunqq(), 
	gapfunmsT(), gapfunqqT(), and B0fun().
    */
    int set_quarks(quark &u, quark &d, quark &s);

    /// The momentum cutoff
    double L;
    
    /// The four-fermion coupling
    double G;
    
    /// The 't Hooft six-fermion interaction coupling
    double K;
    
    /// The bag constant
    double B0;
    
    /** 
	\name The default quark objects

	The masses are automatically set in the constructor to
	\c up_default_mass, \c down_default_mass, and 
	\c strange_default_mass.c
     */
    //@{
    quark def_up; 
    quark def_down;
    quark def_strange;
    //@}

    /// Return string denoting type ("nambujl_eos")
    virtual const char *type() { return "nambujl_eos"; }

    /// Set solver to use in set_parameters()
    virtual int set_solver(mroot<void *,mm_funct<void *> > &s) {
      solver=&s;
      return 0;
    }

    /// Set integration object
    virtual int set_inte(inte<void *,funct<void *> > &i) {
      it=&i;
      return 0;
    }

    /// The default solver
    gsl_mroot_hybrids<void *,mm_funct<void *> > def_solver;
    
  protected:

#ifndef DOXYGEN_INTERNAL

    friend class io_tlate<nambujl_eos>;
    
    /// The default integrator
    gsl_inte_qag<void *,funct<void *> > def_it;
    
    /// The integrator for finite temperature integrals
    inte<void *,funct<void *> > *it;
    
    /// The solver to use for set_parameters()
    mroot<void *,mm_funct<void *> > *solver;
    
    /// Used by calc_B0() to compute the bag constant
    int B0fun(size_t nv, const ovector_view &x, ovector_view &y, void *&pa);
    
    /// Calculates the contribution to the bag constant from quark \c q 
    void njbag(quark &q);

    /// The up quark
    quark *up;
    /// The down quark
    quark *down;
    /// The strange quark
    quark *strange;

    /// The integrand for the quark condensate
    double iqq(double x, void *&pa);
    /// The integrand for the density
    double ide(double x, void *&pa);
    /// The integrand for the energy density
    double ied(double x, void *&pa);
    /// The integrand for the pressure
    double ipr(double x, void *&pa);
  
    /// A structure for passing parameters to the integrands [protected]
    typedef struct njtp_s {
      double ms, m, mu, temper, limit;
    } njtp;

    /// The temperature for calc_temp_p()
    double cp_temp;
    
#endif
    
  };

  template<> int io_tlate<nambujl_eos>::input
    (cinput *co, in_file_format *ins, nambujl_eos *ne);
  template<> int io_tlate<nambujl_eos>::output
    (coutput *co, out_file_format *outs, nambujl_eos *nj);
  template<> const char *io_tlate<nambujl_eos>::type();
  
  typedef io_tlate<nambujl_eos> nambujl_eos_io_type;
  
#ifndef DOXYGENP
}
#endif

#endif
