/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/skyrme_eos.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

skyrme_eos::skyrme_eos() {

  parent_method=false;

  def_nr_neutron.init(939.0/hc_mev_fm,2.0);
  def_nr_proton.init(939.0/hc_mev_fm,2.0);
  def_nr_neutron.non_interacting=false;
  def_nr_proton.non_interacting=false;
  set_n_and_p(def_nr_neutron,def_nr_proton);

  provides=provides_calc_e | provides_calc_temp_e;
}

skyrme_eos::skyrme_eos(string model) {

  load(model);
  def_nr_neutron.init(939.0/hc_mev_fm,2.0);
  def_nr_proton.init(939.0/hc_mev_fm,2.0);
  def_nr_neutron.non_interacting=false;
  def_nr_proton.non_interacting=false;
  set_n_and_p(def_nr_neutron,def_nr_proton);
  parent_method=false;
  
  provides=provides_calc_e | provides_calc_temp_e;
}

int skyrme_eos::calc_temp_e(fermion &ne, fermion &pr, 
			    const double ltemper, thermo &locth) {
  
  double n, x, hamk, ham, ham1, ham2, ham3, ham4, ham5, ham6;
  double dhdnn, dhdnp, na, npa, nna, term, term2, common, gn, gp;
 
  ne.non_interacting=false;
  pr.non_interacting=false;

  //---------------------------------------
  // Some local variables of interest:
  //
  // hamk is just the kinetic part of the hamiltonian 
  //   hbar^2 tau / (2 m^{star})
  // ham{1-6} are remaining parts of the hamiltonian, modulo
  //   factors of density
  // dhdn{n,p} are the partial derivatives of the hamiltonian wrt the 
  //   neutron and proton densities (hold energy densities constant)

  // If the temperature is too small, just use 
  // the zero-temperature code

  if (ne.n<=0.0 && pr.n<=0.0) {
    ne.ms=ne.m;
    pr.ms=pr.m;
    ne.mu=ne.m;
    pr.mu=pr.m;
    ne.pr=0.0;
    pr.pr=0.0;
    ne.ed=0.0;
    pr.ed=0.0;
    locth.pr=0.0;
    locth.ed=0.0;
    locth.en=0.0;
    return gsl_success;
  } else if (ne.n<=0.0) {
    ne.n=0.0;
  } else if (pr.n<=0.0) {
    pr.n=0.0;
  }

  if (ltemper<=0.0) {
    calc_e(ne,pr,locth);
    return 0;
  }

  n=ne.n+pr.n;
  x=pr.n/n;

  // Landau effective masses
  term=0.25*(t1*(1.0+x1/2.0)+t2*(1.0+x2/2.0));
  term2=0.25*(t2*(0.5+x2)-t1*(0.5+x1));
  ne.ms=ne.m/(1.0+2.0*(n*term+ne.n*term2)*ne.m);
  pr.ms=pr.m/(1.0+2.0*(n*term+pr.n*term2)*pr.m);
  
  int ret1=ne.calc_density(ltemper);
  if (ret1!=0) {
    add_err("Neutrons failed in skyrme_eos::calc_temp_e(). Continuing.",
	    gsl_efailed);
  }
  int ret2;
  ret2=pr.calc_density(ltemper);
  if (ret2!=0) {
    add_err("Protons failed in skyrme_eos::calc_temp_e(). Continuing.",
	    gsl_efailed);
  }

  // Single particle potentials and energy density

  na=pow(fabs(n),alpha);
  npa=pow(fabs(pr.n),alpha);
  nna=pow(fabs(ne.n),alpha);

  hamk=ne.ed+pr.ed;
  ham1=0.5*t0*(1.0+0.5*x0);
  ham2=-0.5*t0*(0.5+x0);
  ham3=a*t3/6.0*(1.0+0.5*x3);
  ham4=a*t3*pow(2.0,alpha)/96.0*(1.0-x3);
  ham5=b*t3/12.0*(1.0+0.5*x3);
  ham6=-b*t3/12.0*(0.5+x3);
  
  ham=hamk+ham1*n*n+ham2*(ne.n*ne.n+pr.n*pr.n)+
    ham3*na*ne.n*pr.n+ham4*(nna*ne.n*ne.n+npa*pr.n*pr.n)+
    ham5*n*n*na+ham6*(ne.n*ne.n+pr.n*pr.n)*na;
  
  if (ne.inc_rest_mass) {
    gn=2.0*ne.ms*(ne.ed-ne.n*ne.m);
  } else {
    gn=2.0*ne.ms*ne.ed;
  }
  if (pr.inc_rest_mass) {
    gp=2.0*pr.ms*(pr.ed-pr.n*pr.m);
  } else {
    gp=2.0*pr.ms*pr.ed;
  }
  common=(gn+gp)*term+2.0*ham1*n+ham5*(2.0+alpha)*n*na;
  if (ne.n<=0.0) {
    dhdnn=0.0;
  } else {
    dhdnn=common+ne.nu+gn*term2+
      2.0*ham2*ne.n+ham3*na*pr.n*(alpha*ne.n/n+1.0)+
      ham4*(nna*ne.n*(2.0+alpha))+
      ham6*(2.0*ne.n*na+(ne.n*ne.n+pr.n*pr.n)*alpha*na/n);
  }
  if (pr.n<=0.0) {
    dhdnp=0.0;
  } else {
    dhdnp=common+pr.nu+gp*term2+
      2.0*ham2*pr.n+ham3*na*ne.n*(alpha*pr.n/n+1.0)+
      ham4*(npa*pr.n*(2.0+alpha))+
      ham6*(2.0*pr.n*na+(ne.n*ne.n+pr.n*pr.n)*alpha*na/n);
  }
  
  ne.mu=dhdnn;
  pr.mu=dhdnp;
  
  // Thermodynamics
  locth.ed=ham;
  locth.en=ne.en+pr.en;
  locth.pr=ltemper*locth.en+ne.mu*ne.n+pr.mu*pr.n-locth.ed;

  if (ret1!=0 || ret2!=0) return gsl_efailed;
  
  return gsl_success;
}

int skyrme_eos::calc_e(fermion &ne, fermion &pr, thermo &locth) {

  double x, n, hamk, ham, ham1, ham2, ham3, ham4, ham5, ham6;
  double dhdnn, dhdnp, na, npa, nna, term, term2, common, gn, gp;

  ne.non_interacting=false;
  pr.non_interacting=false;

  //---------------------------------------
  // Some local variables of interest:
  //
  // hamk is just the kinetic part of the hamiltonian 
  //   hbar^2 tau / (2 m^{star})
  // ham{1-6} are remaining parts of the hamiltonian, modulo
  //   factors of density
  // dhdn{n,p} are the total derivatives of the hamiltonian wrt the 
  //   neutron and proton densities (takes into account chain rule
  //   contributions from energy density)

  if (ne.n<=0.0 && pr.n<=0.0) {
    ne.ms=ne.m;
    pr.ms=pr.m;
    ne.mu=ne.m;
    pr.mu=pr.m;
    ne.pr=0.0;
    pr.pr=0.0;
    ne.ed=0.0;
    pr.ed=0.0;
    locth.pr=0.0;
    locth.ed=0.0;
    locth.en=0.0;
    return gsl_success;
  } else if (ne.n<=0.0) {
    ne.n=0.0;
  } else if (pr.n<=0.0) {
    pr.n=0.0;
  }

  n=ne.n+pr.n;
  x=pr.n/n;

  na=pow(fabs(n),alpha);
  npa=pow(fabs(pr.n),alpha);
  nna=pow(fabs(ne.n),alpha);

  term=0.25*(t1*(1.0+x1/2.0)+t2*(1.0+x2/2.0));
  term2=0.25*(t2*(0.5+x2)-t1*(0.5+x1));
  ne.ms=ne.m/(1.0+2.0*(n*term+ne.n*term2)*ne.m);
  pr.ms=pr.m/(1.0+2.0*(n*term+pr.n*term2)*pr.m);

  // We don't record error values, since these functions usually
  // always succeed
  ne.calc_density_zerot();
  pr.calc_density_zerot();
  
  hamk=ne.ed+pr.ed;

  ham1=0.5*t0*(1.0+0.5*x0);
  ham2=-0.5*t0*(0.5+x0);
  ham3=a*t3/6.0*(1.0+0.5*x3);
  ham4=a*t3*pow(2.0,alpha-2.0)/6.0*(1.0-x3);
  ham5=b*t3/12.0*(1.0+0.5*x3);
  ham6=-b*t3/12.0*(0.5+x3);

  ham=hamk+ham1*n*n+ham2*(ne.n*ne.n+pr.n*pr.n)+
    ham3*na*ne.n*pr.n+ham4*(nna*ne.n*ne.n+npa*pr.n*pr.n)+
    ham5*n*n*na+ham6*(ne.n*ne.n+pr.n*pr.n)*na;

  if (ne.inc_rest_mass) {
    gn=2.0*ne.ms*(ne.ed-ne.n*ne.m);
  } else {
    gn=2.0*ne.ms*ne.ed;
  }
  if (pr.inc_rest_mass) {
    gp=2.0*pr.ms*(pr.ed-pr.n*pr.m);
  } else {
    gp=2.0*pr.ms*pr.ed;
  }
  common=(gn+gp)*term+2.0*ham1*n+ham5*(2.0+alpha)*n*na;
  if (ne.n<=0.0) {
    dhdnn=0.0;
  } else {
    dhdnn=common+ne.nu+gn*term2+
      2.0*ham2*ne.n+ham3*na*pr.n*(alpha*ne.n/n+1.0)+
      ham4*(nna*ne.n*(2.0+alpha))+
      ham6*(2.0*ne.n*na+(ne.n*ne.n+pr.n*pr.n)*alpha*na/n);
  }
  if (pr.n<=0.0) {
    dhdnp=0.0;
  } else {
    dhdnp=common+pr.nu+gp*term2+
      2.0*ham2*pr.n+ham3*na*ne.n*(alpha*pr.n/n+1.0)+
      ham4*(npa*pr.n*(2.0+alpha))+
      ham6*(2.0*pr.n*na+(ne.n*ne.n+pr.n*pr.n)*alpha*na/n);
  }
    
  ne.mu=dhdnn;
  pr.mu=dhdnp;
  
  // Fix pr.mu since we have division by zero if pr.n=0
  if (x==0.0) pr.mu=0.0;
  
  locth.ed=ham;
  locth.pr=-locth.ed+ne.mu*ne.n+pr.mu*pr.n;
  locth.en=0.0;

  if (!finite(locth.pr)) {
    set_err("Pressure not finite in calc_e()",1);
    return 1;
  }

  return gsl_success;
}

double skyrme_eos::feoa(double nb) {
  double ret, kr23, beta, t3p;

  if (parent_method) {
    return hadronic_eos::feoa(nb);
  }
  t3p=(a+b)*t3;
  kr23=0.6/(neutron->m+proton->m)*pow(1.5*pi2*nb,2.0/3.0);
  beta=(neutron->m+proton->m)/4.0*(0.25*(3.0*t1+5.0*t2)+t2*x2);
  ret=kr23*(1.0+beta*nb)+0.375*t0*nb+0.0625*t3p*pow(nb,1.0+alpha);

  return ret;
}

double skyrme_eos::fmsom(double nb) {
  double ret, beta;

  if (parent_method) {
    return hadronic_eos::fmsom(nb);
  }
  beta=(neutron->m+proton->m)/4.0*(0.25*(3.0*t1+5.0*t2)+t2*x2);
  ret=1.0/(1.0+beta*nb);

  return ret;
}

double skyrme_eos::fcomp(double nb) {
  double ret, kr23, beta, t3p;

  if (parent_method) {
    return hadronic_eos::fcomp(nb);
  }
  t3p=(a+b)*t3;
  kr23=0.6/(neutron->m+proton->m)*pow(1.5*pi2*nb,2.0/3.0);
  beta=(neutron->m+proton->m)/4.0*(0.25*(3.0*t1+5.0*t2)+t2*x2);

  // This only works at saturation density:
  //  ret=-2.0*kr23+10.0*kr23*beta*nb+
  //    9.0/16.0*alpha*(alpha+1.0)*t3p*pow(nb,1.0+alpha);

  ret=10.0*kr23+27.0*nb*t0/4.0+40.0*kr23*beta*nb+
    9.0/16.0*alpha*(alpha+1.0)*t3p*pow(nb,1.0+alpha)+
    9.0/8.0*t3p*(1.0+alpha)*pow(nb,1.0+alpha);
  
  return ret;
}

double skyrme_eos::fesym(const double nb, const double pf) {
  double ret, kr23;

  if (pf!=0.5 || parent_method) {
    return hadronic_eos::fesym(nb,pf);
  }
  kr23=0.6/(neutron->m+proton->m)*pow(1.5*pi2*nb,2.0/3.0);
  ret=5.0/9.0*kr23+10.0/6.0*(neutron->m+proton->m)*kr23*nb*
    (t2/6.0*(1.0+1.25*x2)-0.125*t1*x1)-
    b*t3/24.0*(0.5+x3)*pow(nb,1.0+alpha)-
    0.25*t0*(0.5+x0)*nb-a/96*pow(nb,1.0+alpha)*t3*
    (2.0-alpha*(3.0+alpha)+x3*(4.0+alpha*(3.0+alpha)));
  
  return ret;
}

double skyrme_eos::fkprime(double nb) {
  double ret, kr23, t3p, beta, lmsom;

  if (parent_method) {
    return hadronic_eos::fkprime(nb);
  }
  t3p=(a+b)*t3;
  kr23=0.6/(neutron->m+proton->m)*pow(1.5*pi2*nb,2.0/3.0);
  beta=0.5*(neutron->m+proton->m)/2.0*(0.25*(3.0*t1+5.0*t2)+t2*x2);
  lmsom=1.0/(1.0+beta*nb);
  ret=2.0*kr23*(9.0-5.0/lmsom)+
    t3p*pow(nb,1.0+alpha)*27.0/16.0*alpha*(alpha*alpha-1.0);

  return ret;
}

int skyrme_eos::calpar(double gt0, double gt3, double galpha,
		       double gt1, double gt2) {

  ovector x(3);

  fixn0=n0;
  fixmsom=msom;
  fixeoa=eoa;
  fixcomp=comp;
  fixesym=esym;

  x[0]=gt0;
  x[1]=gt3;
  x[2]=galpha;

  mm_funct_mfptr<skyrme_eos,void *> fmf(this,&skyrme_eos::calparfun);
  mm_funct_mfptr<skyrme_eos,void *> fmf2(this,&skyrme_eos::calparfun2);

  void *vpx=(void *)this;
  if (eos_mroot->msolve(3,x,vpx,fmf)!=0) {
    set_err("Solution failed in calparfun().",1);
    return 1;
  }
  t0=x[0];
  t3=x[1];
  alpha=x[2];
  
  x[0]=gt1;
  x[1]=gt2;
  if (eos_mroot->msolve(2,x,vpx,fmf2)!=0) {
    set_err("Solution failed in calparfun2().",1);
    return 2;
  }
  t1=x[0];
  t2=x[1];

  return gsl_success;
}

int skyrme_eos::calparfun(size_t nv, const ovector_view &x, ovector_view &y, 
			  void *&vp) {
  skyrme_eos *sk=(skyrme_eos *)vp;
  double t3p, kr23;
  double pres, beta, mnuc=(sk->neutron->m+sk->proton->m)/2.0;

  t0=x[0];
  t3=x[1];
  alpha=x[2];

  t3p=(a+b)*t3;
  kr23=0.3/mnuc*pow(1.5*pi2*fixn0,2.0/3.0);
  beta=(1.0/fixmsom-1.0)/fixn0;
  pres=kr23*(fixn0*fixn0*beta+2.0/3.0*fixn0/fixmsom)+0.375*t0*fixn0*fixn0
    +0.0625*t3p*(1.0+alpha)*pow(fixn0,2.0+alpha);
  eoa=kr23*(1.0+beta*fixn0)+0.375*t0*fixn0+0.0625*t3p*pow(fixn0,1.0+alpha);
  comp=-2.0*kr23+10.0*kr23*beta*fixn0+
    9.0/16.0*alpha*(alpha+1.0)*t3p*pow(fixn0,1.0+alpha);

  y[0]=pres;
  y[1]=eoa-fixeoa;
  y[2]=comp-fixcomp;
    
  return 0;
}

int skyrme_eos::calparfun2(size_t nv, const ovector_view &x, ovector_view &y, 
			   void *&vp) {
  skyrme_eos *sk=(skyrme_eos *)vp;
  double kr23, beta;
  double mnuc=(sk->neutron->m+sk->proton->m)/2.0;
  
  t1=x[0];
  t2=x[1];

  kr23=0.3/mnuc*pow(1.5*pi2*fixn0,2.0/3.0);
  beta=0.5*mnuc*(0.25*(3.0*t1+5.0*t2)+t2*x2);
  msom=1.0/(1.0+beta*fixn0);
  esym=5.0/9.0*kr23+10.0/3.0*mnuc*kr23*fixn0*
    (1.0/6.0*t2*(1.0+1.25*x2)-0.125*t1*x1)-
    1.0/24.0*b*t3*(0.5+x3)*pow(fixn0,1.0+alpha)-
    0.25*t0*(0.5+x0)*fixn0-a/96*pow(fixn0,1.0+alpha)*t3*
    (2.0-alpha*(3.0+alpha)+x3*(4.0+alpha*(3.0+alpha)));
  
  y[0]=msom-fixmsom;
  y[1]=esym-fixesym;
  
  return 0;
}

int skyrme_eos::load(string model, bool external) {
  collection co;
  file_detect fd;
  //text_in_file *iff;

  double_io_type double_io;
  bool_io_type bool_io;
  string_io_type string_io;
  int_io_type int_io;

  bool dpfix;
  string fname;
  if (external) {
    fname=model;
  } else {
    fname=((string)(O2SCL_DATA_DIR))+"/skdata/"+model;
  }
    in_file_format *iff=fd.open(fname.c_str());
    //iff=new text_in_file(fname.c_str());
  
  if (co.fin(iff)!=0) {
    set_err((((string)"Couldn't find file '")+fname+
	     "' in load().").c_str(),-1);
    fd.close();
    //delete iff;
    return -1;
  }
   fd.close();
   //   delete iff;

  double_io.get_def(co,"t0hc",t0);
  double_io.get_def(co,"t2hc",t2);
  double_io.get_def(co,"t3hc",t3);
  t0/=hc_mev_fm; 
  t2/=hc_mev_fm; 
  t3/=hc_mev_fm; 
  double_io.get_def(co,"x0",x0);
  double_io.get_def(co,"x2",x2);
  double_io.get_def(co,"x3",x3);
  if (bool_io.get_def(co,"dpfix",dpfix)!=0) dpfix=false;
  if (dpfix==true) {
    t1=-t2/3.0*(5.0+4.0*x2);
    x1=-(4.0+5.0*x2)/(5.0+4.0*x2);
    alpha=1.0/3.0;
    a=1.0;
    b=0.0;
  } else {
    double_io.get_def(co,"t1hc",t1);
    t1/=hc_mev_fm;
    double_io.get_def(co,"x1",x1);
    double_io.get_def(co,"a",a,1.0);
    double_io.get_def(co,"b",b,0.0);
    double_io.get_def(co,"alpha",alpha);
  }
  double_io.get_def(co,"W0hc",W0);
  W0/=hc_mev_fm;

  return 0;
}

int skyrme_eos::check_landau(double nb, double m) {
  double f0, g0, f0p, g0p, f1, g1, f1p, g1p;
  
  landau_nuclear(nb,m,f0,g0,f0p,g0p,f1,g1,f1p,g1p);
  if (f0<-1.0) return 1;
  if (g0<-1.0) return 2;
  if (f0p<-1.0) return 3;
  if (g0p<-1.0) return 4;
  if (f1<-3.0) return 5;
  if (g1<-3.0) return 6;
  if (f1p<-3.0) return 7;
  if (g1p<-3.0) return 8;
  landau_neutron(nb,m,f0,g0,f1,g1);
  if (f0<-1.0) return 9;
  if (g0<-1.0) return 10;
  if (f1<-3.0) return 11;
  if (g1<-3.0) return 12;

  return 0;
}

int skyrme_eos::landau_nuclear(double nb, double m,
			       double &f0, double &g0, double &f0p,
			       double &g0p, double &f1, double &g1,
			       double &f1p, double &g1p) {
  
  double T0, T1, T2, T3, x, y, z, kf, mstar;
  
  kf=pow(3.0*pi2*nb/2.0,1.0/3.0);
  x=t1*x1;
  y=t2*x2;
  z=t3*x3;
  T0=0.125*(3.0*t1+5.0*t2+4.0*y);
  T1=0.125*(2.0*x+2.0*y-t1+t2);
  T2=0.125*(2.0*x-2.0*y+t1-t2);
  T3=0.125*(t1-t2);
  mstar=1.0/(1.0/m+T0*nb);

  f1=-3.0*T0*mstar*nb;
  g1=-3.0*T1*mstar*nb;
  f1p=3.0*T2*mstar*nb;
  g1p=3.0*T3*mstar*nb;
  f0=(0.75*t0+0.0625*(alpha+1.0)*(alpha+2.0)*t3*pow(nb,alpha))*
    2.0*mstar*kf/pi2-f1;
  g0=(0.25*t0*(2.0*x0-1.0)+1.0/24.0*t3*pow(nb,alpha)*(2.0*x3-1.0))*
    2.0*mstar*kf/pi2-g1;
  f0p=(-0.25*t0*(2.0*x0+1.0)-1.0/24.0*t3*pow(nb,alpha)*(2.0*x3+1.0))*
    2.0*mstar*kf/pi2-f1p;
  g0p=(-0.25*t0-1.0/24.0*t3*pow(nb,alpha))*2.0*mstar*kf/pi2-g1p;
  
  return 0;
}

int skyrme_eos::landau_neutron(double nb, double m,
			       double &f0, double &g0, 
			       double &f1, double &g1) {
  
  double T0, T1, T2, T3, x, y, z, kf, mstar;
  
  kf=pow(3.0*pi2*nb,1.0/3.0);
  x=t1*x1;
  y=t2*x2;
  z=t3*x3;
  T0=0.125*(3.0*t1+5.0*t2+4.0*y);
  T1=0.125*(2.0*x+2.0*y-t1+t2);
  T2=0.125*(2.0*x-2.0*y+t1-t2);
  T3=0.125*(t1-t2);
  mstar=1.0/(1.0/m+(T0-T2)*nb);

  f1=-3.0*(T0-T2)*mstar*nb;
  g1=-3.0*(T1-T3)*mstar*nb;
  
  f0=(0.5*t0*(1.0-x0)+1.0/24.0*(alpha+1.0)*(alpha+2.0)*t3*pow(nb,alpha)*
      (1.0-x3))*mstar*kf/pi2-f1;
  g0=(0.5*t0*(x0-1.0)+1.0/12.0*t3*pow(nb,alpha)*(x3-1.0))*mstar*kf/pi2-g1;
  
  return 0;
}

/*
  int skyrme_eos::calpar_new(double m) {
  double T0=(1.0/msom/m-1.0/m)*n0;
  double kf=pow(3.0*pi2*n0/2.0,1.0/3.0);
  alpha=(comp/9.0+eoa+(0.1/m-2.0/15.0/msom/m)*kf*kf)/
  (-eoa+(0.3/m-0.2/msom/m)*kf*kf);
  t3=16.0/pow(n0,1.0+alpha)/alpha*(-eoa+(0.3/m-0.2/msom/m)*kf*kf);
  t0=8.0/3.0/n0*(eoa-0.3/msom/m*kf*kf-0.0625*t3*pow(n0,1.0+alpha));
  
  return 0;
  }
*/

template<> int io_tlate<skyrme_eos>::input
(cinput *co, in_file_format *ins, skyrme_eos *sk) {
  std::string t;
  co->object_in("hadronic_eos",ins,sk,t);
  ins->double_in(sk->t0,"t0");
  ins->double_in(sk->t1,"t1");
  ins->double_in(sk->t2,"t2");
  ins->double_in(sk->t3,"t3");
  sk->t0/=o2scl_const::hc_mev_fm;
  sk->t1/=o2scl_const::hc_mev_fm;
  sk->t2/=o2scl_const::hc_mev_fm;
  sk->t3/=o2scl_const::hc_mev_fm;
  ins->double_in(sk->x0,"x0");
  ins->double_in(sk->x1,"x1");
  ins->double_in(sk->x2,"x2");
  ins->double_in(sk->x3,"x3");
  ins->double_in(sk->alpha,"alpha");
  ins->double_in(sk->a,"a");
  ins->double_in(sk->b,"b");
  ins->double_in(sk->W0,"W0");
  return 0;
}

template<> int io_tlate<skyrme_eos>::output
(coutput *co, out_file_format *outs, skyrme_eos *sk) {
  std::string t;
  co->object_out("hadronic_eos",outs,sk,0,0,t);
  outs->double_out(sk->t0*o2scl_const::hc_mev_fm,"t0");
  outs->double_out(sk->t1*o2scl_const::hc_mev_fm,"t1");
  outs->double_out(sk->t2*o2scl_const::hc_mev_fm,"t2");
  outs->double_out(sk->t3*o2scl_const::hc_mev_fm,"t3");
  outs->double_out(sk->x0,"x0");
  outs->double_out(sk->x1,"x1");
  outs->double_out(sk->x2,"x2");
  outs->double_out(sk->x3,"x3");
  outs->double_out(sk->alpha,"alpha");
  outs->double_out(sk->a,"a");
  outs->double_out(sk->b,"b");
  outs->double_out(sk->W0,"W0");
  return 0;
}

template<> const char *io_tlate<skyrme_eos>::type()
{return "skyrme_eos"; }
