/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_TOV_EOS_H
#define O2SCL_TOV_EOS_H

#include <cmath>
#include <iostream>
#include <fstream>

#include <o2scl/constants.h>
#include <o2scl/funct.h>
#include <o2scl/multi_funct.h>
#include <o2scl/mm_funct.h>
#include <o2scl/lib_settings.h>
#include <o2scl/collection.h>
#include <o2scl/table.h>
#include <o2scl/base_ioc.h>

#include <o2scl/mroot.h>
#include <o2scl/odestep.h>
#include <o2scl/minimize.h>
#include <o2scl/gsl_min_brent.h>
#include <o2scl/gsl_mroot_hybrids.h>
#include <o2scl/ode_funct.h>
#include <o2scl/gsl_astep.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief A EOS base class for the TOV solver
  */
  class tov_eos {
    
  public:
    
    tov_eos() {
      baryon_column=false;
      verbose=1;
    }

    virtual ~tov_eos() {}
    
    /// control for output (default 1)
    int verbose;

    /** 
	\brief Given the pressure, produce the energy and number densities
	
	If the baryon density is not specified, it should be set to
	zero or \ref baryon_column should be set to false
    */
    virtual int get_eden(double P, double &e, double &nb) {
      return 0;
    }
    
    /// Given the pressure, produce all the remaining quantities 
    virtual int get_aux(double P, size_t &np, ovector_view &auxp) {
      return 0;
    }
    
    /** 
	\brief Fill a list with strings for the names of the remaining 
	quanities
    */
    virtual int get_names(size_t &np, std::vector<std::string> &pnames) {
      return 0;
    }
    
    /** 
	\brief Set to true if the baryon density is provided in the
	EOS (default false)
    */
    bool baryon_column;

  };

  /** 
      \brief An EOS for the TOV solver using simple linear interpolation
      and a default low-density EOS

      Internally, energy and pressure are stored in units of solar
      masses per cubic kilometer and baryon density is stored in units
      of fm^{-3}. The user-specified EOS table is left as is, and unit
      conversion is performed as needed in get_eden() and other
      functions so that results are returned in the units specified by
      set_units().

      The function set_units() needs to be called before either of the
      functions get_eden() or get_eden_ld() are called. The function
      set_units() may be called after calling either the read_table()
      functions or the set_low_density_eos() function.
     
  */
  class tov_interp_eos : public tov_eos {
    
  public:
    
    tov_interp_eos();

    virtual ~tov_interp_eos() {}

    /// Given the pressure, produce the energy and number densities
    virtual int get_eden(double pres, double &ed, double &nb);

    /** 
	\brief Given the pressure, produce the energy and number
	densities from the user-specified EOS
    */
    virtual int get_eden_user(double pres, double &ed, double &nb);

    /** 
	\brief Given the pressure, produce the energy and number
	densities from the low-density EOS
    */
    virtual int get_eden_ld(double pres, double &ed, double &nb);
    
    /// Given the pressure, produce all the remaining quantities 
    virtual int get_aux(double P, size_t &nv, ovector_view &auxp);
    
    /** 
	\brief Fill a list with strings for the names of the remaining 
	quanities
    */
    virtual int get_names(size_t &np, std::vector<std::string> &pnames);

    /// Specify the EOS through a table
    int read_table(table &eosat, std::string s_cole="ed", 
		   std::string s_colp="pr", std::string s_colnb="nb");
    
    /// Specify the EOS through a table in a file
    int read_table_file(std::string eosfn, std::string s_cole="ed", 
			std::string s_colp="pr", std::string s_colnb="nb");
    
    /** 
	\brief Set the low-density EOS
    */
    int set_low_density_eos(bool s_ldeos, std::string s_nvpath,
			    int s_nvcole=0, int s_nvcolp=1, int s_nvcolnb=2);

    /** \brief Set the units of the user-specified EOS
     */
    int set_units(double s_efactor, double s_pfactor, 
		  double s_nfactor);
    
    /** \brief Set the units of the user-specified EOS
     */
    int set_units(std::string leunits, 
		  std::string lpunits, std::string lnunits);
    
    /** 
	\brief Return limiting and transition pressures

	Returns, in order:
	- the highest pressure in the low-density EOS
	- the transition pressure
	- the lowest pressure in the high-density EOS
     */
    int get_transition(double &plow, double &ptrans, double &phi);
    
    /** 
	\brief Set the transition pressure and "width"

	Sets the transition pressure and the width (specified as a
	number greater than unity in \c pw) of the transition between
	the two EOSs. The transition is done smoothly using linear
	interpolation between \f$ P=\mathrm{ptrans}/pmathrm{pw} \f$
	and \f$ P=\mathrm{ptrans} \times pmathrm{pw} \f$.
     */
    int set_transition(double ptrans, double pw);

#ifndef DOXYGEN_INTERNAL

  protected:

    /// A base I/O object for reading EOSs
    base_ioc bio;
    
    /// Check that the EOS is valid
    int check_eos();
    
    /// \name Low-density EOS
    //@{

    /// true if we are using the low-density eos (false)
    bool ldeos;

    /** 
	\brief Low-density EOS switch
	
	This is \c true if the ldeos has been read by set_ldeos. This
	is useful, since then we know whether or not we need
	to free the memory for the LD EOS in the destructor
    */
    bool ldeos_read;

    /// the path to the low-density EOS 
    std::string ldpath;
    /// column in low-density eos for energy density
    int ldcole;
    /// column in low-density eos for pressure
    int ldcolp;
    /// column in low-density eos for baryon density
    int ldcolnb;
    /// file containing low-density EOS
    table *ld_eos;
    
    /// highest pressure in low-density EOS
    double presld;
    /// highest energy density in low-density EOS
    double eld;
    /// highest baryon density in low-density EOS
    double nbld;

    /// Transition pressure
    double prest;
    /// Transition width
    double pwidth;
    //@}
    
    /// \name User EOS
    //@{

    /// file containing eos
    table *eost;

    /// number of lines in eos file
    int nfile;

    /// column for energy density in eos file
    int cole;
    /// column for pressure in eos file
    int colp;
    /// column for baryon density in eos file
    int colnb;

    /// True if an EOS has been specified
    bool eos_read;
    //@}

    /// \name Units
    //@{
    
    /// Units for energy density
    std::string eunits;
    /// Units for pressure
    std::string punits;
    /// Units for baryon density
    std::string nunits;

    /// unit conversion factor for energy density (default 1.0)
    double efactor;
    /// unit conversion factor for pressure (default 1.0)
    double pfactor;
    /// unit conversion factor for baryon density (default 1.0)
    double nfactor;
    
    //@}

    /// Linear EOS interpolation
    void interp(const ovector_view &x, const ovector_view &y, 
		double xx, double &yy, int n1, int n2);

#endif

  };

  /** 
      \brief The Buchdahl EOS for the TOV solver

      Given the eos
      \f[
      \rho = 12 \sqrt{p_{*} P}- 5 P
      \f]
      the TOV equation has an analytical solution
      \f[
      R=(1-\beta) \sqrt{\frac{\pi}{288 p_{*} G (1-2 \beta)}}
      \f]
      where \f$ \beta = GM/R \f$.

      The central pressure and energy density are
      \f[
      P_c = 36 p_{*} \beta^2 
      \f]
      \f[
      {\rho}_c = 72 p_{*} \beta (1-5 \beta/2) 
      \f]

      Physical solutions are obtained only for \f$ P< 25 p_{*}/144 \f$
      and \f$ \beta<1/6 \f$ .

      (Lattimer & Prakash, 2001)
  */
  class tov_buchdahl_eos : public tov_eos {
    
  public:

    tov_buchdahl_eos() {
      Pstar=3.2e-5;
    }

    virtual ~tov_buchdahl_eos() {}

    /** 
	\brief The parameter with units of pressure in units of solar
	masses per km cubed (default value \f$ 3.2 \times 10^{-5} \f$
	)
    */
    double Pstar;
    
    /** 
	\brief Given the pressure, produce the energy and number densities
	
	If the baryon density is not specified, it should be set to
	zero or \ref baryon_column should be set to false
    */
    virtual int get_eden(double P, double &e, double &nb) {
      e=12.0*sqrt(Pstar*P)-5.0*P;
      nb=0.0;
      return 0;
    }
    
    /// Given the pressure, produce all the remaining quantities 
    virtual int get_aux(double P, size_t &np, ovector_view &auxp) {
      np=0;
      return 0;
    }
    
    /** 
	\brief Fill a list with strings for the names of the remaining 
	quanities
    */
    virtual int get_names(size_t &np, std::vector<std::string> &pnames) {
      np=0;
      return 0;
    }
    
#ifdef NEVER_DEFINED
    
    int buchfun(size_t bv, const ovector_view &bx, ovector_view &by, 
		void *&pa) {
      double u, rp;
      u=bx[0];
      rp=bx[1];
      by[0]=rp*(1.0-beta+u)/(1.0-2.0*beta)-buchrad;
      by[1]=beta/biga/rp*sin(biga*rp);
      return gsl_success;
    }

#endif

  };

  /// Standard polytropic EOS \f$ p = K \rho^{1+1/n} \f$
  class tov_polytrope_eos : public tov_eos {
    
  public:

    tov_polytrope_eos() {
      K=1.0;
      n=3.0;
    }

    virtual ~tov_polytrope_eos() {}

    /** 
	\brief Coefficient (default 1.0)
    */
    double K;

    /// Index (default 3.0)
    double n;

    /** 
	\brief Given the pressure, produce the energy and number densities
	
	If the baryon density is not specified, it should be set to
	zero or \ref baryon_column should be set to false
    */
    virtual int get_eden(double P, double &e, double &nb) {
      e=pow(P/K,1.0/(1.0+1.0/n));
      nb=0.0;
      return 0;
    }
    
    /// Given the pressure, produce all the remaining quantities 
    virtual int get_aux(double P, size_t &np, ovector_view &auxp) {
      np=0;
      return 0;
    }
    
    /** 
	\brief Fill a list with strings for the names of the remaining 
	quanities
    */
    virtual int get_names(size_t &np, std::vector<std::string> &pnames) {
      np=0;
      return 0;
    }
    
  };

#ifndef DOXYGENP
}
#endif

#endif


