/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_MIN_FIT_H
#define O2SCL_MIN_FIT_H

#include <gsl/gsl_rng.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlin.h>
#include <o2scl/multi_min.h>
#include <o2scl/multi_funct.h>
#include <o2scl/gsl_mmin_simp.h>
#include <o2scl/fit_base.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Non-linear least-squares fitting class with generic minimizer
    
      This minimizes a generic fitting function using any multi_min
      object, and then uses the GSL routines to calculate the
      uncertainties in the parameters and the covariance matrix.

      This can be useful for fitting problems which might be better
      handled by more complex minimizers than those that are used in
      gsl_fit. For problems with many local minima near the global
      minimum, using a sim_anneal object with this class can sometimes
      produce better results than gsl_fit.
  */
  template<class param_t, class func_t, class vec_t=ovector_view, 
    class mat_t=omatrix_view> class min_fit : 
  public fit_base<param_t,func_t,vec_t,mat_t> {
    public:
    
    min_fit() {
      min_set=false;
      mmp=&def_multi_min;
    }

    virtual ~min_fit() {}
  
    /** 
	\brief A structure for passing information to the GSL functions

	This structure is given so that the user can specify the 
	minimizer to use.
     */
    typedef struct {
      /// The fitting function
      func_t &f;
      /// The user-specified parameter
      param_t &vp;
      /// The number of data
      int ndat;
      /// The x values
      vec_t *xdat;
      /// The y values
      vec_t *ydat;
      /// The y uncertainties
      vec_t *yerr;
      /// The number of fitting parameters
      int npar;
    } func_par;

    /** \brief Fit the data specified in (xdat,ydat) to
	the function \c fitfun with the parameters in \c par.

	The covariance matrix for the parameters is returned in \c
	covar and the value of \f$ \chi^2 \f$ is returned in \c chi2.
    */
    virtual int fit(size_t ndat, vec_t &xdat, vec_t &ydat, vec_t &yerr,
		    size_t npar, vec_t &par, mat_t &covar, double &chi2,
		    param_t &pa, func_t &fitfun) {

      int status, i, iter=0;
      double dtemp;

      // ---------------------------------------------------
      // First minimize with the specified multi_min object
      // ---------------------------------------------------

      func_par fpar={fitfun,pa,ndat,&xdat,&ydat,&yerr,npar};
  
      multi_funct_mfptr_noerr<min_fit,func_par *> mmf(this,&min_fit::min_func);
      func_par *fp=&fpar;
      mmp->mmin(npar,par,dtemp,fp,mmf);
      
      gsl_multifit_fdfsolver *s=gsl_multifit_fdfsolver_alloc
      (gsl_multifit_fdfsolver_lmsder,ndat,npar);

      // -------------------------------------------------------------
      // Now use the GSL routines to calculate the covariance matrix,
      // and chi-squared.
      // -------------------------------------------------------------
  
      gsl_vector *x=(gsl_vector *)(&par);
      gsl_multifit_function_fdf f;
  
      f.f=func;
      f.df=dfunc;
      f.fdf=fdfunc;
      f.n=ndat;
      f.p=npar;
      f.params=&fpar;
     
      s=gsl_multifit_fdfsolver_alloc
	(gsl_multifit_fdfsolver_lmsder,ndat,npar);
      gsl_multifit_fdfsolver_set(s,&f,x);
  
      gsl_matrix *gcovar=gsl_matrix_alloc(npar,npar);
      gsl_multifit_covar(s->J,0.0,gcovar);
      for(i=0;i<((int)npar);i++) {
	par[i]=gsl_vector_get(s->x,i);
	for(int j=0;j<((int)npar);j++) {
	  covar[i][j]=gsl_matrix_get(gcovar,i,j);
	}
      }
      gsl_matrix_free(gcovar);
  
      chi2=gsl_blas_dnrm2(s->f);
      chi2*=chi2;

      gsl_multifit_fdfsolver_free(s);
  
      return 0;
    }
    
    /** \brief Set the multi_min object to use (default is gsl_mmin_nmsimplex)
     */
    int set_multi_min(multi_min<func_par *,
		      multi_funct<func_par *,vec_t> > &mm) {
      mmp=&mm;
      min_set=true;
      return 0;
    }
    
    /// The default minimizer
    gsl_mmin_simp<func_par *,multi_funct<func_par *,vec_t> > def_multi_min;

    /// Return string denoting type ("min_fit")
    virtual const char *type() { return "min_fit"; }

#ifndef DOXYGEN_INTERNAL

    protected:

    /// The minimizer
    multi_min<func_par *,multi_funct<func_par *,vec_t> > *mmp;
    
    /// True if the minimizer has been set by the user
    bool min_set;
    
    /// Evaluate the function
    static int func(const gsl_vector *x, void *pa, gsl_vector *f) {
      func_par *fp=(func_par *)pa;
      int i, j;
      double yi;
      ovector xp(fp->npar);
      for(j=0;j<fp->npar;j++) xp[j]=gsl_vector_get(x,j);
      for(i=0;i<fp->ndat;i++) {
	fp->f(fp->npar,xp,(*fp->xdat)[i],yi,fp->vp);
	gsl_vector_set(f,i,(yi-(*fp->ydat)[i])/(*fp->yerr)[i]);
      }

      return 0;
    }

    /// Evaluate the jacobian
    static int dfunc(const gsl_vector *x, void *pa, gsl_matrix *jac) {
      func_par *fp=(func_par *)pa;
      int i, j;
      double yi, ylo, yhi, xtemp, eps=1.0e-4;
      ovector xp(fp->npar);

      for(j=0;j<fp->npar;j++) xp[j]=gsl_vector_get(x,j);
      for(j=0;j<fp->npar;j++) {
	for(i=0;i<fp->ndat;i++) {
	  xtemp=xp[j];
	  xp[j]+=eps;
	  fp->f(fp->npar,xp,(*fp->xdat)[i],yi,fp->vp);
	  yhi=(yi-(*fp->ydat)[i])/(*fp->yerr)[i];
	  xp[j]=xtemp;
	  fp->f(fp->npar,xp,(*fp->xdat)[i],yi,fp->vp);
	  ylo=(yi-(*fp->ydat)[i])/(*fp->yerr)[i];
	  gsl_matrix_set(jac,i,j,(yhi-ylo)/eps);
	}
      }

      return 0;
    }

    /// Evaluate the function and the jacobian
    static int fdfunc(const gsl_vector *x, void *pa, gsl_vector *f,
		      gsl_matrix *jac) {
      func(x,pa,f);
      dfunc(x,pa,jac);
      return 0;
    }

    /// The function to minimize
    double min_func(size_t np, const vec_t &xp, func_par *&fp) {
      int i;
      double ret=0.0;
      double yi;

      ovector xpnoc(np);
      for(size_t ii=0;ii<np;ii++) xpnoc[ii]=xp[ii];
      
      for(i=0;i<fp->ndat;i++) {
	fp->f(np,xpnoc,(*fp->xdat)[i],yi,fp->vp);
	ret+=pow((yi-(*fp->ydat)[i])/(*fp->yerr)[i],2.0);
      }

      return ret;
    }
    
#endif
  
  };

#ifndef DOXYGENP
}
#endif

#endif
