/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#include <o2scl/multi_funct.h>
#include <o2scl/gsl_mmin_bfgs2.h>
#include <gsl/gsl_multimin.h>
#include <o2scl/test_mgr.h>
#include <o2scl/constants.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

double minfun(size_t n, const ovector_view &x, void *&pa) {
  double ret;
  ret=x[0]*x[0]+(x[1]-2.0)*(x[1]-2.0)+3.0;
  return(ret);
}

// A difficult function to minimize
double spring(size_t nv, const ovector_view &x, void *&pa) {
  double theta=atan2(x[1],x[0]);
  double r=sqrt(x[0]*x[0]+x[1]*x[1]);
  double z=x[2];
  while (z>pi) z-=2.0*pi;
  while (z<-pi) z+=2.0*pi;
  double tmz=theta-z;
  double rm1=r-1.0;
  double ret=exp(tmz*tmz+rm1*rm1)+fabs(x[2]/10.0);

  return ret;
}

double my_f(const gsl_vector *v, void *params) {
  double x, y;
  double *dp = (double *)params;
     
  double x0 = gsl_vector_get(v, 0);
  double x1 = gsl_vector_get(v, 1);
  double x2 = gsl_vector_get(v, 2);

  double theta=atan2(x1,x0);
  double r=sqrt(x0*x0+x1*x1);
  double z=x2;
  while (z>pi) z-=2.0*pi;
  while (z<-pi) z+=2.0*pi;
  double tmz=theta-z;
  double rm1=r-1.0;
  double ret=exp(tmz*tmz+rm1*rm1)+fabs(x2/10.0);

  return ret;
}
     
void my_df(const gsl_vector *v, void *params, gsl_vector *df) {
     
  double fv1, fv2, deriv_h=1.0e-4;

  fv1=my_f(v,params);
  
  gsl_vector *v2=gsl_vector_alloc(3);

  for(size_t i=0;i<3;i++) {
    gsl_vector_set(v2,i,gsl_vector_get(v,i));
  }
  for(size_t i=0;i<3;i++) {
    gsl_vector_set(v2,i,gsl_vector_get(v,i)+deriv_h);
    fv2=my_f(v2,params);
    gsl_vector_set(v2,i,gsl_vector_get(v,i));
    gsl_vector_set(df,i,(fv2-fv1)/deriv_h);
  }
}
     
void my_fdf(const gsl_vector *x, void *params, double *f, gsl_vector *df) {
  *f=my_f(x, params);
  my_df(x, params, df);
}

int main(void) {
  test_mgr t;
  t.set_output_level(2);

  cout.setf(ios::scientific);

  double min=0.0;
  ovector x3(3), x2(2);
  void *vp=0;
  gsl_mmin_bfgs2<void *,multi_funct<void *> > g;

  cout.setf(ios::scientific);

  multi_funct_fptr_noerr<void *> mf(minfun);
  multi_funct_fptr_noerr<void *> mf2(spring);
  
  x2[0]=10.0;
  x2[1]=10.0;
  int ret=g.mmin(2,x2,min,vp,mf);
  cout << ret << " " << min << " " << x2[0] << " " << x2[1] << endl;
  cout << err_hnd->get_str() << endl;
  t.test_abs(x2[0],0.0,1.0e-4,"gsl_mmin_bfgs2 1");
  t.test_rel(x2[1],2.0,2.0e-4,"gsl_mmin_bfgs2 2");
  cout << endl;

  x3[0]=1.0;
  x3[1]=0.0;
  x3[2]=2.0*o2scl_const::pi;
  g.verbose=2;
  ret=g.mmin(3,x3,min,vp,mf2);
  cout << ret << " " << min << " " << x3 << endl;
  cout << err_hnd->get_str() << endl;
  cout << endl;

  {
    size_t iter = 0;
    int status;
     
    const gsl_multimin_fdfminimizer_type *T;
    gsl_multimin_fdfminimizer *s;
     
    gsl_vector *x;
    gsl_multimin_function_fdf my_func;
     
    my_func.f = &my_f;
    my_func.df = &my_df;
    my_func.fdf = &my_fdf;
    my_func.n = 3;
    my_func.params = 0;
     
    /* Starting point, x = (5,7) */
    x = gsl_vector_alloc (3);
    gsl_vector_set (x, 0, 1.0);
    gsl_vector_set (x, 1, 0.0);
    gsl_vector_set (x, 2, 2.0*pi);
     
    T = gsl_multimin_fdfminimizer_vector_bfgs2;
    s = gsl_multimin_fdfminimizer_alloc (T, 3);
     
    gsl_multimin_fdfminimizer_set (s, &my_func, x, 0.01, 1e-4);
     
    do
      {
	iter++;
	status = gsl_multimin_fdfminimizer_iterate (s);
     
	if (status)
	  break;
     
	status = gsl_multimin_test_gradient (s->gradient, 1e-3);

	if (false) {
	  cout << gsl_vector_get(s->x,0) << " "
	       << gsl_vector_get(s->x,1) << " "
	       << gsl_vector_get(s->x,2) << endl;
	}
	
      }
    while (status == GSL_CONTINUE && iter < 200);
    
    cout << status << " " << gsl_vector_get(x,0) << " "
	 << gsl_vector_get(x,1) << " "
	 << gsl_vector_get(x,2) << endl;
    cout << err_hnd->get_str() << endl;

    gsl_multimin_fdfminimizer_free (s);
    gsl_vector_free (x);
    
  }

  t.report();
  return 0;
}
 
