/*
  -------------------------------------------------------------------

  Copyright (C) 2006, 2007, Andrew W. Steiner

  This file is part of O2scl.

  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_GSL_MMIN_SIMP_H
#define O2SCL_GSL_MMIN_SIMP_H

#include <o2scl/multi_min.h>
#include <o2scl/gsl_mmin_simp_b.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Multidimensional minimization by the Simplex method (GSL)

      \todo Not properly generalized to non-GSL vectors (to do this,
      I'll have to store the simplex as a vec_t array rather than a 
      gsl_matrix). Right now, this only works with \c vec_t \c = 
      \ref ovector_view.
      \todo Add a minimize function which allows specification
      of the entire simplex.
      \todo Gracefully ensure memory allocation and deallocation 
      is performed automatically.
      \todo Test mmin_twovec().
      \todo Document how set() chooses the simplex from the initial
      guess and step size
  */
  template<class param_t,class func_t=multi_funct<param_t>,
    class vec_t=ovector_view> class gsl_mmin_simp :
  public multi_min<param_t,func_t,func_t,vec_t>, public gsl_mmin_simp_b 
  {
    
#ifndef DOXYGEN_INTERNAL
    
    protected:
    
    /** 
	\brief Move a corner of a simplex

	Moves a simplex corner scaled by coeff (negative value
	represents mirroring by the middle point of the "other" corner
	points) and gives new corner in xc and function value at xc as
	a return value.
    */
    virtual double move_corner(const double coeff, const simp_state_t *state,
			       size_t corner, gsl_vector *xc, func_t &f, 
			       size_t nvar, param_t &pa) {
      
      gsl_matrix *x1=state->x1;
	
      size_t i,j;
      double newval,mp;
	
      if(x1->size1<2) {
	set_err("Simplex too small in gsl_mmin_simp::move_corner().",
		gsl_efailed);
	return 0.0;
      }
	
      for (j=0;j<x1->size2;j++) {
	mp=0.0;
	for (i=0;i<x1->size1;i++) {
	  if (i!=corner) {
	    mp+=(gsl_matrix_get(x1,i,j));
	  }
	}
	mp/=(double)(x1->size1-1);
	newval=mp-coeff*(mp-gsl_matrix_get(x1,corner,j));
	gsl_vector_set(xc,j,newval);
      }
      
      f(nvar,*((ovector *)xc),newval,pa);
	
      return newval;
    }
      
    /** 
	\brief Contract the simplex towards the best point
	  
	Function contracts the simplex in respect to best valued
	corner. All corners besides the best corner are moved.
	
	The xc vector is used as work space.
    */
    virtual int contract_by_best(simp_state_t *state, size_t best, 
				 gsl_vector *xc, func_t &f,
				 size_t nvar, param_t &pa) {
	
      gsl_matrix *x1=state->x1;
      gsl_vector *y1=state->y1;
  
      size_t i,j;
      double newval;
  
      for (i=0;i<x1->size1;i++) {
	if (i!=best) {
	  for (j=0;j<x1->size2;j++) {
	    newval=0.5*(gsl_matrix_get(x1,i,j)
			+gsl_matrix_get(x1,best,j));
	    gsl_matrix_set(x1,i,j,newval);
	  }
	    
	  /* evaluate function in the new point */
	  gsl_matrix_get_row(xc,x1,i);
	  f(nvar,*((ovector *)xc),newval,pa);
	  gsl_vector_set(y1,i,newval);
	}
      }
  
      return gsl_success;
    }

    /// Number of variables to be minimized over
    size_t dim;

    /// Present minimum vector
    gsl_vector *x;

    /// Function value at minimum
    double fval;

    /// Function
    func_t *func;

    /// Parameters
    param_t *params;

    /// True if set() has been called
    bool set_called;

    /// Size of simplex
    double size;

    /// Simplex state storage
    simp_state_t lstate;

#endif

    public:

    gsl_mmin_simp() {
      dim=0;
      initial_step=1.0;
      print_simplex=0;
    }
    
    virtual ~gsl_mmin_simp() {}
    
    /// The initial stepsize (default 1.0)
    double initial_step;
    
    /** 
	\brief Print simplex information in print_iter() (default 0)
	
	If this is 1 and \ref verbose is greater than 0, 
	then print_iter() will print the function values at all 
	the simplex points. 
    */
    int print_simplex;

    /** \brief Calculate the minimum \c min of \c func w.r.t the
	array \c x of size \c nvar.
    */
    virtual int mmin(size_t nn, vec_t &xx, double &fmin, param_t &pa,
		     func_t &ufunc) {
      
      int ret=0,status,iter=0;
      
      allocate(nn);
      
      ovector ss(nn);
      for (size_t is=0;is<nn;is++) ss[is]=initial_step;
      ret=set(ufunc,pa,xx,ss);
      if(ret!=0) {
	free();
	return ret;
      }
  
      do {
	iter++;
	  
	status=iterate();
	if(status) break;

	if(this->verbose>0) {
	  ovector *tx=(ovector *)(x);
	  print_iter(nn,*tx,fval,iter,size,this->tolx,
		     "gsl_mmin_nmsimplex");
	}
    
	status=gsl_multimin_test_size(size,this->tolx);
	  
      } while(status == GSL_CONTINUE && iter<this->ntrial);
	
      for (size_t i=0;i<nn;i++) xx[i]=gsl_vector_get(x,i);      
      fmin=fval;
  
      if(iter>=this->ntrial) {
	set_err_ret("Exceeded max # of iterations in gsl_mmin_simp::mmin().",
		    gsl_emaxiter);
      }

      free();
      this->last_ntrial=iter;
      
      return status;
    }
    
    /** \brief Calculate the minimum \c min of \c func w.r.t the
	array \c x of size \c nvar, using \c xx and \c xx2 to specify
	the simplex
    */
    virtual int mmin_twovec(size_t nn, vec_t &xx, vec_t &xx2, double &fmin, 
			    param_t &pa, func_t &ufunc) {
      
      int ret=0,i,status,iter=0;
      
      allocate(nn);
	
      ovector ss(nn);
      for (size_t is=0;is<nn;is++) ss[is]=xx2[is]-xx[is];
      ret=set(ufunc,pa,xx,ss);
      if(ret!=0) {
	free();
	return ret;
      }
  
      do {
	iter++;
	  
	status=iterate();
	if(status) break;

	if(this->verbose>0) {
	  ovector *tx=(ovector *)(x);
	  print_iter(nn,*tx,fval,iter,size,this->tolx,
		     "gsl_mmin_nmsimplex");
	}
    
	status=gsl_multimin_test_size(size,this->tolx);
	  
      } while(status == GSL_CONTINUE && iter<this->ntrial);
	
      for (i=0;i<((int)nn);i++) xx[i]=gsl_vector_get(x,i);
      fmin=fval;
  
      if(iter>=this->ntrial) {
	set_err_ret
	  ("Exceeded max # of iterations in gsl_mmin_simp::mmin_twovec().",
	   gsl_emaxiter);
      }

      free();
      this->last_ntrial=iter;

      return status;
    }
    
    /// Allocate the memory
    virtual int allocate(size_t n) {
      int status;
      if(dim!=0) free();
      set_called=false;
      dim=n;

      x=gsl_vector_calloc(n);
      if(x == 0) {
	GSL_ERROR_VAL("failed to allocate space for x",GSL_ENOMEM,0);
      }
  
      lstate.x1=gsl_matrix_alloc(n+1,n);
      if(lstate.x1 == NULL) {
	set_err("Failed to allocate x1 in gsl_mmin_nmsimplex::allocate().",
		gsl_enomem);
      }
  
      lstate.y1=gsl_vector_alloc(n+1);
      if(lstate.y1 == NULL) {
	set_err("Failed to allocate y1 in gsl_mmin_nmsimplex::allocate().",
		gsl_enomem);
      }
  
      lstate.ws1=gsl_vector_alloc(n);
      if(lstate.ws1 == NULL) {
	set_err("Failed to allocate ws1 in gsl_mmin_nmsimplex::allocate().",
		gsl_enomem);
      }
  
      lstate.ws2=gsl_vector_alloc(n);
      if(lstate.ws2 == NULL) {
	set_err("Failed to allocate ws2 in gsl_mmin_nmsimplex::allocate().",
		gsl_enomem);
      }

      return gsl_success;
    }
    
    /// Free the allocated memory
    virtual int free() {

      gsl_matrix_free(lstate.x1);
      gsl_vector_free(lstate.y1);
      gsl_vector_free(lstate.ws1);
      gsl_vector_free(lstate.ws2);
      gsl_vector_free(x);

      dim=0;

      return 0;
    }
    
    /// Set the function and initial guess
    virtual int set(func_t &ufunc, param_t &pa, vec_t &ax, 
		    vec_t &step_size) {
      size_t i;
		
      if(dim==0) {
	set_err_ret("Memory not allocated in gsl_mmin_nmsimplex::set().",
		    gsl_ebadlen);
      }
      
      if(dim>ax.size()) {
	set_err_ret("Initial guess<solver size in gsl_mmin_nmsimplex::set().",
		    gsl_ebadlen);
      }
      
      params=&pa;
   
      // We can't use gsl_vector_memcpy here,since ax may have a
      // different size than dim.
      for (i=0;i<dim;i++) gsl_vector_set(x,i,ax[i]);
 
      func=&ufunc;
  
      double val;
      int status;
  
      gsl_vector *xtemp=lstate.ws1;
  
      /* first point is the original x0 */
	
      ufunc(dim,ax,val,pa);
      gsl_matrix_set_row(lstate.x1,0,x);
      gsl_vector_set(lstate.y1,0,val);
  
      /* following points are initialized to x0+step_size */
 
      for (i=0;i<x->size;i++) {
	status=gsl_vector_memcpy(xtemp,x);
 
	if(status != 0)      {
	  GSL_ERROR("vector memcopy failed",GSL_EFAILED);
	}
	    
	val=gsl_vector_get(xtemp,i)+
	  gsl_vector_get((gsl_vector *)(&step_size),i);
	gsl_vector_set(xtemp,i,val);
	ufunc(dim,*((ovector *)xtemp),val,pa);
	gsl_matrix_set_row(lstate.x1,i+1,xtemp);
	gsl_vector_set(lstate.y1,i+1,val);
      }
 
      /* Initialize simplex size */
	
      size=nmsimplex_size(&lstate);
	
      set_called=true;
 
      return gsl_success;
    }

    /// Perform an iteration
    virtual int iterate() {
	
      simp_state_t *nstate=&lstate;
      
      /* xc and xc2 vectors store tried corner point coordinates */
 
      gsl_vector *xc=nstate->ws1;
      gsl_vector *xc2=nstate->ws2;
      gsl_vector *y1=nstate->y1;
      gsl_matrix *x1=nstate->x1;
 
      size_t n=y1->size;
      size_t i;
      size_t hi=0,s_hi=0,lo=0;
      double dhi,ds_hi,dlo;
      int status;
      double val,val2;
 
      /* get index of highest,second highest and lowest point */
 
      dhi=ds_hi=dlo=gsl_vector_get(y1,0);
 
      for (i=1;i<n;i++) {
	val =(gsl_vector_get(y1,i));
	if(val<dlo) {
	  dlo=val;
	  lo=i;
	} else if (val > dhi) {
	  ds_hi=dhi;
	  s_hi=hi;
	  dhi=val;
	  hi=i;
	} else if(val > ds_hi) {
	  ds_hi=val;
	  s_hi=i;
	}
      }
 
      /* reflect the highest value */
	
      val=move_corner(-1.0,nstate,hi,xc,*func,dim,*params);
  
      if (val<gsl_vector_get(y1,lo)) {
      
	/* reflected point becomes lowest point,try expansion */
      
	val2=move_corner(-2.0,nstate,hi,xc2,*func,dim,*params);
	if (val2<gsl_vector_get(y1,lo)) {
	  gsl_matrix_set_row(x1,hi,xc2);
	  gsl_vector_set(y1,hi,val2);
	} else {
	  gsl_matrix_set_row(x1,hi,xc);
	  gsl_vector_set(y1,hi,val);
	}

      } else if (val > gsl_vector_get(y1,s_hi)) {
	
	/* reflection does not improve things enough */
	
	if(val <= gsl_vector_get(y1,hi)) {
	    
	  /* if trial point is better than highest point,replace
	     highest point */
	    
	  gsl_matrix_set_row(x1,hi,xc);
	  gsl_vector_set(y1,hi,val);
	}
      
	/* try one dimensional contraction */
	  
	val2=move_corner(0.5,nstate,hi,xc2,*func,dim,*params);
      
	if(val2 <= gsl_vector_get(y1,hi)) {

	  gsl_matrix_set_row(nstate->x1,hi,xc2);
	  gsl_vector_set(y1,hi,val2);

	} else {
	  
	  /* contract the whole simplex in respect to the best point */
	  
	  status=contract_by_best(nstate,lo,xc,*func,dim,*params);
	  if(status != 0) {
	    set_err("Function contract_by_best() failed in iterate().",
		    gsl_efailed);
	  }
	}

      } else {
      
	/* trial point is better than second highest point.
	   Replace highest point by it */
      
	gsl_matrix_set_row(x1,hi,xc);
	gsl_vector_set(y1,hi,val);
      }
  
      /* return lowest point of simplex as x */
  
      lo=gsl_vector_min_index(y1);
      gsl_matrix_get_row(x,x1,lo);
      fval=gsl_vector_get(y1,lo);
  
      /* Update simplex size */
  
      size=nmsimplex_size(nstate);

      return gsl_success;
    }
      
    /** 
	\brief Print out iteration information.
	  
	Depending on the value of the variable verbose, this prints out
	the iteration information. If verbose=0, then no information is
	printed, while if verbose>1, then after each iteration, the
	present values of x and y are output to std::cout along with the
	iteration number. If verbose>=2 then each iteration waits for a
	character.
    */
    virtual int print_iter(size_t nv, vec_t &xx, double y, int iter,
			   double value, double limit,
			   std::string comment) {
      if (this->verbose<=0) return 0;
      
      int i;
      char ch;
	
      std::cout << comment << " Iteration: " << iter << std::endl;
      text_out_file outs(&std::cout,79);
      outs.word_out("x:");
      for(i=0;i<((int)nv);i++) outs.double_out(xx[i]);
      outs.end_line();
      if (print_simplex==1) {
	ovector_view *ov=(ovector_view *)(lstate.y1);
	outs.word_out("simp vals:");
	for(i=0;i<((int)nv+1);i++) {
	  outs.double_out((*ov)[i]);
	}
	outs.end_line();
      }
      std::cout << "y: " << y << " Val: " << value << " Lim: " 
      << limit << std::endl;
      if (this->verbose>1) {
	std::cout << "Press a key and type enter to continue. ";
	std::cin >> ch;
      }
	
      return 0;
    }

    /// Return string denoting type("gsl_mmin_simp")
    virtual const char *type() { return "gsl_mmin_simp";}

  };
  
#ifndef DOXYGENP
}
#endif

#endif
