/*
  -------------------------------------------------------------------

  Copyright (C) 2006, 2007, Andrew W. Steiner

  This file is part of O2scl.

  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_CONTOUR_H
#define O2SCL_CONTOUR_H

#include <gsl/gsl_math.h>
#include <o2scl/smart_interp.h>
#include <o2scl/omatrix_tlate.h>
#include <o2scl/pinside.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Calculate contour lines from a two-dimensional data set

      \b Basic \b Usage

      - Specify the data as a two-dimensional square grid of "heights"
      with set_data().
      - Specify the contour levels with set_levels().
      - Compute the contours with calc_contours() which returns
      the number of contours (which is often larger than the number
      of contour levels, since one level can result in multiple contours).
      - Retrieve the contours individually using calls to
      get_contour().

      The data should be stored so that the y-index is first, i.e.
      data[iy][ix]. One can always switch \c x_fun and \c y_fun if
      this is not the case. The data is copied by set_data(), so
      changing the data will not change the contours unless set_data()
      is called again. The functions set_levels() and calc() can be
      called several times for the same data without calling
      set_data() again.

      Linear interpolation is used to decide whether or not a line
      segment and a contour cross. This choice is intentional, since
      (in addition to making the algorithm much simpler) 
      it is the user (and not the class) which is likely best able to
      refine the data. In case a simple refinement scheme is desired,
      the method regrid_data() is provided which uses cubic spline
      interpolation to refine the data and thus make the curves more
      continuous.

      Since linear interpolation is used, the contour calculation
      implicitly assumes that there is not more than one intersection
      of any contour level with any line segment, so if this is the
      case, then either a more refined data set should be specified or
      regrid_data() should be used. For contours which do not close
      inside the region of interest, the results will always end at
      either the minimum or maximum values of \c x_fun or \c y_fun (no
      extrapolation is ever done).

      As an example, for the function
      \f[
      15 e^{-(x-20)^2/400-(y-5)^2/25}+
      40 e^{-(x-70)^2/4900-(y-2)^2/4}
      \f]
      a 10x10 grid gives the contours:
      \image html contourg.png "contourg.png"
      \image latex contourg.eps "contourg.eps" width=9cm

      While after a call to regrid_data(3,3), the contours
      are a little smoother:
      \image html contourg2.png "contourg2.png"
      \image latex contourg2.eps "contourg2.eps" width=9cm

      Mathematica gives a similar result:
      \image html contourg3.png "contourg3.png"
      \image latex contourg3.eps "contourg3.eps" width=9cm

      \b The \b Algorithm:

      This works by viewing the data as defining a square
      two-dimensional grid. The function calc() exhaustively
      enumerates every line segment in the grid which involves a level
      crossing and then organizing the points defined by the
      intersection of a line segment with a level curve into a full
      contour. 

      \b Representing \b Contours \b by \b Fill \b Regions
      
      If the user wants to "shade" the contours to provide a 
      shaded or colored contour plot, then it is useful
      to provide a set of closed contours to be shaded instead 
      of the (possibly) open contours returned by calc_contours().
      After a call to calc_contours(), the function regions() can
      be used to organize the closed contours into regions to 
      be shaded. Open contours are closed by adding points defining
      lines along the edges of the data and closed contours are 
      inverted (if necessary) by adding an external line emanating
      from the closed contour and properly including the boundary 
      edges.

      \todo 
      - Some contours which should be closed are not properly
      closed yet. See the tests for examples which fail.
      - Use twod_intp for regrid_data
      - Work on how memory is allocated
      - Include the functionality to provide regions to be shaded
      instead of just lines. This can be done by providing a method,
      i.e. regions() which converts the curves into regions.

  */
  class contour {
    
  public:
    
    contour();
    ~contour();

    /// \name Basic usage
    //@{

    /** 
	\brief Set the data 

	The types \c vec_t and \c mat_t can be any types which have \c
	operator[] and \c operator[][] for array and matrix indexing.

	Note that this method copies all of the user-specified data to 
	local storage so that changes in the data after calling this function
	will not be reflected in the contours that are generated.
    */
    template<class vec_t, class mat_t>
      int set_data(size_t sizex, size_t sizey, const vec_t &x_fun, 
		   const vec_t &y_fun, const mat_t &udata) {
      int i;
      
      if (sizex<2 || sizey<2) { 
	set_err_ret("Not enough data (must be at least 2x2) in set_data().",
		    gsl_einval);
      }
      
      free_memory();
      
      nx=sizex;
      ny=sizey;
      xfun=new ovector(nx);
      yfun=new ovector(ny);
      for(int is=0;is<nx;is++) (*xfun)[is]=x_fun[is];
      for(int is=0;is<ny;is++) (*yfun)[is]=y_fun[is];
      
      data=new ovector *[ny];
      for(i=0;i<ny;i++) {
	data[i]=new ovector(nx);
	for(int j=0;j<nx;j++) {
	  (*(data[i]))[j]=udata[i][j];
	}
      }
      
      return 0;
    }

    /** 
	\brief Set the contour levels 
	
	This is separate from the function calc_contours() so that
	the user can compute the contours for different data sets using
	the same levels
    */
    template<class vec_t> int set_levels(size_t nlevels, vec_t &ulevels) {
      nlev=nlevels;
      levels.allocate(nlevels);
      for(size_t i=0;i<nlevels;i++) {
	levels[i]=ulevels[i];
      }
      levels_set=true;
      return 0;
    }

    /** 
	\brief Calculate the contours 

	The function calc_contours() returns the total number of
	contours found.  Since there may be more than one disconnected
	contours for the same contour level, or no contours for a
	given level, the total number of contours may be less than or
	greater than the number of levels given by set_levels().

	If an error occurs, zero is returned.
    */
    template<class vec_t> 
      int calc_contours(vec_t &new_levels, bool debug=false) {
      int i,j,k,ileft,iright,m;
      
      for(i=0;i<nlev;i++) new_levels[i]=levels[i];

      if (nx==0) {
	set_err("Data not set in calc_contours().",gsl_einval);
	return 0;
      }
      if (levels_set==false) {
	set_err("Contour levels not set in calc_contours().",gsl_einval);
	return 0;
      }

      // Free previously stored contours
      if (ncurves!=0) {
	csizes.clear();
	vals.clear();
	xc.clear();
	yc.clear();
	ncurves=0;
      }
  
      // Free space for previously computed edges
      for(j=0;j<((int)(re.size()));j++) {
	delete re[j];
	delete be[j];
	delete redges[j];
	delete bedges[j];
      }
      re.clear();
      be.clear();
      redges.clear();
      bedges.clear();

      // Temporary storage for the interpolation arrays
      int nint;
      double *xi, *yi;

      // Some useful variables for looping over the lines
      bool foundline, linedone;
      int fp;

      // Temporary storage for the curves
      std::vector<double> x, y;

      // The interpolation object

      linear_interp<ovector_view> it1;
      linear_interp<ovector_const_reverse> it2;
      linear_interp<ovector_const_subvector> it3;
      linear_interp<ovector_const_subvector_reverse> it4;
      sm_interp *si=new sm_interp(it1,it2,it3,it4);
  
      // Start with zero curves
      ncurves=0;

      // For each level
      for(i=0;i<nlev;i++) {

	// Make space for the edges and add to the list
	rep=new omatrix_int(ny-1,nx);
	bep=new omatrix_int(ny,nx-1);
	redgesp=new omatrix(ny-1,nx);
	bedgesp=new omatrix(ny,nx-1);
	re.push_back(rep);
	be.push_back(bep);
	redges.push_back(redgesp);
	bedges.push_back(bedgesp);

	if (verbose>0) {
	  std::cout << "\nLooking for edges for level: " 
		    << new_levels[i] << std::endl;
	}

	// Examine the each of the rows for an intersection
	find_intersections(i,new_levels[i]);
	
	if (verbose>0) {
	  std::cout << "\nInterpolating edge intersections for level: " 
		    << new_levels[i] << std::endl;
	}

	// Process the right edges
	right_edges(new_levels[i],si);
	
	// Process the bottom edges
	bottom_edges(new_levels[i],si);

	if (verbose>0) {
	  std::cout << "\nPiecing together contour lines for level: " 
		    << new_levels[i] << std::endl;
	}

	// Now go through and one side of the line
	foundline=true;
	while(foundline==true) {
	  for(j=0;j<nx;j++) {
	    for(k=0;k<ny;k++) {
	      foundline=false;
	  
	      // A line beginning with a right edge
	      if (k<ny-1 && (*rep)[k][j]==edge) {
		if (verbose>0) {
		  std::cout << "Starting contour line: " << std::endl;
		  std::cout << "(" << (*xfun)[j] << ", " << (*redgesp)[k][j] 
			    << ")" << std::endl;
		}
		x.push_back((*xfun)[j]);
		y.push_back((*redgesp)[k][j]);
		(*rep)[k][j]++;
	    
		// Go through both sides
		process_line(j,k,dright,x,y,true);
		if (verbose>0) {
		  std::cout << "Computing other side of line." << std::endl;
		}
		process_line(j,k,dright,x,y,false);
		foundline=true;
	      }

	      // A line beginning with a bottom edge
	      if (j<nx-1 && foundline==false && (*bep)[k][j]==edge) {
		if (verbose>0) {
		  std::cout << "Starting contour line: " << std::endl;
		  std::cout << "(" << (*bedgesp)[k][j] << ", " << (*yfun)[k] 
			    << ")" << std::endl;
		}
		x.push_back((*bedgesp)[k][j]);
		y.push_back((*yfun)[k]);
		(*bep)[k][j]++;
	    
		// Go through both sides
		process_line(j,k,dbottom,x,y,true);
		if (verbose>0) {
		  std::cout << "Computing other side of line." << std::endl;
		}
		process_line(j,k,dbottom,x,y,false);
		foundline=true;
	      }

	      // Now repackage the line information in x and y
	      // to xc and yc and update csizes, vals and ncurves:
	      if (foundline==true) {
		csizes.push_back(x.size());
		vals.push_back(new_levels[i]);
		ncurves++;
		std::vector<double> xv, yv;
		for(m=0;m<((int)x.size());m++) {
		  xv.push_back(x[m]);
		  yv.push_back(y[m]);
		}
		xc.push_back(xv);
		yc.push_back(yv);
		x.clear();
		y.clear();
	      }
	    }
	  }
	}
      }
  
      return ncurves;
    }

    /// Return the number of points in the specified contour
    int get_contour_size(int index);
    
    /** 
	\brief Get a contour

	Given the \c index, which is between 0 and the number of contours
	returned by calc_contours() minus 1 (inclusive), this
	returns the level for this contour with the x and y-values
	in \c x and \c y which are both of length \c csize.
	The vectors \c x and \c y must have been previously allocated.
	The user can obtain the necessary size for \c x and \c y by
	calling get_contour_size(). 
    */
    template<class vec_t> int get_contour(int index, double &val, int &csize,
					  vec_t &x, vec_t &y) {
      int i;
      if (index>=ncurves) {
	set_err_ret("No curve with specified index in get_contour().",
		    gsl_efailed);
      }
      
      csize=csizes[index];
      val=vals[index];
      for(i=0;i<csize;i++) {
	x[i]=xc[index][i];
	y[i]=yc[index][i];
      }
      return 0;
    }

    /** 
	\brief Compute closed contour regions (unfinished)
    */
    int regions(bool larger);
    
    //@}

    /// \name Regrid function
    //@{
    /** 
	\brief Regrid the data 
	
	The uses cubic spline interpolation to refine the data set,
	ideally used before attempting to calculate the contour
	levels. If the original number of data points is
	\f$(\mathrm{nx},\mathrm{ny})\f$, then the new number of data
	points is
	\f[
	(\mathrm{xfact}~(\mathrm{nx}-1)+1,
	\mathrm{yfact}~(\mathrm{ny}-1)+1)
	\f]
    */
    int regrid_data(size_t xfact, size_t yfact);
    //@}

    /// \name Obtain internal data
    //@{
    /** 
	\brief Get the data 
	
	This is useful to see how the data has changed after
	a call to regrid_data().
    */
    int get_data(size_t &sizex, size_t &sizey, const ovector *&x_fun, 
		 const ovector *&y_fun, const ovector **&udata) {
      if (nx==0) {
	set_err_ret("Data not set in calc().",gsl_einval);
      }
      sizex=nx;
      sizey=ny;
      x_fun=xfun;
      y_fun=yfun;
      udata=(const ovector **)data;
      return 0;
    }

    /** \brief Return the edges used to compute the contours
     */
    int get_edges(const std::vector<omatrix_int *> *rints,
		  const std::vector<omatrix_int *> *bints,
		  const std::vector<omatrix *> *rpoints,
		  const std::vector<omatrix *> *bpoints);
    
    /** 
	\brief Return the edges used to compute the contours

	This function allocates memory for \c rints, \c bints, 
	\c rpoints, and \c bpoints and fills them with a copy of 
	the data that the class is using. As such, there is no
	need for them to be const.
    */
    int get_edges_for_level(size_t nl, omatrix_int &rints, omatrix_int &bints,
			    omatrix &rpoints, omatrix &bpoints);
    
    //@}

    /// Verbosity parameter
    int verbose;

    /// (default \f$ 10^{-8} \f$)
    double lev_adjust;
    
#ifndef DOXYGEN_INTERNAL

  protected:

    /// \name Edge direction
    //@{
    static const int dright=0;
    static const int dbottom=1;
    //@}

    /// \name Edge status
    //@{
    static const int empty=0;
    static const int edge=1;
    static const int contourp=2;
    static const int endpoint=3;
    //@}

    /// \name Edge found or not found
    //@{
    static const int efound=1;
    static const int enot_found=0;
    //@}

    /// \name User-specified data
    //@{
    int nx, ny;
    ovector *xfun, *yfun, **data;
    //@}

    int new_debug;

    /// \name User-specified contour levels
    //@{
    int nlev;
    ovector levels;
    bool levels_set;
    //@}

    /// \name Generated curves
    //@{
    int ncurves;
    std::vector<int> csizes;
    std::vector<double> vals;
    std::vector< std::vector<double> > xc, yc;
    //@}

    /// \name Storage for edges
    //@{
    std::vector<omatrix *> redges, bedges;
    std::vector<omatrix_int *> re, be;
    omatrix_int *rep;
    omatrix_int *bep;
    omatrix *redgesp;
    omatrix *bedgesp;
    //@}

    /// Object to find if a point is inside a polygon
    pinside pi;

    /// Find next point starting from a point on a right edge
    int find_next_point_right(int j, int k, int &jnext, int &knext, 
			      int &dir_next, int nsw=1);

    /// Find next point starting from a point on a bottom edge
    int find_next_point_bottom(int j, int k, int &jnext, int &knext, 
			       int &dir_next, int nsw=1);

    /// Find all of the intersections of the edges with the contour level
    int find_intersections(size_t ilev, double &level);

    /// Interpolate all right edge crossings 
    int right_edges(double level, o2scl::sm_interp *si);
    
    /// Interpolate all bottom edge crossings
    int bottom_edges(double level, o2scl::sm_interp *si);

    /// Create a contour line from a starting edge
    int process_line(int j, int k, int dir, std::vector<double> &x, 
		     std::vector<double> &y, bool first=true);

    /** 
	\brief Test if a point is inside a closed contour (unfinished)
	
	This returns true if the point (x1,y1) is "inside" the contour
	(i.e. a collection of line segments) given in \c x and \c
	y. The arrays \c x and \c y must be "ordered" so that adjacent
	points are placed at adjacent entries.  The result is
	undefined if this is not the case or if the contours are not
	properly closed. The first and last points in \c x and \c y
	should be the same to indicate a closed contour. The values 
	xscale and yscale should be an approximate scale for the
	contours \c x and \c y.

	\note This function is deprecated and has been replaced by
	\ref pinside
    */
    bool is_point_inside_old(double x1, double y1, const ovector_view &x,
			     const ovector_view &y, double xscale=0.01,
			     double yscale=0.01);

    /// Free memory
    int free_memory();
    
    /** 
	\brief Check if lines cross

	Returns true if the line segment defined by (x1,y1) and (x2,y2)
	and the line segment defined by (x3,y3) and (x4,y4) have
        an intersection point that is between the endpoints of both
	segments. The function handles vertical and horizontal lines
	appropriately. This function will fail if x1==y1 and x2==y2
	or if x3==y3 and x4==y4, i.e. if the points do not really
	define a line. If the function fails, it returns false. 

	\note This function is deprecated and has been replaced by
	\ref pinside
    */
    bool lines_cross_old(double x1, double y1, double x2, double y2,
			 double x3, double y3, double x4, double y4);

    /// Check to ensure the x- and y-arrays are monotonic
    int check_data();

    /** \brief Smooth the contours by adding internal points using cubic 
	interpolation (this doesn't work)
	
	This makes the contours smoother by adding internal points 
	between each contour line segment determined by cubic spline
	interpolation. 

	For more accurate contours, it is better to provide the 
	original data on a finer grid, or use regrid_data().
    */
    int smooth_contours(size_t nfact);

#endif

  };
  
#ifndef DOXYGENP
}
#endif

#endif


