/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_POLY_H
#define O2SCL_POLY_H

/** \file poly.h
    \brief Classes for solving polynomials

    \warning 
    We should be careful about using pow() in functions using
    complex<double> since pow(((complex<double>)0.0),3.0) returns
    (nan,nan). Instead, we should use pow(((complex<double>)0.0),3)
    which takes an integer for the second argument. The sqrt() function,
    always succeeds i.e. sqrt(((complex<double>)0.0))=0.0
    
    \warning
    One has to be careful about using e.g. pow(a,1.0/3.0)
    for complex a since if Re(a)<0 and Im(a)==0 then the function 
    returns NaN.
*/

#include <iostream>
#include <complex>
#include <gsl/gsl_math.h>
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_poly.h>
#include <o2scl/constants.h>
#include <o2scl/err_hnd.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Solve a quadratic polynomial with real coefficients and 
      real roots
  */
  class quadratic_real {
  public:

    virtual ~quadratic_real() {}

    /** \brief Solves the polynomial \f$ a_2 x^2 + b_2 x + c_2 = 0 \f$ 
	giving the two solutions \f$ x=x_1 \f$ and \f$ x=x_2 \f$ .
    */
    virtual int solve_r(const double a2, const double b2, const double c2, 
			double &x1, double &x2) {return gsl_nobase;};

    /// Return a string denoting the type ("quadratic_real")
    const char *type() { return "quadratic_real"; }
  };

  /** 
      \brief Solve a quadratic polynomial with real coefficients and 
      complex roots
  */
  class quadratic_real_coeff : public quadratic_real {

  public:

    virtual ~quadratic_real_coeff() {}

    /** \brief Solves the polynomial \f$ a_2 x^2 + b_2 x + c_2 = 0 \f$ 
	giving the two solutions \f$ x=x_1 \f$ and \f$ x=x_2 \f$ .
    */
    virtual int solve_r(const double a2, const double b2, const double c2, 
			double &x1, double &x2);

    /** \brief Solves the polynomial \f$ a_2 x^2 + b_2 x + c_2 = 0 \f$ 
	giving the two complex solutions \f$ x=x_1 \f$ and \f$ x=x_2 \f$ 
    */
    virtual int solve_rc(const double a2, const double b2, const double c2, 
			 std::complex<double> &x1, std::complex<double> &x2) 
      { return gsl_nobase; }

    /// Return a string denoting the type ("quadratic_real_coeff")
    const char *type() { return "quadratic_real_coeff"; }
  };

  /** 
      \brief Solve a quadratic polynomial with complex coefficients and 
      complex roots
  */
  class quadratic_complex : public quadratic_real_coeff {
  public:

    virtual ~quadratic_complex() {}

    /** \brief Solves the polynomial \f$ a_2 x^2 + b_2 x + c_2 = 0 \f$ 
	giving the two solutions \f$ x=x_1 \f$ and \f$ x=x_2 \f$ .
    */
    virtual int solve_r(const double a2, const double b2, const double c2, 
			double &x1, double &x2);

    /** \brief Solves the polynomial \f$ a_2 x^2 + b_2 x + c_2 = 0 \f$ 
	giving the two complex solutions \f$ x=x_1 \f$ and \f$ x=x_2 \f$ 
    */
    virtual int solve_rc(const double a2, const double b2, const double c2, 
			 std::complex<double> &x1, std::complex<double> &x2);

    /** \brief Solves the complex polynomial \f$ a_2 x^2 + b_2 x + c_2 = 0 \f$ 
	giving the two complex solutions \f$ x=x_1 \f$ and \f$ x=x_2 \f$ 
    */
    virtual int solve_c(const std::complex<double> a2, 
			const std::complex<double> b2, 
			const std::complex<double> c2, 
			std::complex<double> &x1, 
			std::complex<double> &x2) { return gsl_nobase; }

    /// Return a string denoting the type ("quadratic_complex")
    const char *type() { return "quadratic_complex"; }
  };  

  /** 
      \brief Solve a cubic polynomial with real coefficients and real roots
  */
  class cubic_real {
  public:

    virtual ~cubic_real() {}

    /** \brief Solves the polynomial 
	\f$ a_3 x^3 + b_3 x^2 + c_3 x + d_3= 0 \f$ giving the three 
	solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ , and \f$ x=x_3 \f$ .
    */
    virtual int solve_r(const double a3, const double b3, const double c3, 
			const double d3, double &x1, double &x2, double &x3) 
      { return gsl_nobase; }

    /// Return a string denoting the type ("cubic_real")
    const char *type() { return "cubic_real"; }
  };

  /** 
      \brief Solve a cubic polynomial with real coefficients and complex roots
  */
  class cubic_real_coeff : public cubic_real {

  public:

    virtual ~cubic_real_coeff() {}

    /** \brief Solves the polynomial 
	\f$ a_3 x^3 + b_3 x^2 + c_3 x + d_3= 0 \f$ giving the three 
	solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ , and \f$ x=x_3 \f$ .
    */
    virtual int solve_r(const double a3, const double b3, const double c3, 
			const double d3, double &x1, double &x2, double &x3);

    /** \brief Solves the polynomial 
	 \f$ a_3 x^3 + b_3 x^2 + c_3 x + d_3= 0 \f$ 
	giving the real solution \f$ x=x_1 \f$ and two complex solutions 
	 \f$ x=x_1 \f$ , \f$ x=x_2 \f$ , and \f$ x=x_3 \f$ .
    */
    virtual int solve_rc(const double a3, const double b3, const double c3, 
			 const double d3, double &x1, std::complex<double> &x2,
			 std::complex<double> &x3) { return gsl_nobase; }

    /// Return a string denoting the type ("cubic_real_coeff")
    const char *type() { return "cubic_real_coeff"; }
  };

  /** 
      \brief Solve a cubic polynomial with complex coefficients and 
      complex roots
  */
  class cubic_complex : public cubic_real_coeff {

  public:

    virtual ~cubic_complex() {}

    /** \brief Solves the polynomial 
	\f$ a_3 x^3 + b_3 x^2 + c_3 x + d_3= 0 \f$ giving the three 
	solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ , and \f$ x=x_3 \f$ .
    */
    virtual int solve_r(const double a3, const double b3, const double c3, 
			const double d3, double &x1, double &x2, double &x3);

    /** \brief Solves the polynomial 
	\f$ a_3 x^3 + b_3 x^2 + c_3 x + d_3= 0 \f$ giving the real 
	solution \f$ x=x_1 \f$ and two complex solutions 
	\f$ x=x_1 \f$ , \f$ x=x_2 \f$ , and \f$ x=x_3 \f$ .
    */
    virtual int solve_rc(const double a3, const double b3, const double c3, 
			 const double d3, double &x1, std::complex<double> &x2,
			 std::complex<double> &x3);

    /** Solves the complex polynomial 
	\f$ a_3 x^3 + b_3 x^2 + c_3 x + d_3= 0 \f$ 
	giving the three complex solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ , and 
	\f$ x=x_3 \f$ .
    */
    virtual int solve_c(const std::complex<double> a3, 
			const std::complex<double> b3, 
			const std::complex<double> c3, 
			const std::complex<double> d3, 
			std::complex<double> &x1, std::complex<double> &x2, 
			std::complex<double> &x3) { return gsl_nobase; }

    /// Return a string denoting the type ("cubic_complex")
    const char *type() { return "cubic_complex"; }
  };

  /** 
      \brief Solve a quartic polynomial with real coefficients and real roots
  */
  class quartic_real {

  public:

    virtual ~quartic_real() {}
    
    /** Solves the polynomial 
	\f$ a_4 x^4 + b_4 x^3 + c_4 x^2 + d_4 x + e_4 = 0 \f$ 
	giving the four solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ ,
	\f$ x=x_3 \f$ , and \f$ x=x_4 \f$ .
    */
    virtual int solve_r(const double a4, const double b4, const double c4, 
			const double d4, const double e4, 
			double &x1, double &x2, 
			double &x3, double &x4) { return gsl_nobase; }

    /// Return a string denoting the type ("quartic_real")
    const char *type() { return "quartic_real"; }
  };

  /** 
      \brief Solve a quartic polynomial with real coefficients and 
      complex roots
  */
  class quartic_real_coeff : public quartic_real {

  public:

    virtual ~quartic_real_coeff() {}

    /** Solves the polynomial 
	\f$ a_4 x^4 + b_4 x^3 + c_4 x^2 + d_4 x + e_4 = 0 \f$ 
	giving the four solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ ,
	\f$ x=x_3 \f$ , and \f$ x=x_4 \f$ .
    */
    virtual int solve_r(const double a4, const double b4, const double c4, 
			const double d4, const double e4, double &x1, 
			double &x2, double &x3, double &x4);

    /**  Solves the polynomial 
	 \f$ a_4 x^4 + b_4 x^3 + c_4 x^2 + d_4 x + e_4 = 0 \f$ 
	 giving the four complex solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ ,
	 \f$ x=x_3 \f$ , and \f$ x=x_4 \f$ .
    */
    virtual int solve_rc(const double a4, const double b4, const double c4, 
			 const double d4, const double e4, 
			 std::complex<double> &x1, std::complex<double> &x2, 
			 std::complex<double> &x3, std::complex<double> &x4) 
      { return gsl_nobase; }

    /// Return a string denoting the type ("quartic_real_coeff")
    const char *type() { return "quartic_real_coeff"; }
  };

  /** 
      \brief Solve a quartic polynomial with complex coefficients and 
      complex roots
  */
  class quartic_complex : public quartic_real_coeff {

  public:

    virtual ~quartic_complex() {}

    /** Solves the polynomial 
	\f$ a_4 x^4 + b_4 x^3 + c_4 x^2 + d_4 x + e_4 = 0 \f$ 
	giving the four solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ ,
	\f$ x=x_3 \f$ , and \f$ x=x_4 \f$ .
    */
    virtual int solve_r(const double a4, const double b4, const double c4, 
			const double d4, const double e4, double &x1, 
			double &x2, 
			double &x3, double &x4);

    /**  Solves the polynomial 
	 \f$ a_4 x^4 + b_4 x^3 + c_4 x^2 + d_4 x + e_4 = 0 \f$ 
	 giving the four complex solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ ,
	 \f$ x=x_3 \f$ , and \f$ x=x_4 \f$ .
    */
    virtual int solve_rc(const double a4, const double b4, const double c4, 
			 const double d4, const double e4, 
			 std::complex<double> &x1, std::complex<double> &x2, 
			 std::complex<double> &x3, std::complex<double> &x4);

    /** Solves the complex polynomial 
	 \f$ a_4 x^4 + b_4 x^3 + c_4 x^2 + d_4 x + e_4 = 0 \f$ 
	giving the four complex solutions \f$ x=x_1 \f$ , \f$ x=x_2 \f$ ,
	 \f$ x=x_3 \f$ , and \f$ x=x_4 \f$ .
    */
    virtual int solve_c(const std::complex<double> a4, 
			const std::complex<double> b4, 
			const std::complex<double> c4, 
			const std::complex<double> d4, 
			const std::complex<double> e4, 
			std::complex<double> &x1, 
			std::complex<double> &x2, std::complex<double> &x3,
			std::complex<double> &x4) { return gsl_nobase; }

    /// Return a string denoting the type ("quartic_complex")
    const char *type() { return "quartic_complex"; }
  };

  /** \brief Base class for solving a general polynomial with real
      coefficients and complex roots
  */
  class poly_real_coeff : public quadratic_real_coeff,
    public cubic_real_coeff, public quartic_real_coeff {

  public:
    
    virtual ~poly_real_coeff() {}
    
    /** 
	\brief Solve the n-th order polynomial
	
        The coefficients are stored in co[], with the leading coefficient
	as co[0] and the constant term as co[n]. The roots are returned
	in ro[0],...,ro[n-1].
    */
    virtual int solve_rc(int n, const double co[], 
			 std::complex<double> ro[]) { return gsl_nobase; }

    /// Polish the roots 
    virtual int polish_rc(int n, const double co[],
			  std::complex<double> *ro) { return gsl_nobase; }

    /// Return a string denoting the type ("poly_real_coeff")
    const char *type() { return "poly_real_coeff"; }
  };

  /** \brief Base class for solving a general polynomial with complex
      coefficients
  */
  class poly_complex : public quadratic_complex,
    public cubic_complex, public quartic_complex {

    public:

    virtual ~poly_complex() {}
    
    /** 
	\brief Solve the n-th order polynomial
	
        The coefficients are stored in co[], with the leading coefficient
	as co[0] and the constant term as co[n]. The roots are returned
	in ro[0],...,ro[n-1].
    */
    virtual int solve_c(int n, const std::complex<double> co[], 
			std::complex<double> ro[]) { return gsl_nobase; }
    
    /// Polish the roots 
    virtual int polish_c(int n, const std::complex<double> co[],
			 std::complex<double> *ro) { return gsl_nobase; }

    /// Return a string denoting the type ("poly_complex")
    const char *type() { return "poly_complex"; }
  };

  /** 
      \brief Solve a cubic with real coefficients and complex roots (CERNLIB)
  */
  class cern_cubic_real_coeff : public cubic_real_coeff {

  public:

    virtual ~cern_cubic_real_coeff() {}

    virtual int solve_rc(const double a3, const double b3, const double c3, 
			 const double d3, double &x1, 
			 std::complex<double> &x2, std::complex<double> &x3);

    /// The original CERNLIB interface
    virtual int rrteq3(double r, double s, double t, double x[], double &d);

    /// Return a string denoting the type ("cern_cubic_real_coeff")
    const char *type() { return "cern_cubic_real_coeff"; }
  };

  /** 
      \brief Solve a quartic with real coefficients and complex roots (CERNLIB)
  */
  class cern_quartic_real_coeff : public quartic_real_coeff {

  public:

    virtual ~cern_quartic_real_coeff() {}

    virtual int solve_rc(const double a4, const double b4, const double c4, 
			 const double d4, const double e4, 
			 std::complex<double> &x1, std::complex<double> &x2, 
			 std::complex<double> &x3, std::complex<double> &x4);

    /// The original CERNLIB interface
    virtual int rrteq4(double a, double b, double c, double d, 
		       std::complex<double> z[], double &dc, 
		       int &mt);

    /// Return a string denoting the type ("cern_quartic_real_coeff")
    const char *type() { return "cern_quartic_real_coeff"; }

#ifndef DOXYGEN_INTERNAL

  protected:

    /// The object to solve for the associated cubic
    cern_cubic_real_coeff cub_obj;

#endif

  };

  /** 
      \brief Solve a quadratic with real coefficients and complex roots (GSL)
  */
  class gsl_quadratic_real_coeff : public quadratic_real_coeff {

  public:

    virtual ~gsl_quadratic_real_coeff() {}

    virtual int solve_rc(const double a2, const double b2, const double c2, 
			 std::complex<double> &x1, std::complex<double> &x2);

    /// Return a string denoting the type ("gsl_quadratic_real_coeff")
    const char *type() { return "gsl_quadratic_real_coeff"; }

  };

  /** 
      \brief Solve a cubic with real coefficients and complex roots (GSL)
  */
  class gsl_cubic_real_coeff : public cubic_real_coeff {

  public:

    virtual ~gsl_cubic_real_coeff() {}

    virtual int solve_rc(const double a3, const double b3, const double c3, 
			 const double d3, double &x1, 
			 std::complex<double> &x2, 
			 std::complex<double> &x3);

    /// Return a string denoting the type ("gsl_cubic_real_coeff")
    const char *type() { return "gsl_cubic_real_coeff"; }

#ifndef DOXYGEN_INTERNAL

  protected:

  /** 
      \brief An alternative to \c gsl_poly_complex_solve_cubic()

      This is an alternative to the function \c
      gsl_poly_complex_solve_cubic() with some small corrections.
      
      \todo Demonstrate that this works 
      better than the original GSL version.
   */
  int gsl_poly_complex_solve_cubic2(double a, double b, double c, 
				    gsl_complex *z0, gsl_complex *z1, 
				    gsl_complex *z2);

#endif
  
  };

  /** 
      \brief Solve a quartic with real coefficients and real roots (GSL)
  */
  class gsl_quartic_real : public quartic_real {
    
  public:
    
    virtual ~gsl_quartic_real() {}

    virtual int solve_r(const double a4, const double b4, const double c4, 
			const double d4, const double e4, double &x1, 
			double &x2, 
			double &x3, double &x4);

    /// Return a string denoting the type ("gsl_quartic_real")
    const char *type() { return "gsl_quartic_real"; }

  };

  /** 
      \brief Solve a quartic with real coefficients and real roots (GSL)

      \todo Document the distinction between this class and
      \ref gsl_quartic_real
  */
  class gsl_quartic_real2 : public quartic_real {

  public:

    virtual ~gsl_quartic_real2() {}

    virtual int solve_r(const double a4, const double b4, const double c4, 
			const double d4, const double e4, double &x1, 
			double &x2, 
			double &x3, double &x4);

    /// Return a string denoting the type ("gsl_quartic_real2")
    const char *type() { return "gsl_quartic_real2"; }
  };

  /** 
      \brief Solve a general polynomial with real coefficients (GSL)
  */
  class gsl_poly_real_coeff : public poly_real_coeff {

  public:

    gsl_poly_real_coeff();

    virtual ~gsl_poly_real_coeff();

    virtual int solve_rc(int n, const double co[], 
			 std::complex<double> ro[]);

    virtual int solve_rc(const double a3, const double b3, const double c3, 
			 const double d3, double &x1, 
			 std::complex<double> &x2, 
			 std::complex<double> &x3);

    virtual int solve_rc(const double a2, const double b2, const double c2, 
			 std::complex<double> &x1, 
			 std::complex<double> &x2);

    virtual int solve_rc(const double a4, const double b4, const double c4, 
			 const double d4, const double e4, 
			 std::complex<double> &x1, std::complex<double> &x2, 
			 std::complex<double> &x3, std::complex<double> &x4);

    /// Return a string denoting the type ("gsl_poly_real_coeff")
    const char *type() { return "gsl_poly_real_coeff"; }

  protected:

#ifndef DOXYGEN_INTERNAL

    // Workspace for quadratic polynomials
    gsl_poly_complex_workspace *w2;

    // Workspace for cubic polynomials
    gsl_poly_complex_workspace *w3;

    // Workspace for quartic polynomials
    gsl_poly_complex_workspace *w4;

    // Workspace for general polynomials
    gsl_poly_complex_workspace *wgen;

    /// The size of the workspace \ref wgen
    int gen_size;

#endif

  };

  /** 
      \brief Solve a quadratic with complex coefficients and complex roots
  */
  class quadratic_std_complex : public quadratic_complex {

  public:

    virtual ~quadratic_std_complex() {}

    virtual int solve_c(const std::complex<double> a2, 
			const std::complex<double> b2, 
			const std::complex<double> c2, 
			std::complex<double> &x1, std::complex<double> &x2);

    /// Return a string denoting the type ("quadratic_std_complex")
    const char *type() { return "quadratic_std_complex"; }
  };

  /** 
      \brief Solve a cubic with complex coefficients and complex roots
  */
  class cubic_std_complex : public cubic_complex {

  public:

    virtual ~cubic_std_complex() {}

    virtual int solve_c(const std::complex<double> a3, 
			const std::complex<double> b3, 
			const std::complex<double> c3, 
			const std::complex<double> d3, 
			std::complex<double> &x1, std::complex<double> &x2, 
			std::complex<double> &x3);

    /// Return a string denoting the type ("cubic_std_complex")
    const char *type() { return "cubic_std_complex"; }
  };

  /** 
      \brief Solve a quartic with real coefficients and real roots
      
      \todo 3/8/07 - Compilation at the NSCL produced non-finite values
      in solve_r() for some values of the coefficients. This should
      be checked.
      \todo It looks like this code is tested only for a4=1, and if 
      so, the tests should be generalized.
      \todo Also, there is a hard-coded number in here (\f$ 10^{-6} \f$),
      which might be moved to a data member?
  */
  class naive_quartic_real : public quartic_real {

  public:

    virtual ~naive_quartic_real() {}

    virtual int solve_r(const double a4, const double b4, const double c4, 
			const double d4, const double e4, double &x1, 
			double &x2, double &x3, double &x4);

    /// Return a string denoting the type ("naive_quartic_real")
    const char *type() { return "naive_quartic_real"; }
  };

  /** 
      \brief Solve a quartic with complex coefficients and complex roots
  */
  class naive_quartic_complex : public quartic_complex {

  public:

    virtual ~naive_quartic_complex() {}

    virtual int solve_c(const std::complex<double> a4, 
			const std::complex<double> b4, 
			const std::complex<double> c4, 
			const std::complex<double> d4, 
			const std::complex<double> e4, 
			std::complex<double> &x1, 
			std::complex<double> &x2, std::complex<double> &x3,
			std::complex<double> &x4);

    /// Return a string denoting the type ("naive_quartic_complex")
    const char *type() { return "naive_quartic_complex"; }

#ifndef DOXYGENP

  protected:

    /// The object to solve for the associated cubic
    cubic_std_complex cub_obj;
    
#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
