/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_TIMER_H
#define O2SCL_TIMER_H

#include <iostream>
#include <string>
#include <o2scl/string_conv.h>
#include <o2scl/misc.h>
#include <sys/time.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Provide an interface for timing execution using 
      \c gettimeofday()

      \todo Better testing which doesn't use a fixed number of
      mathematical operations, but automatically selects enough
      operations.
  */
  class timer_gettod {
  public:
    
#ifdef NEVER_DEFINED
    
    // typical definition of timeval (for reference only)
    struct timeval {
      // seconds
      time_t         tv_sec;        
      // microseconds
      suseconds_t    tv_usec;  
    };
    
    // typical definition of timezone (for reference only)
    struct timezone {
      // minutes W of Greenwich
      int  tz_minuteswest; 
      // type of dst correction
      int  tz_dsttime;     
    };
    
#endif
    
    timer_gettod() { reset(); }
    
    /// Set time 'zero'
    int reset() { return gettimeofday(&zero,&tz); }

    /// Store the present time
    int set() { return gettimeofday(&mark,&tz); }
    
    /// Return the number of seconds between set() and reset()
    double seconds_elapsed();
    
    /// Return the time between set() and reset()
    int time_elapsed(int &d, int &h, int &m, int &s, int &usec);
    
    /// Time remaining if \c n out of \c tot tasks have been completed
    int time_remaining(int n, int ntot, 
		       int &d, int &h, int &m, int &s, int &usec);
    
    /// Time remaining if \c n out of \c tot tasks have been completed
    std::string time_remaining(int n, int ntot);
    
    /// Convert a time interval to a string
    std::string interval_to_string(int d, int h, int m, int s, int usec);

  protected:

#ifndef DOXYGEN_INTERNAL

    /// The last time the clock was reset
    struct timeval zero;

    /// The most resent time from set()
    struct timeval mark;

    /// The timezone
    struct timezone tz;

#endif

  };

  /** 
      \brief Provide an interface for timing execution using 
      \c clock()
      
      \note Note that the time return by clock() is reset on
      some regular interval (sometimes 72 minutes) and this
      class does not yet account for this.
  */
  class timer_clock {

  public:
    timer_clock();

    /// Number of seconds elapsed
    double time_since();

    /** 
	\brief Time elapsed in days, hours, minutes, seconds, and 
	fractions of seconds
    */
    void time_since(int &d, int &h, int &m, int &s, double &f);
  
    /// Time remaining if \c n out of \c tot tasks have been completed
    void time_remaining(int n, int tot, int &d, int &h, int &m, int &s,
			double &f);

    /// Convert a time interval to a string
    std::string interval_to_string(int d, int h, int m, int s, double f=0.0);

  protected:

#ifndef DOXYGEN_INTERNAL

    /// Desc
    clock_t time;

#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
