/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_EFF_BOSON_H
#define O2SCL_EFF_BOSON_H

#include <string>
#include <iostream>
#include <fstream>
#include <cmath>
#include <o2scl/constants.h>
#include <o2scl/funct.h>
#include <o2scl/mm_funct.h>
#include <o2scl/root.h>
#include <o2scl/mroot.h>
#include <o2scl/inte.h>
#include <o2scl/boson.h>
#include <o2scl/cern_mroot_root.h>
#include <o2scl/gsl_mroot_hybrids.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Boson class from fitting method

      The constructor loads the coefficients from the file \c boselat3 
      by default. If this is not successful, then
      is_loaded() will return false.

      \todo Better documentation (see eff_fermion)
      \todo Remove native error codes
  */
  /*
    The pdf documentation fails if these lines are added.
    
    \todo Use file_detect instead of text_in_file.
  */
  class eff_boson : public boson {

  public:

    /// Create a boson with mass \c m  and degeneracy \c g 
    eff_boson(double m=0.0, double g=0.0);
    virtual ~eff_boson();
  
    /** \brief Load coefficients for finite-temperature approximation
      
	Presently acceptable values of \c fn are: \c boselat3 from
	Lattimer's notes \c bosejel21, \c bosejel22, \c bosejel34, and
	\c bosejel34cons from \ref Johns96.
      
	boselat3
	\include boselat3
	bosejel21
	\include bosejel21
	bosejel22
	\include bosejel22
	bosejel34
	\include bosejel34
	bosejel34cons
	\include bosejel34cons 
    */
    static int loadcoeff(std::string bfile);
  
    virtual int calc_mu(const double temper);
    virtual int calc_density(const double temper);
    virtual int pair_mu(const double temper);
    virtual int pair_density(const double temper);

    /** \brief Set the solver for use in calculating \f$ \psi \f$ */
    int set_psi_root(root<void *,funct<void *> > &rp) {
      psi_root=&rp;
      return 0;
    }

    /** \brief Set the solver for use in calculating the chemical
	potential from the density */
    int set_density_mroot(mroot<void *,mm_funct<void *> > &rp) {
      density_mroot=&rp;
      return 0;
    }

    /** \brief Set the solver for use in calculating the chemical
	potential from the density (meth2=true) */
    int set_meth2_root(root<void *,funct<void *> > &rp) {
      meth2_root=&rp;
      return 0;
    }

    /** \brief The default solver for calc_density() and pair_density()
    */
    gsl_mroot_hybrids<void *,mm_funct<void *> > def_density_mroot;

    /** \brief The default solver for \f$ \psi \f$
     */
    cern_mroot_root<void *,funct<void *> > def_psi_root;

    /** \brief The default solver for calc_density() and pair_density()
    */
    cern_mroot_root<void *,funct<void *> > def_meth2_root;

    virtual const char *type() { return "eff_boson"; }

#ifndef DOXYGEN_INTERNAL

    friend class io_tlate<eff_boson>;

  protected:
  
    /// The coefficients
    static double **Pmnb;
    /// The number of coefficient rows
    static int sizem;
    /// The number of coefficient columns
    static int sizen;
    /// The parameter, \f$ a \f$
    static double parma;
    /// Temporary storage
    static double fix_density;
    /// Temporary storage
    static double stat_temper;
    /// True if coefficients have been loaded
    static bool loaded;

    /// The solver for calc_density()
    mroot<void *,mm_funct<void *> > *density_mroot;
    /// The solver to compute \f$ h \f$ from \f$ \psi \f$.
    root<void *,funct<void *> > *psi_root;
    /// The solver for calc_density()
    root<void *,funct<void *> > *meth2_root;

    /// The function which solves for \f$ h \f$ from \f$ \psi \f$.
    int solve_fun(double x, double &y, void *&pa);

    /// Fix density for calc_density()
    int density_fun(size_t nv, const ovector_view &x, ovector_view &y, 
		    void *&pa);

    /// Fix density for pair_density()
    int pair_density_fun(size_t nv, const ovector_view &x, ovector_view &y, 
			 void *&pa);

#endif
  };

  template<> int io_tlate<eff_boson>::input
    (cinput *co, in_file_format *ins, eff_boson *b);
  template<> int io_tlate<eff_boson>::output
    (coutput *co, out_file_format *outs, eff_boson *b);
  template<> const char *io_tlate<eff_boson>::type();
  
  typedef io_tlate<eff_boson> eff_boson_io_type;

#ifndef DOXYGENP
}
#endif

#endif
