/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_CERN_ROOT_H
#define O2SCL_CERN_ROOT_H

#include <o2scl/root.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief One-dimensional root-finding routine (CERNLIB)
      
      This class attempts to find \f$ x_0 \f$ and \f$ x_1 \f$ in \f$
      [a,b] \f$ such that \f$ f(x_0) f(x_1) \leq 0 \f$, \f$ |f(x_0)|
      \leq|f(x_1)| \f$, and \f$ | x_0-x_1| \leq
      2~\mathrm{tolx}~(1+|x_0|) \f$.
      
      The variable cern_root::tolx defaults to \f$ 10^{-8} \f$ and
      cern_root::ntrial defaults to 200.  
      \comment 
      I'm not sure why I chose these values for tolx and ntrial above,
      as it doesn't seem to me that these values are suggested in
      CERNLIB. However, they're reasonable so I'll leave them in for
      now. 
      \endcomment
	
      solve_bkt() returns 0 for success, \ref gsl_einval if the root
      is not initially bracketed, and \ref gsl_emaxiter if the number
      of function evaluations is greater than cern_root::ntrial.
    
      Based on CERNLIB routine DZEROX which is based on \ref Bus75.
    
  */
  template<class param_t, class func_t=funct<param_t> > 
    class cern_root : public root<param_t,func_t>
    {

#ifndef DOXYGEN_INTERNAL

      protected:
      
      /** 
	  \brief Internal storage for the mode

	  This internal variable is actually defined to be smaller by
	  1 than the "mode" as it is defined in the CERNLIB
	  documentation in order to avoid needless subtraction in
	  solve_bkt().
      */
      int mode;
      
      /// FORTRAN-like function for sign
      inline double sign(double a, double b) {
	if (b>=0.0) return fabs(a);
	return -fabs(a);
      }
      
      friend class io_tlate<cern_root>;
      
#endif
      
      public:
      
      cern_root()
      {
	mode=0;
	this->tolx=1.0e-8;
	this->ntrial=200;
	this->over_bkt=true;
      }
      
      /** 
	  \brief Set mode of solution (1 or 2)
         
	  - \c 1 should be used for simple functions where the cost is
	  inexpensive in comparison to one iteration of solve_bkt(),
	  or functions which have a pole near the root (this is the
	  default).
	  - \c 2 should be used for more time-consuming functions.
 
	  If an integer other than \c 1 or \c 2 is specified, \c 1 is
	  assumed.
         
      */
      int set_mode(int m) {
	if (m!=1 && m!=2) {
	  mode=0;
	  set_err_ret("Invalid mode in cern_root::set_mode().",
		      o2scl::gsl_einval);
	}
	mode=m-1;
	return 0;
      }

      /// Return the type, \c "cern_root".
      virtual const char *type() { return "cern_root"; }

      /** \brief Solve \c func in region \f$ x_1<x<x_2 \f$ returning  
	  \f$ x_1 \f$.
      */
      virtual int solve_bkt(double &x1, double x2, param_t &pa, 
			    func_t &func) 
      {
	
	double im1[2]={2.0,3.0}, im2[2]={-1.0,3.0}, c=0.0, fa, fb;
	double atl, a, b, mft;
	double fc=0.0, d=0.0, fd=0.0, tol, h, hb, w, p, q, fdb, fda, f=0.0;
	bool lmt[2];
	int ie=0, loop, mf;
	char ch;
    
	func(x1,fb,pa);
	func(x2,fa,pa);
	
	if (fa*fb>0.0) {
	  set_err_ret
	    ("Endpoints don't bracket function in cern_root::solve_bkt().",
	     o2scl::gsl_einval);
	}
	
	atl=fabs(this->tolx);
	b=x1;
	a=x2;
	lmt[1]=true;
	mf=2;
	loop=1;
	do {
	  if (loop==1) {
	    c=a;
	    fc=fa;
	    ie=0;
	  } else if (loop==2) {
	    ie=0;
	  }
	  if (fabs(fc)<fabs(fb)) {
	    if (c!=a) {
	      d=a;
	      fd=fa;
	    }
	    a=b;
	    b=c;
	    c=a;
	    fa=fb;
	    fb=fc;
	    fc=fa;
	  }
	  tol=atl*(1.0+fabs(c));
	  h=0.5*(c+b);
	  hb=h-b;
      
	  if (this->verbose>0) {
	    this->print_iter(c,fc,mf-2,fabs(hb),tol,"cern_root");
	  }

	  if (fabs(hb)>tol) {
	    if (ie>im1[mode]) {
	      w=hb;
	    } else {
	      tol*=sign(1.0,hb);
	      p=(b-a)*fb;
	      lmt[0]=(ie<=1);
	      if (lmt[mode]) {
		q=fa-fb;
		lmt[1]=false;
	      } else {
		fdb=(fd-fb)/(d-b);
		fda=(fd-fa)/(d-a);
		p*=fda;
		q=fdb*fa-fda*fb;
	      }
	      if (p<0.0) {
		p=-p;
		q=-q;
	      }
	      if (ie==im2[mode]) p+=p;
	      if (p==0.0 || p<=q*tol) {
		w=tol;
	      } else if (p<hb*q) {
		w=p/q;
	      } else {
		w=hb;
	      }
	    }
	    d=a;
	    a=b;
	    fd=fa;
	    fa=fb;
	    b+=w;
	    mf++;
	    if (mf>this->ntrial) {
	      set_err_ret
		("Too many function calls in cern_root::solve_bkt().",
		 o2scl::gsl_emaxiter);
	    }

	    func(b,fb,pa);

	    if (fb==0.0 || sign(1.0,fc)==sign(1.0,fb)) {
	      loop=1;
	    } else if (w==hb) {
	      loop=2;
	    } else {
	      ie++;
	      loop=3;
	    }
	  } else {
	    loop=0;
	  }
	} while (loop>0);
	x1=c;
	this->last_ntrial=mf;
	return o2scl::gsl_success;
      }
            
    };

#ifndef DOXYGENP
  template<> int io_tlate<cern_root<void *,funct<void *> > >::input
    (cinput *co, in_file_format *ins, cern_root<void *, funct<void *> > *ro);
  template<> int io_tlate<cern_root<void *,funct<void *> > >::output
    (coutput *co, out_file_format *outs, 
     cern_root<void *, funct<void *> > *ro);
  template<> const char *io_tlate<cern_root<void *,funct<void *> > >::type();
#endif
  
  typedef io_tlate<cern_root<void *,funct<void *> > > cern_root_io_type;

#ifndef DOXYGENP
}
#endif
  
#endif
